function [p,pe] = fmm3d(z,m,ze,varargin)
%FMM3D Fast multipole method in 3 dimensions.
%   P = FMM3D(Z,M) evaluates the potential P due to masses M at
%   Cartesian coordinates Z = (x,y,z). Z is an N-by-3 matrix and M is
%   a length N real vector.
%
%   PE = FMM3D(Z,M,ZE) evaluates the potential at the points ZE
%   instead, where ZE is an NE-by-3 matrix.
%
%   [P,PE] = FMM3D(Z,M,ZE) evaluates the potential both at Z and ZE at
%   the same time.
%
%   P = FMM3D(Z,M,...) or [...] = FMM3D(Z,M,ZE,...) allows options to 
%   be passed in the form of property/value-pairs. See the table below 
%   for available properties.
%
%   Property     Value/{Default}     Description
%   -----------------------------------------------------------------------
%   tol          Scalar {1e-6}       Relative tolerance. The choice
%                                    tol = 0 means a direct evaluation.
%
%   out          {'sol'} | 'mesh' |  Output solution/mesh/timings. Type
%                'time'              'type fmm3d' for more
%                                    information.
%
%   ndirect      Integer {100}       The number of particles in each
%                                    box at the finest level.
%
%   nlevels      Integer >= 0        The number of levels in the
%                                    multipole tree, if ndirect is not
%                                    specified.
%
%   theta        Scalar in (0,1)     Value of theta in the
%                {0.5}               theta-criterion for well
%                                    separated sets, R+theta*r <=
%                                    theta*d.
%
%   pcoeff       Integer > 0         Order of multipole
%                                    expansion, if tol is not specified.
%
%   Experimental or partially supported options:
%
%   eta          Scalar in [0,1]     Each box is split around the point 
%                {1}                 eta*xmed+(1-eta)*xmid with
%                                    xmed/xmid the median/midpoint,
%                                    respectively.
%
%   pot          {1}                 1: Coulomb potential,
%                                    P = M/norm(Z).
%
%   smooth       {'dirac'}           Near-field model.
%
%   Example:
%     % N random masses
%     N = 100;
%     z = randn(N,3);
%     m = rand(N,1);
%
%     % direct evaluation of the potential
%     p = fmm3d(z,m,'tol',0);

% J. Bull 2015-09-11 (switch to 3D)
% S. Engblom 2012-05-03 (Revision, 2 more outputs in the mesh-syntax)
% S. Engblom 2009-10-12 (Revision, performance)
% S. Engblom 2009-04-16 (Major revision, adaptivity)
% S. Engblom 2007-07-08 (Major revision)
% S. Engblom 2007-05-15 (Revision)
% S. Engblom 2006-07-10 (Based on the code 'fmm2dlp' by the same author)

% Notes:
%   -The call P = FMM3D(Z,M,Z,...) is interpreted as P = FMM3D(Z,M,...)
%   through comparision of pointers.
%
%   -The syntax M = FMM3D(...,'out','mesh') returns the FMM mesh in a simple
%   format (see below and see also MESHPLOT for an example). This syntax is
%   mainly intended for debugging and pedagogical use.
%
%   [M,C] = FMM3D(...,'out','mesh') additionally returns cell vectors
%   M and C containing mesh- and sparse connection matrices.

%   [M,C,S,T] = FMM2D(...,'out','mesh') additionally returns cell
%   vectors M and C containing mesh- and sparse connection matrices,
%   as well as sparse matrices S and T representing the sorting of
%   potentials Z and evaluation points E into boxes at the finest
%   level of the multipole mesh.

%   M{l} for l = 1..nlevels contains 6-by-Nl real matrices where Nl =
%   8^(l-1) is the number of boxes at level l. M{l}(1:3,:) are the
%   coordinates of the center point for each box and M{l}(4:6,:)
%   contains the components of the vector from center to upper right
%   back corner.
%
%   C{l}(i,j) = 1 if box i and j are strongly connected (interact
%   directly), and C{l}(i,j) = -1 whenever box i and j interact
%   through cluster to cluster interactions at level l. If
%   all(C{l}(i,j+1:8) = -1), then a cluster to cluster interaction via
%   the parent of j is used and C{l}(i,j+1:8) = -2 in order to specify
%   this. At the final level, C{nlevel}(i,j) = 2 if box i and j
%   interact through cluster to particle (or particle to cluster)
%   interactions. Only the strictly lower triangular part of the
%   matrix is stored. If required, C{i} =
%   speye(size(C{i}))+C{i}''+C{i}' yields the complete matrix.
%
%   S(:,i) is 1 for all index j such that the point Z(j) belongs to
%   box i. Similarly, T(:,i) is 1 for all index j such that the
%   evaluation point E(j) belongs to box i.
%
%   See also MESHPLOT, MESHSTATS, MESHQUAL.
%
% Time-format:
%   T = FMM2D(...,'out','time') returns a column vector of length 32
%   containing precise timings of the algorithm. The exact content of
%   T depends on both the input and on the current compile. Currently,
%   the clock() function from time.h is used.
%
%   Output vector, when 'tol' = 0:
%     1:
%     ...
%
%   Output vector, when 'tol' > 0:
%     1:
%     2:
%     3:
%     4:
%     5:
%     ...

error('.MEX-file not found on path.');
