/* cudaeval.cu -- CUDA host code for direct evaluation. */

/* S. Engblom and A. Goude 2011-10-19 */

//#ifdef CUDASUPPORT
#include "fmm.h"
#include "cuda.h"
#include "cudaeval.h"

/*------------------------------------------------------------------------*/
#define CUDABLOCKSPERMP 8 //blocks per multiprocessor. Used for determining number of threads in some algorithms
/*------------------------------------------------------------------------*/
static __inline__ __device__ double fetch_double(texture<int2,1> t,int i)
{
  int2 v = tex1Dfetch(t,i);
  return __hiloint2double(v.y,v.x);
}
/*static __inline__ __device__ void fetch_double2(double2 *dest,texture<int4,1> t,int i)
{
  int4 v = tex1Dfetch(t,i);
  *dest= make_double2(__hiloint2double(v.y,v.x),__hiloint2double(v.w,v.z));
  }*/

#define TEXTUREFETCH(dest,texref,index) dest=fetch_double(texref,index)

int cudaalloccount; // remembers the number of cuda allocations

cudaError_t cudaMallocDebug(void **addr,size_t size) 
// GPU memory allocation debug, can track memory leaks
{
  cudaalloccount++;
  cudaError_t ret = cudaMalloc(addr,size);
  if (*addr == NULL) cudaalloccount--;
  return ret;
}
/*------------------------------------------------------------------------*/
cudaError_t cudaFreeDebug(void* addr)
// corresponding to CudaMallocDebug, all allocations should be
// performed through these calls
{
  if (addr != NULL) cudaalloccount--;
  return cudaFree(addr);
}
/*------------------------------------------------------------------------*/
void printcudaalloccount(char *message)
// for tracking memory leaks
{
  mexPrintf("%s %d memory allocations on GPU.\n",message,cudaalloccount);
}
/*------------------------------------------------------------------------*/
int getcudaalloccount()
// since the variable currently only is declared in this file.
{
  return cudaalloccount;
}
/*------------------------------------------------------------------------*/
void resetalloccount()
// since the variable currently only is declared in this file.
{
  cudaalloccount=0;;
}
/*------------------------------------------------------------------------*/
void cudasafe2(cudaError_t error,const char *message,const char* command,const char *file,int line)
// checks if a cuda function succeded, if not, print error message and
// abort (Cuda functions fail a lot more often than normal cpu
// functions. it is important to check the return values)
{
  char errormessage[1000];
  if (error != cudaSuccess) {
    snprintf(errormessage,999,"%s : %s (%i) \nCommand: %s\nFile %s at line %d",message,
             cudaGetErrorString(error),error,command,file,line);
    cudaGetLastError(); // remove the error message
    cudaThreadExit();
    resetalloccount();//cudaThreadExit() should deallocate variables
    mexErrMsgTxt(errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void cudasafeMalloc(void** addr,size_t size)
// safe memory allocation, cudasafe(cudaMalloc(...),"message") can
// also be used
{
  char errormessage[1000];
  cudaError_t error = cudaMallocDebug(addr,size);
  if(error != cudaSuccess) {
    sprintf(errormessage,"cudaMalloc (%u bytes): %s (%i) \n",
            (unsigned)size,cudaGetErrorString(error),error);
    cudaGetLastError(); //remove the error message
    cudaThreadExit();
    mexErrMsgTxt(errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void checkcudaerror(const char *message)
{
  //only check for error and abort if error is found
  char errormessage[1000];
  cudaError_t error=cudaGetLastError();
  if(error!=cudaSuccess) {
    sprintf(errormessage,"%s : %s (%i) \n",message,
            cudaGetErrorString(error),error);
    cudaThreadExit();
    mexErrMsgTxt(errormessage); 
  }
}
/*------------------------------------------------------------------------*/
void cudareset()
{
  cudaThreadExit();
}
/*------------------------------------------------------------------------*/
void cudastart()
{
  char errormessage[1000];
  cudaError_t error=cudaGetLastError();
  if(error!=cudaSuccess) {
    sprintf(errormessage,"Cuda initialization error: %s (%i) \n",
            cudaGetErrorString(error),error);
    cudaThreadExit();
    mexWarnMsgTxt(errormessage); 
  }
}

/*------------------------------------------------------------------------*/

//textures have to be global
texture<int2,1> tex;
texture<int2,1> tey;
texture<int2,1> tez;
texture<int2,1> tzx;
texture<int2,1> tzy;
texture<int2,1> tzz;
texture<int2,1> tmm;
texture<int> tixptr;
texture<int> tjxptr;

/*------------------------------------------------------------------------*/

// ALL OF THESE NOT NEEDED FOR DIRECT_EVAL_CUDA
#define MAXBLOCKS 16384
#define MAXTHREADSEVALUATION 64 //max threads for fmm direct evaluation and multipole evaluation
#define DIRECTMAXTHREADS 256 //max threads for tol=0, i.e. direct summation
#define REORDERMAXTHREADS 256 //max threads for CUDA reordering
#define M2PSMAXTHREADS 32 //number of threads in each block for M2PS interaction
#define P2PMAXTHREADS 64 //number of threads in each block for P2P interaction (needs to be atleast P2PMINTHREADS)
#define P2PMINTHREADS 32 //number of threads in each block for P2P interaction (lowest value)
#define SHIFT_M2M_MAXTHREADS 64 //number of threads in each block for P2P interaction (needs to be atleast SHIFT_M2M_MINTHREADS)
#define SHIFT_M2M_MINTHREADS 32 //number of threads in each block for P2P interaction (lowest value)

//cuda kernels has to be declared in same file as they are called
#include "cudaevalkernels.h"

/*------------------------------------------------------------------------*/
//prototype
void directsumkernellaunch(const double *m,double *GPUXr,
                           int N,int NE,int pot,enum SMOOTHER smooth,int kernelSelector);
/*------------------------------------------------------------------------*/
/*direct evaluation (tol==0). This function is independent of the others and
  performs its own memory allocation, cuda calls cleanup etc. */
void direct_eval_cuda(int N,
		      const double *zx,const double *zy,const double *zz,
		      const double *mm,int NE,
		      const double *ex,const double *ey, const double *ez,
		      double *pr,
		      double *qr,
		      int pot,enum SMOOTHER smooth)

//function to start the GPU direct summation calculations
{
  /* **** kernelSelector ****
     1. DIRAC, harmonic potential eval at zx,zy,zz
     2. DIRAC, harmonic potential eval at ex,ey,ez
   */
  int kernelSelector; // choosing eval kernel type
  cudavariables ctmp;
  cudavariables *GPUvars=&ctmp; //simply to always use GPUvars as a pointer
  if(N==0) /*no potential points, code will give Unspecified Driver Error otherwise*/
    return;
  cudastart();
  checkcudaerror("Check eval_cuda start");
  /* Timing not implemented, needs cudatiming and some extra functions
  TimeCreateAndStart(GPUvars->fulltimestart,GPUvars->fulltimestop);*/
  
  //start by making necessary allocations and copy memory to GPU
  // evaluation att (ex,ey,ez)
  if(qr != NULL) { 
    cudasafeMalloc((void**)&GPUvars->ex,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->ey,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->ez,NE*sizeof(double));
    cudasafeMalloc((void**)&GPUvars->qr,NE*sizeof(double));
    cudasafe( cudaMemcpy(GPUvars->ex, ex, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy ex" );
    cudasafe( cudaMemcpy(GPUvars->ey, ey, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy ey" );
    cudasafe( cudaMemcpy(GPUvars->ez, ez, NE*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy ez" );
  }
  else {
    GPUvars->ex=NULL;
    GPUvars->ey=NULL;
    GPUvars->ez=NULL;
    GPUvars->qr=NULL;
  }
  // evaluation at zx,zy,zz
  if(pr!=NULL) {
    cudasafeMalloc((void**)&GPUvars->pr,N*sizeof(double));
  }
  cudasafeMalloc((void**)&GPUvars->zx,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->zy,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->zz,N*sizeof(double));
  cudasafeMalloc((void**)&GPUvars->mm,N*sizeof(double));
  
  // Smoother not used?
  //  MPexp_smooth_(pot, smooth, xopt, cont, &cutoff, &shape, &scale);

  cudasafe( cudaMemcpy(GPUvars->zx, zx, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zx" );
  cudasafe( cudaMemcpy(GPUvars->zy, zy, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zy" );
  cudasafe( cudaMemcpy(GPUvars->zz, zz, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy zz" );
  cudasafe( cudaMemcpy(GPUvars->mm, mm, N*sizeof(double), cudaMemcpyHostToDevice), "cudaMemcpy mm" );

  cudasafe(cudaBindTexture(NULL,tzx,GPUvars->zx,N*sizeof(double)),"cudaBindTexture tzx");
  cudasafe(cudaBindTexture(NULL,tzy,GPUvars->zy,N*sizeof(double)),"cudaBindTexture tzy");
  cudasafe(cudaBindTexture(NULL,tzz,GPUvars->zz,N*sizeof(double)),"cudaBindTexture tzz");
  cudasafe(cudaBindTexture(NULL,tmm,GPUvars->mm,N*sizeof(double)),"cudaBindTexture tmm");

    //move textures to make it possible to use same code in both cases.
  // ***TODO What does this do???
  if(pr!=NULL) {
    cudasafe(cudaBindTexture(NULL,tex,GPUvars->ex,N*sizeof(double)),"cudaBindTexture tex");
    cudasafe(cudaBindTexture(NULL,tey,GPUvars->ey,N*sizeof(double)),"cudaBindTexture tey");
    cudasafe(cudaBindTexture(NULL,tez,GPUvars->ez,N*sizeof(double)),"cudaBindTexture tez");
    
    //launch GPU kernel - pr != null, eval at zx,zy,zz
    kernelSelector = 1;
    directsumkernellaunch(mm,GPUvars->pr,N,N,pot,smooth,kernelSelector);
  }
  if(qr != NULL) {
    
    cudasafe(cudaBindTexture(NULL,tex,GPUvars->ex,NE*sizeof(double)),"cudaBindTexture tex");
    cudasafe(cudaBindTexture(NULL,tey,GPUvars->ey,NE*sizeof(double)),"cudaBindTexture tey");
    cudasafe(cudaBindTexture(NULL,tez,GPUvars->ez,NE*sizeof(double)),"cudaBindTexture tez");
    
    //launch GPU kernel - qr != null, eval at ez,ey,ez
    kernelSelector = 2;
    directsumkernellaunch(mm,GPUvars->qr,N,NE,pot,smooth,kernelSelector);
  }

  //start the cleanup
  //***TODO: FIX CLEANUP FOR qr,pr ???
  // Complains about command for qr cleanup

  if(qr != NULL) {
    cudasafe( cudaMemcpy(qr,GPUvars->qr, NE*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy qr" );
    cudaFreeDebug(GPUvars->ex);
    cudaFreeDebug(GPUvars->ey);
    cudaFreeDebug(GPUvars->ez);
    cudaFreeDebug(GPUvars->qr);
  }

  if(pr!=NULL) {
    cudasafe( cudaMemcpy(pr,GPUvars->pr, N*sizeof(double), cudaMemcpyDeviceToHost), "cudaMemcpy pr" );
    cudaFreeDebug(GPUvars->pr);
  }
  cudaUnbindTexture(tzx);
  cudaUnbindTexture(tzy);
  cudaUnbindTexture(tzz);
  cudaUnbindTexture(tmm);
  cudaUnbindTexture(tex);
  cudaUnbindTexture(tey);
  cudaUnbindTexture(tez);
    
  cudaFreeDebug(GPUvars->zx);
  cudaFreeDebug(GPUvars->zy);
  cudaFreeDebug(GPUvars->zz);
  cudaFreeDebug(GPUvars->mm);

}
/*------------------------------------------------------------------------*/
void cudaSafeMemcpy(void *dst,const void *src,size_t count,
                    enum cudaMemcpyKind kind,const char *str)
// wrapper
{
  cudasafe(cudaMemcpy(dst,src,count,kind),str);
}
/*------------------------------------------------------------------------*/
void cudaSafeMemset(void *dst,int value,size_t count,const char *str)
// wrapper
{
  cudasafe(cudaMemset(dst,value,count),str);
}
/*------------------------------------------------------------------------*/
cudaError_t cudaGetLastErrorDummy()
// wrapper
{
  return cudaGetLastError();
}
/*------------------------------------------------------------------------*/
void getGPUinfo(cudavariables* GPUvars)
{
  int device;
  cudasafe(cudaGetDevice(&device),"cudaGetDevice");
  cudasafe(cudaGetDeviceProperties(&GPUvars->info,device),"cudaGetDeviceProperties");
}

/*------------------------------------------------------------------------*/

//selects correct direct summation kernel and starts it
void directsumkernellaunch(const double *m,double *GPUXr,
                           int N,int NE,int pot,SMOOTHER smooth,int kernelSelector)
{
  cudaDeviceProp info;
  int device;
  cudasafe(cudaGetDevice(&device),"cudaGetDevice");
  cudasafe(cudaGetDeviceProperties(&info,device),"cudaGetDeviceProperties");
  if(info.multiProcessorCount<1) //safety as some devices report 0 multiprocessors
      info.multiProcessorCount=1;
  if(info.warpSize<1) //safety
      info.warpSize=32;
  //load balancing. Each multiprocessor can run up to 8 blocks (current models, this information is sadly not available in the info structure)
  //  
  // ***TODO: AROUND HERE ARE OPTIMIZATIONS TO BE MADE FOR NEWER GPU ARCHITECTURES.
  
  int nthreads=imin(DIRECTMAXTHREADS,ceil((double)NE/(double)info.multiProcessorCount/(double)CUDABLOCKSPERMP/(double)info.warpSize)*info.warpSize);
  switch(kernelSelector) {
  case 1:
    // ***TODO -> fix better function names!
    kernelType1<<<imin(MAXBLOCKS, (NE+nthreads-1)/nthreads), nthreads,nthreads*4*sizeof(double)>>>(GPUXr,N); // eval at zx,zy,zz
  case 2:
    kernelType2<<<imin(MAXBLOCKS, (NE+nthreads-1)/nthreads), nthreads,nthreads*4*sizeof(double)>>>(GPUXr,N,NE); // eval at ex,ey,ez
  }
    
}
/*
  checkcudaerror("Kernel launch error direct interact");
  CHECKCUDAERROR
*/
  /*------------------------------------------------------------------------*/


/*
 * Below are the cuda function bodys for mpexp direct interaction. 
 * They should first transfer data to GPU sort of like the direct interact,
 * then the lists should be sorted, can 3D adapted FMM2dgpu reorder_cuda1-4 functions.
 *
 * Rough call hierarchy:
 * Transfer memory -> sort lists (cuda kernels) -> call evaluation kernel ->
 * transfer memory back -> cleanup
 */

void mpexp_direct_interact_cuda2(double *pr,
			   const double *zx,const double *zy,const double *zz,
			   const double *mm,
			   const int *ix,
			   int begin1,int end1,int begin2,int end2,
			   int pot,enum SMOOTHER smooth)
{
  //Function goes here
}

void mpexp_direct_interact_cuda(double *qr,
			  const double *ex,const double *ey,const double *ez,
			  const int *jx,int begin1,int end1,
			  const double *zx,const double *zy,const double *zz,
			  const double *mm,
			  const int *ix,int begin2,int end2,
			  int pot,enum SMOOTHER smooth){
  //Function goes here
}

/*------------------------------------------------------------------------*/

//#endif /* CUDASUPPORT */