/* direct.c */

/* S. Engblom 2010-02-18
   J. Westberg 2016-05-02 - Added mpexp_directinteracts from fmm.h
   J. Bull 2019-05-02 - added direct interactions between MPI processes */

#include "direct.h"
#ifdef OMP
  #include <omp.h>
#endif
double ddistance(const double dx,const double dy,const double dz);
void rinv2(int N,
	   const double *restrict zx,const double *restrict zy,const double *restrict zz,
	   const double *restrict mm,
	   double *pr,
	   SMOOTHER smooth);
void rinv(int N,
	  const double *restrict zx,const double *restrict zy,const double *restrict zz,
	  const double *restrict mm,
	  int NE,
	  const double *restrict ex,const double *restrict ey,const double *restrict ez,
	  double *qr,
	  SMOOTHER smooth);

/*------------------------------------------------------------------------*/

void directInteract(int N,
        const double *zx,const double *zy,const double *zz,
		    const double *mm,
		    int NE,
        const double *ex,const double *ey,const double *ez,
        double *pr,double *qr,
        int pot,SMOOTHER smooth)
/* Direct evaluation of the potential from N pointmasses (mm)
   at the positions (zx,zy,zz). The evaluation is done
   at the points (zx,zy,zz) if (p) is not NULL and/or at the points
   (ex,ey,ez) if (q) is not NULL. The result is stored in the vectors
   (p) and (q), respectively, which must be allocated and
   cleared prior to call. Hard-coded to use harmonic potential. */
{
  // evaluation at (zx,zy,zz)
  if (pr != NULL) {
    rinv2(N,zx,zy,zz,mm,pr,smooth);
  }

  // evaluation at (ex,ey,ez)
  if (qr != NULL) {
    rinv(N,zx,zy,zz,mm,NE,ex,ey,ez,qr,smooth);
  }
}
/*------------------------------------------------------------------------*/
// distance between 2 points (duplicate of fmm.c function)
inline double ddistance(const double dx,const double dy,const double dz)
{
  return sqrt(dx*dx + dy*dy + dz*dz);
}
/*------------------------------------------------------------------------*/
/* Now follows the 4 corresponding functions for the Harmonic
   potential. */
void rinv2(int N,
	   const double *restrict zx,const double *restrict zy,const double *restrict zz,
	   const double *restrict mm,
	   double *pr,
	   SMOOTHER smooth)
{

#ifdef FMM_MPI
  // dense N*N calculation. Each process operates on a row in the global matrix
#endif

  // TODO: use openmp for this double loop
  for (int i = 0; i < N; i++) {
    for (int j = i+1; j < N; j++) {
      const double r = ddistance(zx[i]-zx[j],zy[i]-zy[j],zz[i]-zz[j]);
      const double P = 1.0/r;
      pr[i] += mm[j]*P;
      pr[j] += mm[i]*P;
    }
  }
}
/*------------------------------------------------------------------------*/
void rinv(int N,
	  const double *restrict zx,const double *restrict zy,const double *restrict zz,
	  const double *restrict mm,
	  int NE,
	  const double *restrict ex,const double *restrict ey,const double *restrict ez,
	  double *qr,
	  SMOOTHER smooth)
{

  for (int j = 0; j < N; j++) {
    for (int i = 0; i < NE; i++) {
      const double r = ddistance(zx[j]-ex[i],zy[j]-ey[i],zz[j]-ez[i]);
      const double P = 1.0/r;
      qr[i] += mm[j]*P;
    }
  }
}
/*------------------------------------------------------------------------*/

/* Moving function bodies from fmm.c  to here allows for parallel definition
 * for cpu / gpu functions using the same direct.h header
 */
#define BEGIN(begin,i) (begin == -1 ? i+1 : begin)

void mpexp_directInteract2(double * restrict pr,
			   const double *restrict zx,
			   const double * restrict zy,
			   const double * restrict zz,
			   const double * restrict mm,
			   const int * restrict ix,
			   const int begin1,const int end1,
			   const int begin2,const int end2,
			   int pot,SMOOTHER smooth)
/* Direct interaction between potentials inside a box (begin2 == -1)
   or between two boxes. */
{
#ifdef OMP
  omp_set_num_threads(1);
#pragma omp parallel for
#endif
  for (int i = begin1; i < end1; i++) {
    for (int j = BEGIN(begin2,i); j < end2; j++) {
      const double r = ddistance(zx[ix[i]]-zx[ix[j]], zy[ix[i]]-zy[ix[j]], zz[ix[i]]-zz[ix[j]]);
      pr[ix[i]] += mm[ix[j]]/r;
      pr[ix[j]] += mm[ix[i]]/r;
    }
  }
}

void mpexp_directInteract(double *qr,
			  const double *restrict ex,const double *restrict ey,const double *restrict ez,
			  const int *restrict jx,int begin1,int end1,
			  const double *restrict zx,const double *restrict zy,const double *restrict zz,
			  const double *restrict mm,
			  const int *restrict ix,int begin2,int end2,
			  int pot,SMOOTHER smooth)
/* Direct evaluation of field from potentials inside one box computed
   at points (ex,ey,ez) inside another box. */
{
  for (int i = begin1; i < end1; i++) {
    for (int j = begin2; j < end2; j++) {
      qr[jx[i]] += mm[ix[j]]/ddistance(fabs(ex[jx[i]]-zx[ix[j]]),fabs(ey[jx[i]]-zy[ix[j]]),fabs(ez[jx[i]]-zz[ix[j]]));
    }
  }
}

#ifdef FMM_MPI
void mpexp_directInteractHalo(double *pr,
			  const double *restrict zx,const double *restrict zy,const double *restrict zz,
			  const int *restrict ix,int begin1,int end1,
			  const double *restrict recvlptr,
			  int begin2,int end2,
			  int pot,SMOOTHER smooth,int rank)
/* Direct evaluation of field from potentials inside non-owned box computed
   at potentials (zx,zy,zz) inside this box. */
{
  for (int i = begin1; i < end1; i++) {
    for (int j = begin2; j < end2; j++) {
      pr[ix[i]] += recvlptr[4*j+3]/ddistance(zx[ix[i]]-recvlptr[4*j], zy[ix[i]]-recvlptr[4*j+1], zz[ix[i]]-recvlptr[4*j+2]);
    }
  }
}
#endif
