/* fmm.c */

/* S. Engblom 2010-02-18
   J. Bull 2015-11-18 */

#include "fmm.h"
#include "fmmshift.h"
#include "direct.h"
#define CACHE
/* #defines: (preferably handled via make)
   ---------------------------------------
   FMM_INLINECOMPLEX
                 Avoid complex arithmetic in various places in the
                 FMM-algorithm and use inline versions instead. Care
                 has been taken to do this in such a way that the
                 result does not differ from C99's native complex
                 aritmetic. When in doubt, try #undefining
                 FMM_INLINECOMPLEX.

   FMM_INLINECOMPLEX_DIRECT
                 Not implemented in 3D code.
                 Inline complex arithmetic in direct evaluations. On
                 some platforms, it is faster by a factor of up to 8
                 not to rely on certain math-functions and/or complex
                 arithmetics, but to write the complex arithmetic
                 inline. On the one hand this implies possible
                 overflow as the evaluations of complex division and
                 absolute values demands properly scaling; on the
                 other hand this will only be observed with very
                 extreme scaling of the input since the operations are
                 only used in near-field computations. Again, when in
                 doubt, try #undefining FMM_INLINECOMPLEX_DIRECT.

   FMM_SHRINK    Postsplit shrinking of boxes on/off: this option is
                 kept here for the purpose of visual
                 demonstrations. Also good to have this option if
                 time-dependent FMM is implemented.

   It is recommended to #define all these.   
*/

/* Criterion for near/far approximation: two boxes interact through
   cluster-cluster approximations provided that the function
   mpexp_theta() below returns true. Generally, small theta (say, less
   than 0.5) means that less coefficients are needed but that the
   interaction lists get larger, and vice versa. */
double theta = 0.5; // 0 < theta < 1

// high-level routines: public
void MPexp_setup(struct MPexp *this1,int N,int NE,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
                 const double *ex,const double *ey,const double *ez,
                 int pot,double tol,int pcoeff,int *Ndirect,int *Nlevels,
                 double eta,bool meshonly);
void MPexp_free(struct MPexp *this1);

// evaluate when eval and source the same
void MPexp_eval2(const struct MPexp *this1,
                 double *pr,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
//                 std::vector< std::vector<double> > recvl,
                 int pot,SMOOTHER smooth);
// evaluate when eval and source different
void MPexp_eval(const struct MPexp *this1,
                double *qr,
                const double *ex,const double *ey,const double *ez,
                const double *zx,const double *zy,const double *zz,
                const double *mm,
                int pot,SMOOTHER smooth);
#ifdef FMM_MPI
void MPexp_eval2send(const struct MPexp *this1,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
                 std::vector< std::vector<double> > *sendl,
                 std::vector< std::vector<int> > *sendlist,
                 MPI_Request *& requests);
int MPexp_procno(const struct MPexp *this1, int p);
#endif
// helper functions: don't use outside fmm.c
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
       double *zmin,double *zmax,
                   int N,const double *x,const double *y,const double *z);
void MPexp_box_(struct MPexp *this1,int N,int NE,
                const double *zx,const double *zy,const double *zz,
                const double *ex,const double *ey,const double *ez);
void MPexp_partition_(int begin,int *im,int end,
          double *z0,double *min,double *max,
                      int *ix,const double *z,int Nmax,double eta);
void MPexp_split_(int begin,int *im,int end,
      double z0,double *min,double *max,
                  int *ix,const double *z);
void MPexp_aspectratio(double x, double y, double z,
     double dx, double dy, double dz,
     const double *zx,const double *zy,const double *zz,
     const double **px0,
     const double *ex,const double *ey,const double *ez,
     const double **qx0,
     int *max);
void MPexp_cutbox(int *ix,const double *px,int begin1,int *im,int end1,
     int *jx,const double *qx,int begin2,int *jm,int end2,
     double eta,double Nmax0,double NE,
     int dim0,int dim1,int dim2,
     double *xyz,double *dxyz,
     double *xyz0,double *dxyz0,
     double *xyz1,double *dxyz1);
void MPexp_sort_(struct MPexp *this1,
                 int N,const double *zx,const double *zy,const double *zz,
                 int NE,const double *ex,const double *ey,const double *ez,
                 int *maxm2l,int *levm2l,int Ndirect,double eta);
double MPexp_distance(const double dx,const double dy,const double dz);

// *** DEBUG:
#ifdef FMM_MPI
#ifndef NOMEX
void MPexp_halo2sparse(struct MPexp *this1, int level, int Nf,
     const struct mpHalo *H, struct mpSparse *HS);
#endif
#endif

// low-level routines (box level)
void mpexp_p2m(struct mpexp *this1,dcmplx *Ynm,int pcoeff,
               const double zx,const double zy,const double zz,
               const double mm,
               int pot);
void mpexp_l2p(const struct mpexp *this1,dcmplx *Ynm,int pcoeff,
               double *vr,
               const double *zx,const double *zy,const double *zz,
               const int *ix,int begin,int end);
bool mpexp_theta(const struct mpexp *this1,const struct mpexp *that);

/*------------------------------------------------------------------------*/
void fmm3dInteract(int rank,int nprocs,int N,
                   const double *zx,const double *zy,const double *zz,
                   const double *mm,
                   int NE,
                   const double *ex,const double *ey,const double *ez,
                   double *pr,
                   double *qr,
                   int pot,SMOOTHER smooth,
                   double tol,int pcoeff,double theta_,
                   int *Ndirect,int *Nlevels,double eta)
/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (m) at the
   positions (zx,zy,zz). The evaluation is done at the points (zx,zy,zz) if
   (p) is not NULL and/or at the points (ex,ey,ez) if (qx,qy,qz) is not
   NULL (it is required that NE is zero if (q) are NULL). The
   result is added to the vectors (p) and (q), respectively,
   which must be allocated prior to call. */
{
  // early return: cannot allocate anything
  if (N <= 0) return;

  // theta-criterion: theta is passed as a global variable
  theta = theta_;

  // multipole structure
  struct MPexp MP;
  
#ifdef FMM_MPI
  MP.rank = rank;
  MP.nprocs = nprocs;
  double starttime = MPI_Wtime();
#endif

  // allocate and setup representation
  MPexp_setup(&MP,N,NE,zx,zy,zz,mm,ex,ey,ez,pot,tol,pcoeff,Ndirect,Nlevels,
              eta,false);

  // timing for tree setup and traversal
#ifdef FMM_MPI
  ofstream fp;
  if (!rank)
  {
    fp.open("perf.log", ofstream::app);
  
    if(!fp.is_open()) {
      cout << "perf.log is unavailable" << endl;
    }
    fp << MP.nprocs << " 2 " << N << " " << tol << " " << *Nlevels << " " << MPI_Wtime() - starttime << endl;
    fp.close();
  }

  starttime = MPI_Wtime();
#endif

  StartTime;
  // evaluation at potentials (zx,zy,zz)
  if (pr != NULL) {
    MPexp_eval2(&MP,pr,zx,zy,zz,mm,pot,smooth);
  }
  // evaluation at coordinates (ex,ey,ez)
  if (qr != NULL) {
    MPexp_eval(&MP,qr,ex,ey,ez,zx,zy,zz,mm,pot,smooth);
  }
  StopTime;
  PrintTime("rank %d, P2P: %f\n");
  
  // deallocate
  MPexp_free(&MP);
  
  // timing for evaluation
#ifdef FMM_MPI
  if (!rank)
  {
    fp.open("perf.log", ofstream::app);
  
    if(!fp.is_open()) {
      cout << "perf.log is unavailable" << endl;
    }
    fp << MP.nprocs << " 3 " << N << " " << tol << " " << *Nlevels << " " << MPI_Wtime() - starttime << endl;
    fp.close();
  }
#endif
}
/*------------------------------------------------------------------------*/
#ifdef FMM_MPI
#ifndef NOMEX
void fmm3dGetMesh(mxArray *m[],mxArray *c[],mxArray *h[],mxArray *s[],mxArray *t[],
                  int rank,int nprocs,int N,
                  const double *zx,const double *zy,const double *zz,
                  const double *mm,
                  int NE,
                  const double *ex,const double *ey,const double *ez,
                  int pot,SMOOTHER smooth,
                  double tol,int pcoeff,double theta_,
                  int *Ndirect,int *Nlevels,double eta)
/* The MPI-version of fmm3dGetMesh(). */
{
  if (N <= 0) return;
  theta = theta_;

  // optional output
  int write_c = 1, write_h = (nprocs > 1), write_s = 0, write_t = 0;
  
  // setup multipole tree
  struct MPexp MP;

  MP.rank = rank;
  MP.nprocs = nprocs;

  MPexp_setup(&MP,N,NE,zx,zy,zz,mm,ex,ey,ez,pot,tol,pcoeff,Ndirect,Nlevels,
              eta,true);
  
  // create cell vector and copy the FMM mesh
  m[0] = mxCreateCellMatrix(1,MP.nlevel+1);
  for (int i = 0; i <= MP.nlevel; i++) {
    const int N = MP.lptr[i+1]-MP.lptr[i];
    mxArray *cell = mxCreateDoubleMatrix(6,N,mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *msh = mxGetDoubles(cell);
#else
    double *msh = mxGetPr(cell);
#endif
    struct mpexp *offset = &MP.root[MP.lptr[i]];

    for (int j = 0; j < N; j++) {
      // midpoint of mpexp offset
      msh[6*j] = offset[j].x0;
      msh[6*j+1] = offset[j].y0;
      msh[6*j+2] = offset[j].z0;
      // radius w.r.t. cell midpoint
      msh[6*j+3] = offset[j].dx0;
      msh[6*j+4] = offset[j].dy0;
      msh[6*j+5] = offset[j].dz0;
    }
    // set ith entry in LHS to cell values
    mxSetCell(m[0],i,cell);
  }

  // additionally, the connection matrices
  if(write_c) {
    c[0] = mxCreateCellMatrix(1,MP.nlevel+1);
    for (int i = 0; i <= MP.nlevel; i++) {
      const int N = MP.lptr[i+1]-MP.lptr[i];

      const mpSparse *S = &MP.connect[i];
      mxArray *cell = mxCreateSparse(N,N,S->jcptr[N],mxREAL);
      mwIndex *jc = mxGetJc(cell),*ir = mxGetIr(cell);
#if MX_HAS_INTERLEAVED_COMPLEX
      double *pr = mxGetDoubles(cell);
#else
      double *pr = mxGetPr(cell);
#endif
      // mwIndex not defined? use '-DmwIndex=int' in the mex-command!

      jc[0] = 0;
      for (int j = 0,k; j < N; j++) {
        jc[j+1] = S->jcptr[j+1];
        
        // near connections
        for (k = jc[j]; k < S->kcptr[j]; k++) {
          ir[k] = S->ir[k];
          pr[k] = 1;
        }
        // far connections
        for ( ; k < S->jcptr[j+1]; k++) {
          ir[k] = S->ir[k];
          pr[k] = -1;
        }
      }
      mxSetCell(c[0],i,cell);
    }
  }

  // the halo matrices
  if(write_h) {
    h[0] = mxCreateCellMatrix(1,MP.nlevel+1);
    
    // (nprocs-1) block of N*N sparse arrays
    struct mpSparse *HS = (struct mpSparse*)mxMalloc(sizeof(struct mpSparse));
    
    for (int i = 0; i <= MP.nlevel; i++) {
      int N = MP.lptr[i+1]-MP.lptr[i];
      mxArray *level = mxCreateCellMatrix(1,MP.nprocs-1);
      const struct mpHalo *H = &MP.halo[i];

      // convert mpHalo on level i to mpSparse format
      MPexp_halo2sparse(&MP, i, N, H, HS);
      
      for(int p=0,q=0; p<MP.nprocs-1; p++) {
        // cyclic index
        int index = MP.rank+p - (MP.nprocs-1)*(MP.rank+p>MP.nprocs-2);
        mxArray *cell = mxCreateSparse(N,N,(HS->jcptr[q+N]),mxREAL);
        mwIndex *jc = mxGetJc(cell),*ir = mxGetIr(cell);
#if MX_HAS_INTERLEAVED_COMPLEX
        double *pr = mxGetDoubles(cell);
#else
        double *pr = mxGetPr(cell);
#endif
        
        jc[0] = HS->jcptr[q];
        for (int j = 0,k; j < N; j++, q++) {
          jc[j+1] = HS->jcptr[q+1];

          // near connections
          for (k = HS->jcptr[q]; k < HS->kcptr[q]; k++) {
            ir[k] = HS->ir[k];
            pr[k] = 1;
          }

          // far connections
          for ( ; k < HS->jcptr[q+1]; k++) {
            ir[k] = HS->ir[k];
            pr[k] = -1;
          }
        }
        // put proc data into cell array level
        mxSetCell(level,index,cell);
      }
      // put level i into cell array h
      mxSetCell(h[0],i,level);
    }
    // mxFree not thread-safe?
    // free up memory
    //mxFree(HS->jcptr);
    //mxFree(HS);
  }
  
  // additionally, a sparse matrix where each column is a list of all
  // potentials in each box at the finest level
  const int Nf = 1 << (MP.nlevel*3);
  if(write_s) {
    s[0] = mxCreateSparse(N,Nf,N,mxREAL);
    mwIndex *jc = mxGetJc(s[0]),*ir = mxGetIr(s[0]);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *p = mxGetDoubles(s[0]);
#else
    double *p = mxGetPr(s[0]);
#endif
    
    jc[0] = 0;
    for (int i = 0; i < Nf; i++) {
      jc[i+1] = MP.ixptr[i+1];
      for (int j = jc[i]; j < jc[i+1]; j++) {
        ir[j] = MP.ix[j];
        p[j] = 1.0;
      }
    }
  }
  
  // additionally, a sparse matrix where each column is a list of
  // all evaluation points in each box at the finest level
  if(write_t) {
    t[0] = mxCreateSparse(NE,Nf,NE,mxREAL);
    mwIndex *jc = mxGetJc(t[0]),*ir = mxGetIr(t[0]);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *p = mxGetDoubles(t[0]);
#else
    double *p = mxGetPr(t[0]);
#endif
    
    jc[0] = 0;
    for (int i = 0; i < Nf; i++) {
      jc[i+1] = MP.jxptr[i+1];
      for (int j = jc[i]; j < jc[i+1]; j++) {
        ir[j] = MP.jx[j];
        p[j] = 1.0;
      }
    }
  }

  // deallocate and return
  MPexp_free(&MP);
}
/*------------------------------------------------------------------------*/
#endif // NOMEX
#else
void fmm3dGetMesh(int nlhs,mxArray *plhs[],
                  int N,
                  const double *zx,const double *zy,const double *zz,
                  const double *mm,
                  int NE,
                  const double *ex,const double *ey,const double *ez,
                  int pot,SMOOTHER smooth,
                  double tol,int pcoeff,double theta_,
                  int *Ndirect,int *Nlevels,double eta)
/* Returns the multipole mesh in a cell vector of complex
   mxArrays. Additionally, all sparse connection matrices are
   optionally returned. For debugging and pedagogical purposes. */
{
  if (N <= 0) return;
  theta = theta_;

  // setup multipole tree
  struct MPexp MP;
  MPexp_setup(&MP,N,NE,zx,zy,zz,mm,ex,ey,ez,pot,tol,pcoeff,Ndirect,Nlevels,
              eta,true);

  // create cell vector and copy the FMM mesh
  plhs[0] = mxCreateCellMatrix(1,MP.nlevel+1);
  for (int i = 0; i <= MP.nlevel; i++) {
    mxArray *cell = mxCreateDoubleMatrix(6,MP.lptr[i+1]-MP.lptr[i],mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *msh = mxGetDoubles(cell);
#else
    double *msh = mxGetPr(cell);
#endif
    struct mpexp *offset = &MP.root[MP.lptr[i]];

    for (int j = 0; j < MP.lptr[i+1]-MP.lptr[i]; j++) {
      // midpoint of mpexp offset
      msh[6*j] = offset[j].x0;
      msh[6*j+1] = offset[j].y0;
      msh[6*j+2] = offset[j].z0;
      // radius w.r.t. cell midpoint
      msh[6*j+3] = offset[j].dx0;
      msh[6*j+4] = offset[j].dy0;
      msh[6*j+5] = offset[j].dz0;
    }
    // set ith entry in LHS to cell values
    mxSetCell(plhs[0],i,cell);
  }

  // additionally, the connection matrices
  if (nlhs > 1) {
    plhs[1] = mxCreateCellMatrix(1,MP.nlevel+1);
    for (int i = 0; i <= MP.nlevel; i++) {
      const struct mpSparse *S = &MP.connect[i];
      const int N = MP.lptr[i+1]-MP.lptr[i];
      mxArray *cell = mxCreateSparse(N,N,S->jcptr[N],mxREAL);
      mwIndex *jc = mxGetJc(cell),*ir = mxGetIr(cell);
#if MX_HAS_INTERLEAVED_COMPLEX
      double *pr = mxGetDoubles(cell);
#else
      double *pr = mxGetPr(cell);
#endif
      // mwIndex not defined? use '-DmwIndex=int' in the mex-command!

      jc[0] = 0;
      for (int j = 0,k; j < N; j++) {
        jc[j+1] = S->jcptr[j+1];
        
        // near connections
        for (k = jc[j]; k < S->kcptr[j]; k++) {
          ir[k] = S->ir[k];
          pr[k] = 1.0;
        }
        // far connections
        for ( ; k < S->jcptr[j+1]; k++) {
          ir[k] = S->ir[k];
          pr[k] = -1.0;
        }
      }
      mxSetCell(plhs[1],i,cell);
    }
  }

  /* additionally, a sparse matrix where each column is a list of all
   potentials in each box at the finest level */
  const int Nf = 1 << (MP.nlevel*3);
  if (nlhs > 2) {
    plhs[2] = mxCreateSparse(N,Nf,N,mxREAL);
    mwIndex *jc = mxGetJc(plhs[2]),*ir = mxGetIr(plhs[2]);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *p = mxGetDoubles(plhs[2]);
#else
    double *p = mxGetPr(plhs[2]);
#endif
    
    jc[0] = 0;
    for (int i = 0; i < Nf; i++) {
      jc[i+1] = MP.ixptr[i+1];
      for (int j = jc[i]; j < jc[i+1]; j++) {
        ir[j] = MP.ix[j];
        p[j] = 1.0;
      }
    }
  }
  
  /* additionally, a sparse matrix where each column is a list of
   all evaluation points in each box at the finest level */
  if (nlhs > 3) {
    plhs[3] = mxCreateSparse(NE,Nf,NE,mxREAL);
    mwIndex *jc = mxGetJc(plhs[3]),*ir = mxGetIr(plhs[3]);
#if MX_HAS_INTERLEAVED_COMPLEX
    double *p = mxGetDoubles(plhs[3]);
#else
    double *p = mxGetPr(plhs[3]);
#endif
    
    jc[0] = 0;
    for (int i = 0; i < Nf; i++) {
      jc[i+1] = MP.jxptr[i+1];
      for (int j = jc[i]; j < jc[i+1]; j++) {
        ir[j] = MP.jx[j];
        p[j] = 1.0;
      }
    }
  }

  // deallocate and return
  MPexp_free(&MP);
}
#endif
/*------------------------------------------------------------------------*/
void MPexp_setup(struct MPexp *this1,int N,int NE,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
                 const double *ex,const double *ey,const double *ez,
                 int pot,double tol,int pcoeff,int *Ndirect,int *Nlevels,
                 double eta,bool meshonly)
/* Setups and allocates a multipole tree for N potentials at (zx,zy,zz)
   to be evaluated at the NE points (ex,ey,ez) to tolerance tol. The
   discretization is determined so that the number of potentials in
   each box at the finest level is <= Ndirect. If the boolean variable
   meshonly is true, then only the multipole tree and the connectivity
   information is determined.
   
   Argument pcoeff and Nlevels, respectively, overrides tol and
   Ndirect. Use the value -1 to signal a void value.
*/
{

  // local value used in PrintTime
  int rank = 0;
#ifdef FMM_MPI
  rank = this1->rank;
#endif
  
  StartTime;
  // size of the domain
  MPexp_box_(this1,N,NE,zx,zy,zz,ex,ey,ez);
  
  // number of multipole terms
  if (pcoeff != -1)
    /* (note: overrides tol) */
    this1->pcoeff = pcoeff-1; // -1 since this1->pcoeff understands 0 offset
  else
    this1->pcoeff = (int)floor(log2(4.0*tol)/log2(theta)-1.0)+(pot != 0);
  /* the magical constants 4 is a (not overly optimistic) estimate
     obtained from numerous experiments with the model error ~
     constant * theta^(p+1) and theta = 0.5 */
  if (this1->pcoeff < 1) this1->pcoeff = 1;

  // number of levels
  if (*Nlevels != -1) {
    /* (note: overrides Ndirect) */
    this1->nlevel = *Nlevels;
    const int Nf = 1 << (this1->nlevel*3);
    *Ndirect = (double)N/Nf;
    if (*Ndirect <= 0) *Ndirect = 1;
  }
  else
    this1->nlevel = ceil((1.0/3.0)*log2(5.0/8.0*(double)N/(*Ndirect)));
  /* the magical constant 5/8 comes from solving for the number of
     levels the equation E(N/Nf) = Ndirect, where E is expectation, N
     is considered a uniform random number on [1,M] (with M tending to
     infinity) and Nf is defined below */
  if (this1->nlevel < 0) this1->nlevel = 0;
  *Nlevels = this1->nlevel;
  
  /* used frequently below: total number of coefficients, number of
     boxes at the finest level and total number of boxes */
  const int P = this1->pcoeff+1;
  const int psq = P*P;
  const int Nf = 1 << (this1->nlevel*3);
  const int Nt = (Nf-1)/7+Nf;

#if defined FMM_MPI
  if(this1->rank==0) {
#endif
    printf("tol %f, pot %d, N %d, NE %d, Nlevels %d, Ndirect %d, pcoeff %d\n",tol,pot,N,NE,*Nlevels,*Ndirect,this1->pcoeff);
#if defined FMM_MPI
  }
#endif

  // pointer to levels
  this1->lptr = (int*) malloc((this1->nlevel+2)*sizeof(int));
  this1->lptr[0] = 0;
  for (int level = 0,nb = 1; level <= this1->nlevel; level++, nb <<= 3) {
    this1->lptr[level+1] = this1->lptr[level]+nb;
  }
  
  // multipole tree
  this1->root = (struct mpexp*) malloc(Nt*sizeof(struct mpexp));

  // multipole coefficients
  dcmplx *coeffs1 = (dcmplx*) calloc(psq*Nt,sizeof(dcmplx));
  dcmplx *coeffs2 = (dcmplx*) calloc(psq*Nt,sizeof(dcmplx));
  for (int j = 0; j < Nt; j++) {
    this1->root[j].coeff1 = coeffs1; coeffs1 += psq;
    this1->root[j].coeff2 = coeffs2; coeffs2 += psq;
  }

  // permutation of input
  this1->ixptr = (int*) malloc((Nf+1)*sizeof(int));
  this1->ix = (int*) malloc(N*sizeof(int));
  if (NE) {
    this1->jxptr = (int*) malloc((Nf+1)*sizeof(int));
    this1->jx = (int*) malloc(NE*sizeof(int));
  }
  else
    this1->jxptr = this1->jx = NULL;

  // connections between boxes at each level
  this1->connect = (struct mpSparse*) malloc((this1->nlevel+1)*sizeof(struct mpSparse));

#ifdef FMM_MPI

  // halo defining connections between boxes on different processes
  this1->halo = (struct mpHalo*) malloc((this1->nlevel+1)*sizeof(struct mpHalo));
  
#endif

  StopTime;
  PrintTime("rank %d, Alloc: %f\n");
  
  StartTime;
  
  /* setup the multipole mesh, sort all potentials and evaluation
     points and determine all connections.
     TODO: send halo and connected box coords to other procs only once,
     then save time during M2L and P2P. */
  int maxm2l,levm2l;
  MPexp_sort_(this1,N,zx,zy,zz,NE,ex,ey,ez,&maxm2l,&levm2l,*Ndirect,eta);
  StopTime;
  PrintTime("rank %d, Tree: %f\n");
  
  if (meshonly) return;
  
  /* Start of the actual multipole algorithm */

#ifdef FMM_MPI

  // set up + alloc quantities for parallel M2L
  int comm_size = this1->nprocs-1;
  MPI_Status status;
  MPI_Request *requests;
  requests = (MPI_Request*) malloc(2*comm_size*this1->nlevel*sizeof(MPI_Request));

  // buffers for send/receive coords and coeffs with MPI
  std::vector< std::vector<double> > sendl(comm_size*this1->nlevel);
  std::vector< std::vector<double> > recvl(comm_size*this1->nlevel);

#endif

  /* Step #1: Initialization at the finest level. */
  
  // prepare to shift
  shift_alloc(P,pot,maxm2l);
  
  // allocate harmonic coeffs for use in mpexp_p2m
  dcmplx *Ynm_wksp = (dcmplx *) malloc(psq*sizeof(dcmplx));

  struct mpexp *offset = &this1->root[this1->lptr[this1->nlevel]];
  
  // initialize coeff1 (multipole expansion) in boxes at finest (leaf) level
  StartTime;
  for (int j = 0; j < Nf; j++) {
    for (int i = this1->ixptr[j]; i < this1->ixptr[j+1]; i++) {
      mpexp_p2m(&offset[j],Ynm_wksp,this1->pcoeff,
                zx[this1->ix[i]],zy[this1->ix[i]],zz[this1->ix[i]],
                mm[this1->ix[i]],pot);
    }
  }
  StopTime;
  PrintTime("rank %d, P2M: %f\n");
  free(Ynm_wksp);

  /* Step #2: Upward pass of the fast multipole algorithm. After the
     finest level has been initialized, this loop performs the
     'upward' pass where all multipole contributions are shifted and
     added onto all coarser levels. */
  StartTime;

  for (int l = this1->nlevel,nb = Nf; l > 0; l--, nb >>= 3) {
    struct mpexp *parent = &this1->root[this1->lptr[l-1]];
    struct mpexp *child = &this1->root[this1->lptr[l]];

    // loop over all children at this level
    for (int j = 0; j < nb; j += 8) {
    
      // M2M shift from 8 children to 1 parent at once
      shift_m2mZ(parent,&child[j]);

      parent++;
    }
  }
  StopTime;
  PrintTime("rank %d, M2M: %f\n");
  
  /* Step #3: Downward pass of the fast multipole method. This shifts
     expansions from the interaction list of each box to the center of
     the box and adds up the result. Before continuing to finer
     levels, each parent shifts and adds its expansion to each of its
     children. */

  /* Substep 3H: swap data with connected procs.
  Then compute M2L shifts from distant boxes to owned boxes. */

#ifdef FMM_MPI
  StartTime;

  int levm2lh = 1;
  int bufdim = 3+2*psq;
  
  // received mpexp
  struct mpexp *child_j = (struct mpexp*) malloc(sizeof(struct mpexp));
  child_j->coeff1 = (dcmplx*) calloc(psq,sizeof(dcmplx));
  
  // for halos we need an extra level in the downward passes:
  // do M2L and L2L shifts at the whole-subdomain level (level 0)
  // TODO: for large number of procs we need multiple extra levels (supertree)
  for (int l = levm2lh-1,nb = 1 << 3*levm2lh; l < this1->nlevel; l++, nb <<= 3) {
    struct mpexp *child = &this1->root[this1->lptr[l+1]];
    struct mpHalo *H = &this1->halo[l+1];
    
    // unsorted lists
    int *iboxes = (int *) malloc(nb*sizeof(int));
    int *jboxes = (int *) malloc(nb*sizeof(int));
    // map from send list index to ibox
    int *sendmap = (int *) malloc(nb*sizeof(int));
    // map from jbox to recv list index
    int *recvmap = (int *) malloc(comm_size*nb*sizeof(int));
    
    // loop over procs and send coords, coeffs
    for(int p=0,n=0; p<comm_size; p++) {
      // the other proc
      int i = MPexp_procno(this1, p);
      
      // increment index n ready for weak connections
      n += H->sptr[p];
      
      // init temp arrays
      for(int j=0; j<nb; j++) {
        iboxes[j] = -1;
        jboxes[j] = -1;
      }
      
      // loop over weak connections in halo (index n)
      for(int j=H->sptr[p]; j<H->wptr[p]; j++, n++) {
        int ii = H->ibox[n];
        int jj = H->jbox[n];
        // store iboxes, jboxes in ascending order with gaps
        iboxes[ii] = ii;
        jboxes[jj] = jj;
      }
      
      // counters for no. of weakly connected iboxes, jboxes
      int icount=0,jcount=0;
      // remove gaps and construct send/recv maps
      for(int j=0; j<nb; j++) {
        int ii = iboxes[j];
        int jj = jboxes[j];
        if(ii != -1) {
          sendmap[icount] = ii;
          icount++;
        }
        if(jj != -1) {
          recvmap[p*nb+jj] = jcount;
          jcount++;
        }
      }
      // allocate send/recv buffers for coordinates and coefficients
      sendl[l*comm_size+p].reserve(bufdim*icount);
      recvl[l*comm_size+p].resize(bufdim*jcount);
      
      // fill send buffer with coords
      for(int j=0; j<icount; j++) {
        int ii = sendmap[j];
        sendl[l*comm_size+p].push_back(child[ii].x0);
        sendl[l*comm_size+p].push_back(child[ii].y0);
        sendl[l*comm_size+p].push_back(child[ii].z0);
        
        // fill send buffer with multipole coeffs
        for(int k=0; k<psq; k++) {
          sendl[l*comm_size+p].push_back(creal(child[ii].coeff1[k]));
          sendl[l*comm_size+p].push_back(cimag(child[ii].coeff1[k]));
        }
      }
      
      // send/recv box centre and multipole coefficients
      if(sendl[l*comm_size+p].size() > 0) {
        MPI_Isend(sendl[l*comm_size+p].data(),sendl[l*comm_size+p].size(),MPI_DOUBLE,i,2,MPI_COMM_WORLD,&requests[l*comm_size+p]);
      }
      if(recvl[l*comm_size+p].size() > 0) {
        MPI_Irecv(recvl[l*comm_size+p].data(),recvl[l*comm_size+p].size(),MPI_DOUBLE,i,2,MPI_COMM_WORLD,&requests[(this1->nlevel+l)*comm_size+p]);
      }
    }
    
    // MPI flag
    int sum = 0;
    std::vector<int> beenthere(comm_size);
    for(int p=0; p<comm_size; p++) beenthere[p] = 0;
    
    // loop over procs until we find a message that has arrived
    while(sum<comm_size) {
      
      for(int p=0,n=0; p<comm_size; p++) {
        
        // the other proc
        int i = MPexp_procno(this1, p);
        
        // increment index to start of weak connections
        n += H->sptr[p];
        
        if(recvl[l*comm_size+p].size() > 0) {
          
          int complete = 0;
          // test if coeffs have arrived. Once they have, requests is dereferenced.
          MPI_Test(&requests[(this1->nlevel+l)*comm_size+p],&complete,&status);
          
          // if so, and we haven't been here before, proceed with asymmetric M2L shift.
          if(complete && !beenthere[p]) {
            
            // loop over weak connections (index n)
            for (int j = H->sptr[p]; j < H->wptr[p]; j++, n++) {
              int ii = H->ibox[n];
              int jj = H->jbox[n];
              
              // get index of jbox in recv buffers
              int m = recvmap[p*nb+jj];
              
              // construct struct mpexp from received data
              child_j->x0 = recvl[l*comm_size+p][bufdim*m];
              child_j->y0 = recvl[l*comm_size+p][bufdim*m+1];
              child_j->z0 = recvl[l*comm_size+p][bufdim*m+2];
              
              // reconstruct complex coeffs
              for(int k=0; k<psq; k++)
              child_j->coeff1[k] = recvl[l*comm_size+p][bufdim*m+3+2*k] + I*recvl[l*comm_size+p][bufdim*m+3+2*k+1];
              
              // perform M2L shift from jboxes to iboxes
              double sph2[3];
              cart2spher(child[ii].x0-child_j->x0, child[ii].y0-child_j->y0,
                         child[ii].z0-child_j->z0,sph2);
              
              shift_m2lZsimple(this1->rank,sph2,NULL,child[ii].coeff2,child_j->coeff1,NULL);
            }
            sum++;
            beenthere[p] = 1;
          }
          else {
            // increment from start of weak connections to start of next proc's strong connections
            n += H->wptr[p]-H->sptr[p];
          }
        }
        // if there are unconnected procs, count them once so that sum is equal to comm_size.
        else {
          if(!beenthere[p]) {
            // increment from start of weak connections to start of next proc's strong connections
            n += H->wptr[p]-H->sptr[p];
            sum++;
            beenthere[p] = 1;
          }
        }
      }
    }
    free(iboxes);
    free(jboxes);
    free(sendmap);
    free(recvmap);
  }
  
  free(child_j->coeff1);
  free(child_j);
  
  // Shift local coeffs from top level to levm2l-1
  for (int l = levm2lh-1,nb = 1 << 3*levm2lh; l < levm2l-1; l++, nb <<= 3) {
    struct mpexp *parent = &this1->root[this1->lptr[l]];
    struct mpexp *child = &this1->root[this1->lptr[l+1]];
    for (int j = 0; j < nb; j += 8) {
      
      // Shift from 1 parent to 8 children at once
      shift_l2lZ(&child[j],parent);
      
      parent++;
    }
  }
  
  // time M2Lh+M2L together (comment out these 2 lines and next StartTime)
//  StopTime;
//  PrintTime("rank %d, M2Lh: %f\n");
#endif

  // Continue the downward pass of the fast multipole method on current proc.
//  StartTime;
  
  for (int l = levm2l-1,nb = 1 << 3*levm2l; l < this1->nlevel; l++, nb <<= 3) {
    struct mpexp *child = &this1->root[this1->lptr[l+1]];
    struct mpSparse *C = &this1->connect[l+1];

    /* tricky part: loop over all boxes (parents) at this level and
       symmetrically shift results to and from the interaction list of
       each child */
    for (int j = 0; j < nb; j += 8) {

      // loop through the interaction list of each child
      for (int k = 0; k < 8; k++) {

        // M2L to child from all connections at once
        shift_m2lZ(&child[j+k],&child[0],C->ir,C->kcptr[j+k],C->jcptr[j+k+1],1);
      }
    }
  }
  StopTime;
  PrintTime("rank %d, M2L: %f\n");

  // Tests show that it's marginally faster to do L2L in a separate loop to M2L.
  StartTime;
  
  for (int l = levm2l-1,nb = 1 << 3*levm2l; l < this1->nlevel; l++, nb <<= 3) {
    struct mpexp *parent = &this1->root[this1->lptr[l]];
    struct mpexp *child = &this1->root[this1->lptr[l+1]];

    for (int j = 0; j < nb; j += 8) {

      // Shift from 1 parent to 8 children at once
      shift_l2lZ(&child[j],parent);
      parent++;

    }
  }
  
  StopTime;
  PrintTime("rank %d, L2L: %f\n");

#ifdef FMM_MPI

  // clean up
  // deallocate send/recv buffers on all levels
  for (int l = levm2lh-1,nb = 1 << 3*levm2lh; l < this1->nlevel; l++) {
    for (int p=0; p<comm_size; p++) {
      if(sendl[l*comm_size+p].size() > 0) {
        // Wait for Isend of coeffs and box coords to complete.
        MPI_Wait(&requests[l*comm_size+p],&status);
        sendl[l*comm_size+p].clear();
      }
      if(recvl[l*comm_size+p].size() > 0) {
        recvl[l*comm_size+p].clear();
      }
    }
  }

  free(requests);

#endif
}
/*------------------------------------------------------------------------*/
void MPexp_free(struct MPexp *this1)
/* Deallocates the whole tree. */
{
  for (int l = this1->nlevel; l >= 0; l--)
    free(this1->connect[l].jcptr);
  free(this1->connect);
  free(this1->jx);
  free(this1->jxptr);
  free(this1->ix);
  free(this1->ixptr);
  free(this1->root->coeff2);
  free(this1->root->coeff1);
  free(this1->root);
  free(this1->lptr);
#ifdef FMM_MPI
  if(this1->nprocs > 1) {
    MPI_Barrier(MPI_COMM_WORLD);
  }
  for (int l = this1->nlevel; l >= 0; l--) {
    free(this1->halo[l].ibox);
  }
  free(this1->halo);
#endif
}
/*------------------------------------------------------------------------*/
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
       double *zmin,double *zmax,
                   int N,const double *x,const double *y,const double *z)
/* Finds min/max of two vectors. Note: it is assumed that suitable
   initial values in xmin/xmax are already set. */
{
  // loop unrolled to get rid of a few N/2 comparisions
  for (int i = 0; i < N-1; i += 2) {
    double min,max;

    if (x[i] < x[i+1]) {
      min = x[i];
      max = x[i+1];
    }
    else {
      min = x[i+1];
      max = x[i];
    }
    if (min < *xmin) *xmin = min;
    if (max > *xmax) *xmax = max;

    if (y[i] < y[i+1]) {
      min = y[i];
      max = y[i+1];
    }
    else {
      min = y[i+1];
      max = y[i];
    }
    if (min < *ymin) *ymin = min;
    if (max > *ymax) *ymax = max;

    if (z[i] < z[i+1]) {
      min = z[i];
      max = z[i+1];
    }
    else {
      min = z[i+1];
      max = z[i];
    }
    if (min < *zmin) *zmin = min;
    if (max > *zmax) *zmax = max;
  }
  if (N&1) {
    if (x[N-1] < *xmin)
      *xmin = x[N-1];
    else if (x[N-1] > *xmax)
      *xmax = x[N-1];

    if (y[N-1] < *ymin)
      *ymin = y[N-1];
    else if (y[N-1] > *ymax)
      *ymax = y[N-1];

    if (z[N-1] < *zmin)
      *zmin = z[N-1];
    else if (z[N-1] > *zmax)
      *zmax = z[N-1];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_box_(struct MPexp *this1,int N,int NE,
                const double *zx,const double *zy,const double *zz,
                const double *ex,const double *ey,const double *ez)
/* Determines and sets the dimensions of the computational domain. */
{
  double eps=0.0;
  this1->xmin = zx[0]-eps;
  this1->xmax = zx[0]+eps;
  this1->ymin = zy[0]-eps;
  this1->ymax = zy[0]+eps;
  this1->zmin = zz[0]-eps;
  this1->zmax = zz[0]+eps;
  MPexp_minmax_(&this1->xmin,&this1->xmax,
                &this1->ymin,&this1->ymax,
    &this1->zmin,&this1->zmax,
                N-1,zx+1,zy+1,zz+1);

  // bounding box might become larger:
  if (NE) 
    MPexp_minmax_(&this1->xmin,&this1->xmax,
                  &this1->ymin,&this1->ymax,
      &this1->zmin,&this1->zmax,
                  NE,ex,ey,ez);
}
/*------------------------------------------------------------------------*/
#define SWAPI(X,Y) { int tmpi = X; X = Y; Y = tmpi; }
void MPexp_partition_(int begin,int *im,int end,
                      double *z0,double *min,double *max,
                      int *ix,const double *z,int Nmax,double eta)
/* Partitions the set of points z[ix[begin]] through z[ix[end]]
   (exclusive) in two sets indicated by [begin,im) and [im,end) such
   that all points are <= or >= than z0 (which is both input and
   output). Additionally, whenever min and max are nonzero, the
   minimum and maximum number are also determined. The division is
   done such that Nlarge <= Nmax, where Nlarge is the larger number of
   particles in the two lists. It is assumed that this split is
   possible and hence that Nmax is not too small.

   The routine is a workhorse and performance is important; for each
   box which is not a leaf it will be called seven times. */
{
  int il,ir,left = begin,right = end;
  int Nsmall,Nlarge;

  // first split according to input z0 (geometric midpoint)
  const double z0_ = *z0; // compiler: this is really a constant!
  il = left-1; ir = right;
  do il++; while (il < right && z[ix[il]] < z0_);
  do ir--; while (left <= ir && z0_ < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    /* after the first swap, there is no need to check for index out
       of bounds */
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0_);
      do ir--; while (z0_ < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }

  // proceed with largest interval [left,il) or [il,right)
  Nsmall = il-begin;
  Nlarge = end-il;
  if (Nsmall > Nlarge) {
    Nlarge = Nsmall;
    *im = right = il;
  }
  else
    *im = left = il;
  const int im0 = *im; // saved for later

  while (Nlarge > Nmax) {
    if (right-left >= 3) {
      /* "median of 3": find pivot and create boundaries such that
         left <= left+1 <= right-1 */
      const int mid = (left+right)>>1;
      SWAPI(ix[left+1],ix[mid])
      if (z[ix[left]] > z[ix[right-1]])
        SWAPI(ix[left],ix[right-1])
      if (z[ix[left+1]] > z[ix[right-1]])
        SWAPI(ix[left+1],ix[right-1])
      else if (z[ix[left]] > z[ix[left+1]])
        SWAPI(ix[left],ix[left+1])
      const double z1_ = *z0 = z[ix[left+1]];

      for (il = left+1,ir = right-1; ; ) {
        do il++; while (z[ix[il]] < z1_);
        do ir--; while (z1_ < z[ix[ir]]);
        if (il >= ir) break;

        SWAPI(ix[il],ix[ir])
      }
      SWAPI(ix[left+1],ix[ir]) // ir is the correct place for the pivot

      // the pivot belongs to the shortest interval:
      Nsmall = ir-begin; // [begin,ir)
      Nlarge = end-ir-1; // [ir+1,end)
      if (Nsmall > Nlarge) {
        Nlarge = Nsmall;
        *im = right = ir;
      }
      else
        /* note: if Nsmall == Nlarge (can only happen if right-left is
           odd), then the right array will be shorter than the left by
           one */
        *im = left = ir+1;
    }
    else {
      // code above does not work properly for less than three elements
      if (z[ix[left]] > z[ix[right-1]])
        SWAPI(ix[left],ix[right-1])
      *im = right-1;
      *z0 = z[ix[right-1]];
      break;
      /*** prefer a direct sort when right-left <= 7 or so? */ 
    }
  }

  // determine min and max
  if (min != NULL) {
    *min = *z0;
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    *max = *z0;
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }

  // re-adjust to get a better connectivity balance (recall that the
  // theta criterion is a geometric criterion)
  if (eta < 1.0) {
    const double z1_ = eta**z0+(1-eta)*z0_; // new 'balanced' midpoint

    // determine boundaries using the result from the first
    // (geometric) split
    if (*z0 <= z0_) {
      il = (left = *im)-1;
      ir = right = im0;
    }
    else {
      il = (left = im0)-1;
      ir = right = *im;
    }

    // as in MPexp_split_() below
    do il++; while (il < right && z[ix[il]] < z1_);
    do ir--; while (left <= ir && z1_ < z[ix[ir]]);
    if (il < ir) {
    SWAPI(ix[il],ix[ir])
      for ( ; ; ) {
        do il++; while (z[ix[il]] < z1_);
        do ir--; while (z1_ < z[ix[ir]]);
        if (il >= ir) break;
        SWAPI(ix[il],ix[ir])
      }
    }
    *im = il;
    *z0 = z1_;
  }
}
/*------------------------------------------------------------------------*/
void MPexp_split_(int begin,int *im,int end,
      double z0,double *min,double *max,
                  int *ix,const double *z)
/* Split as above, but the coordinate z0 around which to split is
   already known. It is also assumed that min/max, whenever nonzero,
   are set to suitable initial guesses of the true min/max (typically
   the ones you get from MPexp_partition_() above). */
{
  int il = begin-1,ir = end;
  do il++; while (il < end && z[ix[il]] < z0);
  do ir--; while (begin <= ir && z0 < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0);
      do ir--; while (z0 < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }
  *im = il;

  if (min != NULL) {
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_aspectratio(double x, double y, double z,
     double dx, double dy, double dz,
     const double *zx,const double *zy,const double *zz,
     const double **px0,
     const double *ex,const double *ey,const double *ez,
     const double **qx0,
     int *dims)
/* takes in xyz0,dxyz0 of a box (in x,y,z order)
 and returns indices of array sorted in descending size order.
 also returns pointers (px,qx) to the component
 of coords (zx,zy,zz), (ex,ey,ez) that matches largest side. */
{
  double min = (dz < dy ? dz : dy);
  min = (dx < min ? dx : min);
  double max = (dz > dy ? dz : dy);
  max = (dx > max ? dx : max);
  
  // cube
  if(max/min == 1.0) {
    dims[0] = 0;
    dims[1] = 1;
    dims[2] = 2;
    *px0 = zx;
    *qx0 = ex;
  }
  // cuboid
  else {
    if(dx == max) {
      dims[0] = 0;
      *px0 = zx;
      *qx0 = ex;
      if(dy == min) {
        dims[1] = 2;
        dims[2] = 1;
      }
      else {
        dims[1] = 1;
        dims[2] = 2;
      }
    }
    else if(dy == max) {
      dims[0] = 1;
      *px0 = zy;
      *qx0 = ey;
      if(dx == min) {
        dims[1] = 2;
        dims[2] = 0;
      }
      else {
        dims[1] = 0;
        dims[2] = 2;
      }
    }
    else {
      dims[0] = 2;
      *px0 = zz;
      *qx0 = ez;
      if(dx == min) {
        dims[1] = 1;
        dims[2] = 0;
      }
      else {
        dims[1] = 0;
        dims[2] = 1;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void MPexp_cutbox(int *ix,const double *px,int begin1,int *im,int end1,
                  int *jx,const double *qx,int begin2,int *jm,int end2,
                  double eta,double Nmax0,double NE,
                  int dim0,int dim1,int dim2,
                  double *xyz,double *dxyz,
                  double *xyz0,double *dxyz0,
                  double *xyz1,double *dxyz1)
/* takes in xyz,dxyz of parent in x,y,z order
 and returns xyz0,dxyz0,xyz1,dxyz1 of 2 boxes in x,y,z order */
{
  double x = xyz[dim0]; // longest - this will be split
  double y = xyz[dim1]; // middle
  double z = xyz[dim2]; // shortest
  double dx = dxyz[dim0];
  double dy = dxyz[dim1];
  double dz = dxyz[dim2];
  double x0,y0,z0,x1,y1,z1;
  double dx0,dy0,dz0,dx1,dy1,dz1;
  double split = x;
  double min,max;
  
  MPexp_partition_(begin1,im,end1,&split,&min,&max,ix,px,Nmax0,eta);
  if (NE)
    MPexp_split_(begin2,jm,end2,split,&min,&max,jx,qx);
  
#ifdef FMM_SHRINK
  x0 = 0.5*(min+split);
  y0 = y;
  z0 = z;
  dx0 = x0-min;
  dy0 = y0-(y-dy);
  dz0 = z0-(z-dz);
  x1 = 0.5*(split+max);
  y1 = y;
  z1 = z;
  dx1 = max-x1;
  dy1 = (y+dy)-y1;
  dz1 = (z+dz)-z1;
#else
  x0 = 0.5*(x-dx+split);
  y0 = y;
  z0 = z;
  dx0 = x0-(x-dx);
  dy0 = y0-(y-dy);
  dz0 = z0-(z-dz);
  x1 = 0.5*(split+x+dx);
  y1 = y;
  z1 = z;
  dx1 = (x+dx)-x1;
  dy1 = (y+dy)-y1;
  dz1 = (z+dz)-z1;
#endif

  // setup coordinate arrays of resulting boxes
  xyz0[dim0] = x0;
  xyz0[dim1] = y0;
  xyz0[dim2] = z0;
  xyz1[dim0] = x1;
  xyz1[dim1] = y1;
  xyz1[dim2] = z1;
  dxyz0[dim0] = dx0;
  dxyz0[dim1] = dy0;
  dxyz0[dim2] = dz0;
  dxyz1[dim0] = dx1;
  dxyz1[dim1] = dy1;
  dxyz1[dim2] = dz1;
}
/*------------------------------------------------------------------------*/
void MPexp_sort_(struct MPexp *this1,
                 int N,const double *zx,const double *zy,const double *zz,
                 int NE,const double *ex,const double *ey,const double *ez,
                 int *maxm2l,int *levm2l,int Ndirect,double eta)
/* Sorts the N potentials (zx,zy,zz) in the multipole tree this which is
   simultaneously constructed and, for NE > 0, also assigns each
   evaluation point (ex,ey,ez) a multipole box. After each level has been
   thus constructed the connectivity information is also
   determined. The integer maxm2l is set to the maximum number of
   cluster to cluster interactions in the whole tree (the maximum
   number of -1's in any column of all the connectivity matrices) and
   levm2l is the first level where a cluster to cluster interaction
   occurs. This function is critical to performance. */
{
  const int Nf = this1->lptr[this1->nlevel+1]-this1->lptr[this1->nlevel];
  int *ix,*ixptr,*ixptr0,*ixptr1;
  int *jx,*jxptr,*jxptr0,*jxptr1;
  struct mpSparse *C = this1->connect;
  int rank = 0;
#ifdef FMM_MPI
  rank = this1->rank;
#endif
  
  // (ixptr,ix) to be built
  ix = this1->ix;
  for (int i = 0; i < N; i++) ix[i] = i;
  ixptr = this1->ixptr;
  ixptr0 = &ixptr[Nf-1];
  ixptr0[0] = 0;
  ixptr0[1] = N;
  
  // ditto (jxptr,jx)
  if (NE) {
    jx = this1->jx;
    for (int i = 0; i < NE; i++) jx[i] = i;
    jxptr = this1->jxptr;
    jxptr0 = &jxptr[Nf-1];
    jxptr0[0] = 0;
    jxptr0[1] = NE;
  }
  else
    jx = jxptr = NULL;
  
  // root box midpoint and size
  this1->root[0].x0  = 0.5*(this1->xmin+this1->xmax);
  this1->root[0].y0  = 0.5*(this1->ymin+this1->ymax);
  this1->root[0].z0  = 0.5*(this1->zmin+this1->zmax);
  this1->root[0].dx0 = 0.5*(this1->xmax-this1->xmin);
  this1->root[0].dy0 = 0.5*(this1->ymax-this1->ymin);
  this1->root[0].dz0 = 0.5*(this1->zmax-this1->zmin);
  
  /* connectivity: each matrix is strictly lower triangular since a
   box is always in near connection with itself and since
   connectivity is a symmetric relation */
  C[0].jcptr = (int*) malloc(5*sizeof(int));
  C[0].jcptr[0] = 0;
  C[0].jcptr[1] = 0;
  // convenient in fmm3dGetMesh() above:
  C[0].kcptr = &C[0].jcptr[2];
  C[0].kcptr[0] = 0;
  C[0].kcptr[1] = 0; // note: 0 could be the final level!
  C[0].ir = &C[0].kcptr[2];
  C[0].ir[0] = 0;
  
#ifdef FMM_MPI
//  StartTime;
  // Very sparse inter-proc global connectivity matrix:
  // the 'from' box indices of proc j are found in ibox[j],
  // the 'to' box indices for proc j are found in jbox[j].
  // Strongly connected boxes are first, then weakly
  // connected boxes. Finally, pointers to end of strong
  // and weak lists. Each process connected to this one
  // has a contiguous chunk of memory in H.*/
  mpHalo *H = this1->halo;
  // temporary local storage for non-owned box
  mpexp *boxj = (mpexp*) malloc(sizeof(mpexp));
  MPI_Status status;
  MPI_Request *requests;
  MPI_Request *requests2;
  int comm_size = this1->nprocs-1;
  
  // storage for boxes for MPI send/recv
  // on level 0
  vector<double> send0(6);
  std::vector< std::vector<double> > recv0(comm_size);
  // on all remaining levels
  std::vector< std::vector<double> > sendl(comm_size);
  std::vector< std::vector<double> > recvl(comm_size);
  requests = (MPI_Request*) malloc(2*comm_size*sizeof(MPI_Request));
  requests2 = (MPI_Request*) malloc(2*comm_size*sizeof(MPI_Request));

  H[0].halo_nnz = comm_size;
  H[0].ibox = (int*) malloc(4*H[0].halo_nnz*sizeof(int));
  H[0].jbox = &H[0].ibox[H[0].halo_nnz];
  H[0].sptr = &H[0].jbox[H[0].halo_nnz];
  H[0].wptr = &H[0].sptr[H[0].halo_nnz];
  
  // Get coords of all other proc bounding boxes
  send0[0] = this1->root[0].x0;
  send0[1] = this1->root[0].y0;
  send0[2] = this1->root[0].z0;
  send0[3] = this1->root[0].dx0;
  send0[4] = this1->root[0].dy0;
  send0[5] = this1->root[0].dz0;
  
  // Non-blocking sens and receive data.
  // round robin indexing.
  // p is index (0:nprocs-2) in halo array.
  for(int p=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    recv0[p].resize(6);
    MPI_Isend(send0.data(),send0.size(),MPI_DOUBLE,i,0,MPI_COMM_WORLD,&requests2[p]);
    MPI_Irecv(recv0[p].data(),recv0[p].size(),MPI_DOUBLE,i,0,MPI_COMM_WORLD,&requests2[p+comm_size]);
  }
  
  // Receive data.
  for(int p=0; p<comm_size; p++) {
    
    MPI_Wait(&requests2[p+comm_size],&status);
    
    boxj->x0 = recv0[p][0];
    boxj->y0 = recv0[p][1];
    boxj->z0 = recv0[p][2];
    boxj->dx0 = recv0[p][3];
    boxj->dy0 = recv0[p][4];
    boxj->dz0 = recv0[p][5];
    boxj->coeff1 = boxj->coeff2 = NULL;
    
    // only 1 box at level 0
    H[0].ibox[p] = 0;
    // local index of box on other proc
    H[0].jbox[p] = 0;
    // connection according to theta criterion.
    // all procs are defined as near or far connections.
    if (!mpexp_theta(&this1->root[0],boxj)) {
      // near connection
      H[0].sptr[p] = 1;
      H[0].wptr[p] = 1;
    }
    else {
      // far connection
      H[0].sptr[p] = 0;
      H[0].wptr[p] = 1;
    }
    
    boxj->x0 = NULL;
    boxj->y0 = NULL;
    boxj->z0 = NULL;
    boxj->dx0 = NULL;
    boxj->dy0 = NULL;
    boxj->dz0 = NULL;
  }
//  StopTime;
//  PrintTime("rank %d, halo setup: %f\n");
#endif

  // vars for box splitting
  double aspect;
  int dims[3];
  double x,y,z,dx,dy,dz;
  double xyz[3],xyz0[3],xyz1[3],xyz00[3],xyz01[3],xyz10[3],xyz11[3];
  double dxyz[3],dxyz0[3],dxyz1[3],dxyz00[3],dxyz01[3],dxyz10[3],dxyz11[3];
  double xyz000[3],xyz001[3],xyz010[3],xyz011[3];
  double xyz100[3],xyz101[3],xyz110[3],xyz111[3];
  double dxyz000[3],dxyz001[3],dxyz010[3],dxyz011[3];
  double dxyz100[3],dxyz101[3],dxyz110[3],dxyz111[3];
  
  // loop over levels
  *maxm2l = 0;
  *levm2l = this1->nlevel+1;
  for (int l = 1,Nmax = N,Nmax0,nc = 0; l <= this1->nlevel; l++) {
    struct mpexp *parent = &this1->root[this1->lptr[l-1]];
    struct mpexp *child = &this1->root[this1->lptr[l]];
    const int N0 =  this1->lptr[l]-this1->lptr[l-1];
    const int N1 = this1->lptr[l+1]-this1->lptr[l];
    
    /* Step #1: divide each box on level l into 8 children such that
     about 1/8th of the potentials are located in each child box. */
    
    /* expanding arrays; they are written at in the lower part of the
     _same_ memory area (parents use the "0-part" and children the
     "1-part") */
    ixptr1 = &ixptr[Nf-N1];
    if (NE) jxptr1 = &jxptr[Nf-N1];
    
    /* determine the maximum number of coordinates after each split
     (the map is x --> ceil(0.5*x)) */
    if ((Nmax0 = (Nmax >> 1)+(Nmax&1)) <= Ndirect)
      Nmax0 = Nmax = Ndirect;
    else if ((Nmax = (Nmax0 >> 1)+(Nmax0&1)) <= Ndirect)
      Nmax = Ndirect;
    /* the intent with using Ndirect as a cutoff is to save some work
     at the final level whenever the number of particles is
     already sufficiently low */
    
    // loop over all parent boxes
    for (int i = 0,j = 0; i < N0; i++,j += 8) {
      
      // box to be split
      x  = parent[i].x0;
      y  = parent[i].y0;
      z  = parent[i].z0;
      dx = parent[i].dx0;
      dy = parent[i].dy0;
      dz = parent[i].dz0;

      /* partition the coordinates [ixptr0[i],ixptr[i+1]) in box i
       into 8 sections indexed by [ixptr1[8*i+n],ixptr1[8*i+n+1])
       for n = 0..7 */
      ixptr1[8*i] = ixptr0[i];
      if (NE) jxptr1[8*i] = jxptr0[i];
      
      /* cut the longest side followed by the next-longest
       followed by the smallest;
       this is a flexible Lebesgue ordering.
       TODO: compare to Hilbert ordering. */
      
      xyz[0] = x;
      xyz[1] = y;
      xyz[2] = z;
      dxyz[0] = dx;
      dxyz[1] = dy;
      dxyz[2] = dz;

      // ptrs to coordinate arrays
      const double *px,*qx;
      // ptrs to eval point permutation arrays
      int *qxptr = NULL;
      int begin2,end2;
      
      if(NE) {
        qxptr = &jxptr1[8*i];
        begin2 = jxptr0[i];
        end2 = jxptr0[i+1];
      }
      
      // takes in xyz0,dxyz0 of a box (in x,y,z order)
      // and returns indices of the largest to smallest sides.
      // also returns pointers (px,qx) to the component
      // of coords (zx,zy,zz), (ex,ey,ez) that matches largest side.
      MPexp_aspectratio(x,y,z,dx,dy,dz,zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      // adaptive split: longest side first.
      
      // takes in xyz,dxyz of parent and returns xyz0,dxyz0,xyz1,dxyz1 of 2 boxes in x,y,z order.
      MPexp_cutbox(ix,px,ixptr0[i],&ixptr1[8*i+4],ixptr0[i+1],
                   jx,qx,begin2,&qxptr[4],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz[0],&dxyz[0],&xyz0[0],
                   &dxyz0[0],&xyz1[0],&dxyz1[0]);
      
      // split longest side of box 0
      MPexp_aspectratio(xyz0[0],xyz0[1],xyz0[2],dxyz0[0],dxyz0[1],dxyz0[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        end2 = jxptr1[8*i+4];
      }
      
      MPexp_cutbox(ix,px,ixptr0[i],&ixptr1[8*i+2],ixptr1[8*i+4],
                   jx,qx,begin2,&qxptr[2],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz0[0],&dxyz0[0],&xyz00[0],
                   &dxyz00[0],&xyz01[0],&dxyz01[0]);
      
      // split longest side of box 1
      MPexp_aspectratio(xyz1[0],xyz1[1],xyz1[2],dxyz1[0],dxyz1[1],dxyz1[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        begin2 = jxptr1[8*i+4];
        end2 = jxptr0[i+1];
      }
      
      MPexp_cutbox(ix,px,ixptr1[8*i+4],&ixptr1[8*i+6],ixptr0[i+1],
                   jx,qx,begin2,&qxptr[6],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz1[0],&dxyz1[0],&xyz10[0],
                   &dxyz10[0],&xyz11[0],&dxyz11[0]);
      
      // split longest side of box 00
      MPexp_aspectratio(xyz00[0],xyz00[1],xyz00[2],dxyz00[0],dxyz00[1],dxyz00[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        begin2 = jxptr0[i];
        end2 = jxptr1[8*i+2];
      }
      
      MPexp_cutbox(ix,px,ixptr0[i],&ixptr1[8*i+1],ixptr1[8*i+2],
                   jx,qx,begin2,&qxptr[1],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz00[0],&dxyz00[0],&xyz000[0],
                   &dxyz000[0],&xyz001[0],&dxyz001[0]);
      
      // split longest side of box 01
      MPexp_aspectratio(xyz01[0],xyz01[1],xyz01[2],dxyz01[0],dxyz01[1],dxyz01[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        begin2 = jxptr1[8*i+2];
        end2 = jxptr1[8*i+4];
      }
      
      MPexp_cutbox(ix,px,ixptr1[8*i+2],&ixptr1[8*i+3],ixptr1[8*i+4],
                   jx,qx,begin2,&qxptr[3],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2]
                   ,&xyz01[0],&dxyz01[0],&xyz010[0],
                   &dxyz010[0],&xyz011[0],&dxyz011[0]);
      
      // split longest side of box 10
      MPexp_aspectratio(xyz10[0],xyz10[1],xyz10[2],dxyz10[0],dxyz10[1],dxyz10[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        begin2 = jxptr1[8*i+4];
        end2 = jxptr1[8*i+6];
      }
      
      MPexp_cutbox(ix,px,ixptr1[8*i+4],&ixptr1[8*i+5],ixptr1[8*i+6],
                   jx,qx,begin2,&qxptr[5],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz10[0],&dxyz10[0],&xyz100[0],
                   &dxyz100[0],&xyz101[0],&dxyz101[0]);
      
      // split longest side of box 11
      MPexp_aspectratio(xyz11[0],xyz11[1],xyz11[2],dxyz11[0],dxyz11[1],dxyz11[2],
                        zx,zy,zz,&px,ex,ey,ez,&qx,&dims[0]);
      
      if(NE) {
        begin2 = jxptr1[8*i+6];
        end2 = jxptr0[i+1];
      }
      
      MPexp_cutbox(ix,px,ixptr1[8*i+6],&ixptr1[8*i+7],ixptr0[i+1],
                   jx,qx,begin2,&qxptr[7],end2,
                   eta,Nmax0,NE,dims[0],dims[1],dims[2],
                   &xyz11[0],&dxyz11[0],&xyz110[0],
                   &dxyz110[0],&xyz111[0],&dxyz111[0]);
      
      // produced boxes
      child[j].x0 = xyz000[0];
      child[j].y0 = xyz000[1];
      child[j].z0 = xyz000[2];
      child[j].dx0 = dxyz000[0];
      child[j].dy0 = dxyz000[1];
      child[j].dz0 = dxyz000[2];
      
      child[j+1].x0 = xyz001[0];
      child[j+1].y0 = xyz001[1];
      child[j+1].z0 = xyz001[2];
      child[j+1].dx0 = dxyz001[0];
      child[j+1].dy0 = dxyz001[1];
      child[j+1].dz0 = dxyz001[2];
      
      child[j+2].x0 = xyz010[0];
      child[j+2].y0 = xyz010[1];
      child[j+2].z0 = xyz010[2];
      child[j+2].dx0 = dxyz010[0];
      child[j+2].dy0 = dxyz010[1];
      child[j+2].dz0 = dxyz010[2];
      
      child[j+3].x0 = xyz011[0];
      child[j+3].y0 = xyz011[1];
      child[j+3].z0 = xyz011[2];
      child[j+3].dx0 = dxyz011[0];
      child[j+3].dy0 = dxyz011[1];
      child[j+3].dz0 = dxyz011[2];
      
      child[j+4].x0 = xyz100[0];
      child[j+4].y0 = xyz100[1];
      child[j+4].z0 = xyz100[2];
      child[j+4].dx0 = dxyz100[0];
      child[j+4].dy0 = dxyz100[1];
      child[j+4].dz0 = dxyz100[2];
      
      child[j+5].x0 = xyz101[0];
      child[j+5].y0 = xyz101[1];
      child[j+5].z0 = xyz101[2];
      child[j+5].dx0 = dxyz101[0];
      child[j+5].dy0 = dxyz101[1];
      child[j+5].dz0 = dxyz101[2];
      
      child[j+6].x0 = xyz110[0];
      child[j+6].y0 = xyz110[1];
      child[j+6].z0 = xyz110[2];
      child[j+6].dx0 = dxyz110[0];
      child[j+6].dy0 = dxyz110[1];
      child[j+6].dz0 = dxyz110[2];
      
      child[j+7].x0 = xyz111[0];
      child[j+7].y0 = xyz111[1];
      child[j+7].z0 = xyz111[2];
      child[j+7].dx0 = dxyz111[0];
      child[j+7].dy0 = dxyz111[1];
      child[j+7].dz0 = dxyz111[2];
    }
    
    // arrays grow to the left
    ixptr0 = ixptr1;
    if (NE) jxptr0 = jxptr1;
    
#ifdef FMM_MPI

    /* Step 3 (pre-send data): Process halo connections on level l-1
    and send child boxes. The latency is partly hidden
    by computing the connectivity matrix. */

    // map (hash table) from jbox to index of recvlist
    int *recvmap = (int *) malloc(comm_size*N1*sizeof(int));
    int iboxes[N0],jboxes[N0],slist[N0],rlist[N0];
    
    // 3.1 get list of boxes owned by this proc which are strongly connected to other procs.
    // separate list for each of the receiving procs.
    // get indices of children of strongly connected parents
    for(int p=0,m=0; p<comm_size; p++) {
      // the other proc
      int i = MPexp_procno(this1, p);
      
      for(int j=0; j<N0; j++) {
        iboxes[j] = -1;
        jboxes[j] = -1;
      }
      
      // store iboxes, jboxes in increasing order with gaps
      for(int j=0; j<H[l-1].sptr[p]; j++, m++) {
        int ii = H[l-1].ibox[m];
        int jj = H[l-1].jbox[m];
        iboxes[ii] = ii;
        jboxes[jj] = jj;
      }
      
      slist[0] = 0;
      // ptr to end of nonzero values in slist
      int *slistptr = &slist[0];
      rlist[0] = 0;
      // ptr to end of nonzero values in rlist
      int *rlistptr = &rlist[0];
      int icount=0,jcount=0;
      
      // make send + recv lists
      for(int j=0; j<N0; j++) {
        
        int ii = iboxes[j];
        int jj = jboxes[j];
        
        // if ibox has not been added to iboxes,
        // add it to slist and increment ptr
        if(ii != -1) {
          slistptr[0] = ii;
          slistptr++;
          icount++;
        }
        if(jj != -1) {
          rlistptr[0] = jj;
          rlistptr++;
          jcount++;
        }
      }
      
      // assign space for coords
      sendl[p].reserve(6*8*icount);
      recvl[p].resize(6*8*jcount);
      
      for(int k=0; k<jcount; k++) {
        for(int q=0; q<8; q++) {
          recvmap[p*N1+8*rlist[k]+q] = 8*k+q;
        }
      }
      
      // add coordinates of children to sendlist
      for(int k=0; k<icount; k++) {
        for(int q=0; q<8; q++) {
          int cindex = 8*slist[k]+q;
          sendl[p].push_back(child[cindex].x0);
          sendl[p].push_back(child[cindex].y0);
          sendl[p].push_back(child[cindex].z0);
          sendl[p].push_back(child[cindex].dx0);
          sendl[p].push_back(child[cindex].dy0);
          sendl[p].push_back(child[cindex].dz0);
        }
      }
      
      // 3.2 Send coords of children
      // TODO: send struct mpexp.
      if(sendl[p].size() > 0) {
        MPI_Isend(sendl[p].data(),sendl[p].size(),MPI_DOUBLE,i,1,MPI_COMM_WORLD,&requests[p]);
      }
      if(recvl[p].size() > 0) {
        MPI_Irecv(recvl[p].data(),recvl[p].size(),MPI_DOUBLE,i,1,MPI_COMM_WORLD,&requests[p+comm_size]);
      }
      
      // increment index to start of strong connections with next proc
      m += H[l-1].wptr[p]-H[l-1].sptr[p];
    }

#endif

    /* Step #2: once all boxes at a level have been determined on,
     their connectivity can be sorted out. */
    
    /* allocate connectivity matrix using that all near connections to
     higher index boxes at the previous level implies 8 connections
     for each of 8 children at this level (and 28 additional
     connections to higher indices {0->[1 2 3 4 5 6 7],
     1->[2 3 4 5 6 7],...,6->[7]}
     between the children of each parent) */
    C[l].jcptr = (int*) malloc((2*N1+1+28*N0+8*8*nc)*sizeof(int));
    C[l].jcptr[0] = 0;
    C[l].kcptr = &C[l].jcptr[N1+1];
    C[l].ir = &C[l].kcptr[N1];
    
    // setup connections
    nc = 0;  // counter for near connections
    for (int i = 0,j = 0; i < N0; i++) { // parent i
      for (int c = 0; c < 8; c++,j++) { // children j
        // forward- and backward counters in each column
        int fwd = C[l].jcptr[j];
        int bwd = fwd+7-c+8*(C[l-1].kcptr[i]-C[l-1].jcptr[i]);
        C[l].jcptr[j+1] = bwd;
        
        /* [@Lthetanoteplots] set up near connections to siblings > j
         without checking the theta-criterion; the connection could
         in principle be of the far type (except to the neighbour
         sibling), but this should occur only very rarely */
        for (int d = 1; d < 8-c; d++)
          C[l].ir[fwd++] = j+d;
        
        // parents (near) connections
        for (int ii = C[l-1].jcptr[i]; ii < C[l-1].kcptr[i]; ii++) {
          /* the children of the near connections of box i become
           either near connections or far connections of box j */
          const int bwd0 = bwd;
          const int K = C[l-1].ir[ii];
          
          for (int d = 0, k = K << 3; d < 8; d++,k++) {
            // connection according to theta criterion
            if (!mpexp_theta(&child[j],&child[k])) {
              // near connection
              C[l].ir[fwd++] = k;
              //printf("child %d near child %d\n",j,k);
            }
            else {
              // far connection
              C[l].ir[--bwd] = k;
              //printf("child %d far from child %d\n",j,k);
            }
          }          
        }
        C[l].kcptr[j] = bwd;
        /* this splits the column into near connections
         [jcptr[j],kcptr[j]) and far connections
         [kcptr[j],jcptr[j+1]) */
        nc += bwd-C[l].jcptr[j];
        if (C[l].jcptr[j+1]-bwd > *maxm2l)
          *maxm2l = C[l].jcptr[j+1]-bwd;
        if (l < *levm2l && *maxm2l > 0)
          *levm2l = l;
      }
    }
    
    /* Step #3: finish computing halos. */
#ifdef FMM_MPI
    
    // allocate space for halos on level l.
    // Only need space for 64x previous level strong connections.
    int index = 0;
    for(int i=0; i<comm_size; i++) {
      index += H[l-1].sptr[i];
    }
    // allocate halo members
    H[l].halo_nnz = 64*index;
    H[l].ibox = (int*) malloc((2*H[l].halo_nnz+2*comm_size)*sizeof(int));
    H[l].jbox = &H[l].ibox[H[l].halo_nnz];
    H[l].sptr = &H[l].jbox[H[l].halo_nnz];
    H[l].wptr = &H[l].sptr[comm_size];
    
    // 3.3 Compute connections between children of current proc and received proc(s).
    // accumulate wptr from procs already visited
    // TODO: use a while loop to test for first available message, as in Setup and Eval2, to make this routine scale better.
    int count = 0;
    for(int p=0,m=0; p<comm_size; p++) {
      
      int fwd = 0;
      int bwd = 64*H[l-1].sptr[p];
      
      if(recvl[p].size() > 0) {
        // wait for Irecv
        MPI_Wait(&requests[p+comm_size],&status);
        
        // this loop structure is most efficient if the no. of connections is small
        // and the number of boxes is large.
        // otherwise, it may be better to loop over sendlist and recvlist.
        for(int j=0; j<H[l-1].sptr[p]; j++, m++) {
          int ii = H[l-1].ibox[m];
          int jj = H[l-1].jbox[m];
          
          // double loop over children of ibox and jbox
          for(int k=0; k<8; k++) {
            int ibox = 8*ii+k;
            
            for(int q=0; q<8; q++) {
              int jbox = 8*jj+q;
              // look up index of jbox in map
              int ll = recvmap[p*N1+jbox];
              
              // retrieve coordinates of jbox
              boxj->x0 = recvl[p][6*ll];
              boxj->y0 = recvl[p][6*ll+1];
              boxj->z0 = recvl[p][6*ll+2];
              boxj->dx0 = recvl[p][6*ll+3];
              boxj->dy0 = recvl[p][6*ll+4];
              boxj->dz0 = recvl[p][6*ll+5];
              boxj->coeff1 = boxj->coeff2 = NULL;
              
              // check theta criterion
              // near connection
              if (!mpexp_theta(&child[ibox],boxj)) {
                // add boxes to strong section of next level
                H[l].ibox[fwd+count] = ibox;
                H[l].jbox[fwd+count] = jbox;
                fwd++;
              }
              // far connection
              else {
                // add boxes to weak section of next level
                bwd--;
                H[l].ibox[bwd+count] = ibox;
                H[l].jbox[bwd+count] = jbox;
              }
              boxj->x0 = NULL;
              boxj->y0 = NULL;
              boxj->z0 = NULL;
              boxj->dx0 = NULL;
              boxj->dy0 = NULL;
              boxj->dz0 = NULL;
            }
          }
        }
      }
      
      // increment index to start of strong connections with next proc
      m += H[l-1].wptr[p]-H[l-1].sptr[p];
      
      // set pointer to end of strong connections
      H[l].sptr[p] = fwd;
      
      // ptr to end of weak connections
      H[l].wptr[p] = fwd+64*H[l-1].sptr[p]-bwd;
      
      // accumulate wptr count
      count += H[l].wptr[p];
    }

    // clean up before next level
    for(int p=0; p<comm_size; p++) {
      // wait for Isend
      if(sendl[p].size() > 0) {
        MPI_Wait(&requests[p],&status);
      }
      sendl[p].clear();
      recvl[p].clear();
    }
    free(recvmap);

#endif
    
  } // end of loop over levels
  
#ifdef FMM_MPI
  // Wait for sends of root boxes to complete
  for(int p=0; p<comm_size; p++) {
    MPI_Wait(&requests2[p],&status);
  }
  // clean up
  free(boxj);
  free(requests);
  free(requests2);
#endif
}
/*------------------------------------------------------------------------*/
// distance between 2 points
double MPexp_distance(const double dx,const double dy,const double dz)
{
  return sqrt(dx*dx + dy*dy + dz*dz);
}
/*------------------------------------------------------------------------*/
#ifdef FMM_MPI
/*
 Convert halo from struct mpHalo to struct mpSparse.
 Keep as it may be useful for debugging.
*/
void MPexp_halo2sparse(struct MPexp *this1, int level, int Nf,
     const struct mpHalo *H, struct mpSparse *HS)
{
  int comm_size = this1->nprocs-1;
  // allocate mpSparse members in one contiguous block
  HS->jcptr = (int*) malloc((1+H->halo_nnz+(2*Nf)*comm_size)*sizeof(int));
  // jcptr, kcptr, ir have blocks of memory for each connected proc
  HS->jcptr[0] = 0;
  HS->kcptr = &HS->jcptr[Nf*comm_size+1];
  HS->ir = &HS->kcptr[Nf*comm_size];
  
  // init to 0
  for(int i=0; i<Nf*comm_size; i++) {
    HS->jcptr[i] = 0;
    HS->kcptr[i] = 0;
  }
  HS->jcptr[Nf*comm_size] = 0;

  for(int i=0, m=0, j; i<comm_size; i++) {
    int offset = Nf*i;

    // fill strong column ptr (no. of jbox per ibox)
    for(j=0; j<H->sptr[i]; j++, m++) {
      ++(HS->jcptr[offset + H->ibox[m]+1]);
      ++(HS->kcptr[offset + H->ibox[m]]);
    }
    
    // fill weak column ptr (no. of jbox per ibox)
    for( ; j<H->wptr[i]; j++, m++) {
      ++(HS->jcptr[offset + H->ibox[m]+1]);
    }
  }

  // accumulate jcptr and kcptr entries per proc
  int cumul=0;
  for(int i=1; i<Nf*comm_size; i++) {
    cumul += HS->jcptr[i];
    (HS->jcptr[i]) = cumul;
    (HS->kcptr[i]) += cumul;
  }
  // last entry in jcptr
  (HS->jcptr[Nf*comm_size]) += cumul;

  // counter in each column for indexing ir
  int scount[Nf];
  
  // fill row ptr (jbox indices)
  for(int i=0, m=0; i<comm_size; i++) {
    int offset = Nf*i;
    
    for(int j=0; j<Nf; j++) {
      scount[j] = 0;
    }
    
    // fill column ptr (no. of jboxes per ibox)
    for(int j=0; j<H->wptr[i]; j++, m++) {
      int ind = HS->jcptr[offset + H->ibox[m]] + scount[H->ibox[m]];
      HS->ir[ind] = H->jbox[m];
      // no. of times ibox has appeared per proc
      scount[H->ibox[m]]++;
    }
  }
  
  // print result
  if(this1->rank==0 && 0) {
    for(int i=0,k=0; i<comm_size; i++) {
      for(int j=0; j<Nf; j++,k++) {
        
        printf("level %d proc %d, ibox %d, pos %d, strong %d:%d, weak %d:%d\n",level,i,j,k,HS->jcptr[k],HS->kcptr[k],HS->kcptr[k],HS->jcptr[k+1]);
        printf("ir strong:\n");
        int l;
        for(l=HS->jcptr[k]; l<HS->kcptr[k]; l++) {
          printf("%d, ",HS->ir[l]);
        }
        printf("\nir weak:\n");
        for( ; l<HS->jcptr[k+1]; l++) {
          printf("%d, ",HS->ir[l]);
        }
        printf("\n");
      }
    }
  }
}
#endif
/*------------------------------------------------------------------------*/
#ifdef FMM_MPI
void MPexp_eval2send(const struct MPexp *this1,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
                 std::vector< std::vector<double> > *sendl,
                 std::vector< std::vector<int> > *sendlist,
                 MPI_Request *& requests)
/* OBSOLETE: Send points for direct evaluation to neighbouring processes. */
{
  const int Nf = this1->lptr[this1->nlevel+1]-this1->lptr[this1->nlevel];
  const int *restrict ix = this1->ix, *restrict ixptr = this1->ixptr;

  int comm_size = this1->nprocs-1;
  const struct mpHalo *HF = &this1->halo[this1->nlevel];
  
  MPI_Status status;
  
  // loop over procs
  for(int p=0,n=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    
    // only do proc if it has strong connections
    if(HF->sptr[p] > 0) {
      
      // list containing a flag (1 if connected) and npts in each box
      (*sendlist)[p].resize(2*Nf);
      
      // loop over strong connections in halo
      for(int j=0; j<HF->sptr[p]; j++, n++) {
        int ii = HF->ibox[n];
        // set flag if it's a connected box
        (*sendlist)[p][2*ii] = 1;
        // set npts
        (*sendlist)[p][2*ii+1] = ixptr[ii+1]-ixptr[ii];
      }
      
      // add up no. of points to send/receive
      int isum = 0;
      for(int j=0; j<Nf; j++)
      isum += (*sendlist)[p][2*j+1];
      
      // allocate coordinates send buffer
      (*sendl)[p].resize(4*isum);
      
      // fill send buffer with coords: zx[ix[...]], zy[...], zz[...], mm[...].
      for(int j=0, m=0; j<Nf; j++) {
        if((*sendlist)[p][2*j] == 1) {
          // also set a pxptr index in sendlist for retrieval on other proc
          (*sendlist)[p][2*j] = m;
          for (int k = ixptr[j]; k < ixptr[j+1]; k++, m++) {
            (*sendl)[p][4*m] = zx[ix[k]];
            (*sendl)[p][4*m+1] = zy[ix[k]];
            (*sendl)[p][4*m+2] = zz[ix[k]];
            (*sendl)[p][4*m+3] = mm[ix[k]];
          }
        }
      }
      
      // send sendlist
      MPI_Isend((*sendlist)[p].data(),(*sendlist)[p].size(),MPI_INT,i,2*this1->rank*this1->nprocs+i,MPI_COMM_WORLD,&requests[p]);
      MPI_Isend((*sendl)[p].data(),(*sendl)[p].size(),MPI_DOUBLE,i,2*this1->rank*this1->nprocs+i+comm_size,MPI_COMM_WORLD,&requests[p+comm_size]);

      cout << "proc " << this1->rank << " P2P send size " << (*sendl)[p].size() << " to proc " << i << " with tags " << 2*this1->rank*this1->nprocs+i << "," << 2*this1->rank*this1->nprocs+i+comm_size << endl;
    }
    // increment to start of next proc's strong connections
    n += HF->wptr[p]-HF->sptr[p];
  }
}
/*------------------------------------------------------------------------*/
int MPexp_procno(const struct MPexp *this1, int p) {
  /* This function returns the rank ID of a process given 
  the cyclic index p from 0 to nprocs-1 over halo arrays */
  return this1->rank+1+p - (this1->nprocs)*(this1->rank+p>this1->nprocs-2);
}
#endif
/*------------------------------------------------------------------------*/
void MPexp_eval2(const struct MPexp *this1,
                 double *pr,
                 const double *zx,const double *zy,const double *zz,
                 const double *mm,
                 int pot,SMOOTHER smooth)
/* Evaluation of the potential at the potentials themselves. Function
   is critical to performance. */
{
  // hopefully, some optimizing compiler can use this:
  const int Nf = this1->lptr[this1->nlevel+1]-this1->lptr[this1->nlevel];
  const struct mpexp *restrict offset = &this1->root[this1->lptr[this1->nlevel]];
  const int *restrict ix = this1->ix, *restrict ixptr = this1->ixptr;
  const struct mpSparse *restrict C = &this1->connect[this1->nlevel];
  const int *restrict jcptr = C->jcptr, *restrict kcptr = C->kcptr, *restrict ir = C->ir;

#ifdef FMM_MPI
  int comm_size = this1->nprocs-1;
  double start_1[comm_size];
  const struct mpHalo *HF = &this1->halo[this1->nlevel];
  
  MPI_Status status;
  MPI_Request *requests;
  requests = (MPI_Request*) malloc(4*comm_size*sizeof(MPI_Request));
  std::vector< std::vector<double> > recvl(comm_size);
  std::vector< std::vector<int> > recvlist(comm_size);
  std::vector< std::vector<double> > sendl(comm_size);
  std::vector< std::vector<int> > sendlist(comm_size);

  // loop over procs
  for(int p=0,n=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    
    // only do proc if it has strong connections
    if(HF->sptr[p] > 0) {
      
      // list containing a flag (1 if connected) and npts in each box
      sendlist[p].resize(2*Nf);
      
      // loop over strong connections in halo
      for(int j=0; j<HF->sptr[p]; j++, n++) {
        int ii = HF->ibox[n];
        // set flag if it's a connected box
        sendlist[p][2*ii] = 1;
        // set npts
        sendlist[p][2*ii+1] = ixptr[ii+1]-ixptr[ii];
      }
      
      // add up no. of points to send/receive
      int isum = 0;
      for(int j=0; j<Nf; j++)
      isum += sendlist[p][2*j+1];
      
      // allocate coordinates send buffer
      sendl[p].resize(4*isum);
      
      // fill send buffer with coords: zx[ix[...]], zy[...], zz[...], mm[...].
      for(int j=0, m=0; j<Nf; j++) {
        if(sendlist[p][2*j] == 1) {
          // also set a pxptr index in sendlist for retrieval on other proc
          sendlist[p][2*j] = m;
          for (int k = ixptr[j]; k < ixptr[j+1]; k++, m++) {
            sendl[p][4*m] = zx[ix[k]];
            sendl[p][4*m+1] = zy[ix[k]];
            sendl[p][4*m+2] = zz[ix[k]];
            sendl[p][4*m+3] = mm[ix[k]];
          }
        }
      }
      
      // send sendlist
      MPI_Isend(sendlist[p].data(),sendlist[p].size(),MPI_INT,i,2*this1->rank*this1->nprocs+i,MPI_COMM_WORLD,&requests[p]);
      MPI_Isend(sendl[p].data(),sendl[p].size(),MPI_DOUBLE,i,2*this1->rank*this1->nprocs+i+comm_size,MPI_COMM_WORLD,&requests[p+comm_size]);
    }
    // increment to start of next proc's strong connections
    n += HF->wptr[p]-HF->sptr[p];
  }
  
#endif

  // now proceed with P2P on this process
  dcmplx *Ynm = (dcmplx *)malloc((this1->pcoeff+1)*(this1->pcoeff+1)*sizeof(dcmplx));
#ifdef OMP
  omp_set_num_threads(1);
  #pragma omp parallel for schedule(guided)
#endif

  for (int j = 0; j < Nf; j++) {

    // far-field: contribution from distant potentials (local coeffs)
    mpexp_l2p(&offset[j],Ynm,this1->pcoeff,pr,zx,zy,zz,ix,ixptr[j],ixptr[j+1]);

    // near-field: symmetric interaction within this proc's list of near connections
    for (int jj = jcptr[j]; jj < kcptr[j]; jj++) {
      if (jcptr[j] == kcptr[j]){
        continue;
      }
      const int k = ir[jj];

      mpexp_directInteract2(pr,zx,zy,zz,mm,ix,
                            ixptr[j],ixptr[j+1],
                            ixptr[k],ixptr[k+1],
                            pot,smooth);
    }

    // near-field: direct interaction within box
    mpexp_directInteract2(pr,zx,zy,zz,mm,ix,ixptr[j],ixptr[j+1],
                          -1,ixptr[j+1],pot,smooth);
  }
  free(Ynm);

  // near-field: direct interaction with non-owned boxes
#ifdef FMM_MPI

  // non-blocking receive data
  for(int p=0,q=0; p<comm_size; p++) {
    
    if(HF->sptr[p] > 0) {

      // the other proc
      int i = MPexp_procno(this1, p);

      // TODO: figure out better way to receive
      recvlist[p].resize(2*Nf);
      // receive recvlist
      MPI_Irecv(recvlist[p].data(),2*Nf,MPI_INT,i,2*i*this1->nprocs+this1->rank,MPI_COMM_WORLD,&requests[p+2*comm_size]);

      // Probe message from proc i, then allocate recv buffer. Probe is blocking.
      int jsum = 0;
//      for(int j=0; j<Nf; j++) jsum += recvlist[p][2*j]>0;
      MPI_Probe(i, 2*i*this1->nprocs+this1->rank+comm_size, MPI_COMM_WORLD, &status);
      MPI_Get_count(&status, MPI_DOUBLE, &jsum);
      recvl[p].resize(jsum);
      MPI_Irecv(recvl[p].data(),recvl[p].size(),MPI_DOUBLE,i,2*i*this1->nprocs+this1->rank+comm_size,MPI_COMM_WORLD,&requests[p+3*comm_size]);
    }
  }


  // MPI flag
  int sum = 0;
  std::vector<int> beenthere(comm_size);
  for(int p=0; p<comm_size; p++) beenthere[p] = 0;
  
  // loop over procs until we find something that has arrived
  while(sum<comm_size) {
  
    // MPI_Waitany() method:
    /*int pp = MPI_Waitany(&requests[comm_size],&stats[comm_size]);
    // TODO: change halo so we can use MPI_Waitany().
    }
    sum++;*/
    
    // the long-winded method:
    // this method is necessary because of the way that
    // q is incremented with p.
    // if we want to choose any p, it would be easier
    // to store the halos as separate arrays for each process
    // rather than one long array.
    for(int p=0,q=0; p<comm_size; p++) {

      if(HF->sptr[p] > 0) {

        // the other proc
        int i = MPexp_procno(this1, p);

        // test if lists have arrived
        int complete = 0;
        MPI_Test(&requests[p+2*comm_size],&complete,&status);
        
        // test if points have arrived
        int complete2 = 0;
        MPI_Test(&requests[p+3*comm_size],&complete2,&status);

        // if so, and we haven't been here before, proceed with P2Ph.
        if(complete && complete2 && !beenthere[p]) {

          // call directInteract: loop over strong connections
          for (int j=0; j < HF->sptr[p]; j++, q++) {
            int ii = HF->ibox[q];
            int jj = HF->jbox[q];
            int pxptr = recvlist[p][2*jj];
            int npts = recvlist[p][2*jj+1];
            
            mpexp_directInteractHalo(pr,zx,zy,zz,
                                     ix,ixptr[ii],ixptr[ii+1],
                                     &recvl[p][0],pxptr,pxptr+npts,
                                     pot,smooth,this1->rank);
          }
          sum++;
          beenthere[p] = 1;
        }
        else {
          // increment to start of weak connections
          q += HF->sptr[p];
        }
      }
      // if there are unconnected procs, count them once so that sum is equal to comm_size.
      else {
        if(!beenthere[p]) {
          sum++;
          beenthere[p] = 1;
        }
      }
      
      // increment from start of weak connections to start of next proc's strong connections
      q += HF->wptr[p]-HF->sptr[p];
    }
  }
  
  for (int p=0; p<comm_size; p++) {
    if(HF->sptr[p] > 0) {
      MPI_Wait(&requests[p],&status);
      MPI_Wait(&requests[p+comm_size],&status);
    }
    recvlist[p].clear();
    recvl[p].clear();
    sendlist[p].clear();
    sendl[p].clear();
  }
  free(requests);

#endif
}
/*------------------------------------------------------------------------*/
void MPexp_eval(const struct MPexp *this1,
                double *qr,
                const double *ex,const double *ey,const double *ez,
                const double *zx,const double *zy,const double *zz,
                const double *mm,
                int pot,SMOOTHER smooth)
/* Evaluation of the potential at the points (ex,ey,ez). Function is
   critical to performance. */
{
  const int Nf = this1->lptr[this1->nlevel+1]-this1->lptr[this1->nlevel];
  const struct mpexp *restrict offset = &this1->root[this1->lptr[this1->nlevel]];
  const int *restrict ix = this1->ix, *restrict ixptr = this1->ixptr;
  const int *restrict jx = this1->jx, *restrict jxptr = this1->jxptr;
  const struct mpSparse *C = &this1->connect[this1->nlevel];
  const int *restrict jcptr = C->jcptr, *restrict kcptr = C->kcptr, *restrict ir = C->ir;

#ifdef FMM_MPI
  const struct mpHalo *HF = &this1->halo[this1->nlevel];
#endif

  dcmplx *Ynm = (dcmplx *)malloc((this1->pcoeff+1)*(this1->pcoeff+1)*sizeof(dcmplx));

  for (int j = 0,jj; j < Nf; j++) {

    // far-field: contribution from distant potentials
    mpexp_l2p(&offset[j],Ynm,this1->pcoeff,qr,ex,ey,ez,jx,jxptr[j],jxptr[j+1]);

    // near-field: symmetric interaction within list of near connections
    for (jj = jcptr[j]; jj < kcptr[j]; jj++) {
      const int k = ir[jj];
      mpexp_directInteract(qr,ex,ey,ez,
                           jx,jxptr[j],jxptr[j+1],
                           zx,zy,zz,mm,
                           ix,ixptr[k],ixptr[k+1],
                           pot,smooth);
      mpexp_directInteract(qr,ex,ey,ez,
                           jx,jxptr[k],jxptr[k+1],
                           zx,zy,zz,mm,
                           ix,ixptr[j],ixptr[j+1],
                           pot,smooth);
    }

    // near-field: direct interaction within box
    mpexp_directInteract(qr,ex,ey,ez,
                         jx,jxptr[j],jxptr[j+1],
                         zx,zy,zz,mm,
                         ix,ixptr[j],ixptr[j+1],
                         pot,smooth);
  }
  free(Ynm);
  
  // near-field: direct interaction with non-owned boxes
#ifdef FMM_MPI
  
  int comm_size = this1->nprocs-1;
  MPI_Status status;
  MPI_Request *requests;
  requests = (MPI_Request*) malloc(4*comm_size*sizeof(MPI_Request));
  std::vector< std::vector<double> > sendl(comm_size);
  std::vector< std::vector<double> > recvl(comm_size);
  std::vector< std::vector<int> > sendlist(comm_size);
  std::vector< std::vector<int> > recvlist(comm_size);
  
  // loop over procs
  int sendsize = 0;
  for(int p=0,n=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    
    sendlist[p].resize(2*Nf);
    recvlist[p].resize(2*Nf);
    
    // loop over strong connections in halo
    for(int j=0; j<HF->sptr[p]; j++, n++) {
      int ii = HF->ibox[n];
      // set flag if it's a connected box
      sendlist[p][2*ii] = 1;
      // set npts
      sendlist[p][2*ii+1] = ixptr[ii+1]-ixptr[ii];
    }
    
    // add up no. of points to send/receive
    int isum = 0;
    for(int j=0; j<Nf; j++)
    isum += sendlist[p][2*j+1];
    
    // allocate coordinates send buffer
    sendl[p].resize(4*isum);
    sendsize += isum;
    
    // fill send buffer with coords: zx[ix[...]], zy[...], zz[...], mm[...].
    for(int j=0, m=0; j<Nf; j++) {
      if(sendlist[p][2*j] == 1) {
        // also set a pxptr index in sendlist for retrieval on other proc
        sendlist[p][2*j] = m;
        for (int k = ixptr[j]; k < ixptr[j+1]; k++, m++) {
          sendl[p][4*m] = zx[ix[k]];
          sendl[p][4*m+1] = zy[ix[k]];
          sendl[p][4*m+2] = zz[ix[k]];
          sendl[p][4*m+3] = mm[ix[k]];
        }
      }
    }
    
    // send sendlist and points
    MPI_Isend(sendlist[p].data(),2*Nf,MPI_INT,i,1,MPI_COMM_WORLD,&requests[p]);
    //      cout << "rank " << this1->rank << " P2P send size " << sendl[proc].size() << " to proc " << proc << endl;
    if(sendl[p].size() > 0) {
      MPI_Isend(sendl[p].data(),sendl[p].size(),MPI_DOUBLE,i,1,MPI_COMM_WORLD,&requests[p+2*comm_size]);
    }
    
    // receive recvlist
    MPI_Irecv(recvlist[p].data(),2*Nf,MPI_INT,i,1,MPI_COMM_WORLD,&requests[p+comm_size]);
    
    // increment to start of next proc's strong connections
    n += HF->wptr[p]-HF->sptr[p];
  }

  // loop over procs again.
  for(int p=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    
    // Wait for Isend of sendlist
    MPI_Wait(&requests[p],&status);
    // Wait for Irecv of recvlist
    MPI_Wait(&requests[p+comm_size],&status);
    
    // add up no. of points to be received
    int jsum = 0;
    for(int j=0; j<Nf; j++)
    jsum += recvlist[p][2*j+1];
    recvl[p].resize(4*jsum);
    
    // receive points
    if(recvl[p].size() > 0) {
      MPI_Irecv(recvl[p].data(),recvl[p].size(),MPI_DOUBLE,i,1,MPI_COMM_WORLD,&requests[p+3*comm_size]);
    }
  }
  
  // loop over procs a 3rd time.
  for(int p=0,q=0; p<comm_size; p++) {
    // the other proc
    int i = MPexp_procno(this1, p);
    
    // Wait for points Isend/Irecv to complete
    if(sendl[p].size() > 0) {
      MPI_Wait(&requests[p+2*comm_size],&status);
    }
    if(recvl[p].size() > 0) {
      MPI_Wait(&requests[p+3*comm_size],&status);
    }
    
    // call directInteract: loop over strong connections
    for (int j=0; j < HF->sptr[p]; j++, q++) {
      int ii = HF->ibox[q];
      int jj = HF->jbox[q];
      int pxptr = recvlist[p][2*jj];
      int npts = recvlist[p][2*jj+1];
      
      mpexp_directInteractHalo(qr,ex,ey,ez,
                               jx,jxptr[ii],jxptr[ii+1],
                               &recvl[p][0],pxptr,pxptr+npts,
                               pot,smooth,this1->rank);
    }
    
    // increment to start of next proc's strong connections
    q += HF->wptr[p]-HF->sptr[p];
  }

  free(requests);
  for (int i=0; i<comm_size; i++) {
    sendlist[i].clear();
    recvlist[i].clear();
    sendl[i].clear();
    recvl[i].clear();
  }
  
#endif
}
/*------------------------------------------------------------------------*/
void mpexp_p2m(struct mpexp *this1,dcmplx *Ynm,int pcoeff,
    const double zx,const double zy,const double zz,
                const double mm,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mm) at (zx,zy,zz) about (xo,y0,z0). */
{
  // multipole coefficients
  dcmplx *coeff = this1->coeff1;
  double sph[3];
  cart2spher(zx-this1->x0, zy-this1->y0, zz-this1->z0,sph);

  evalMultipole(sph[0], sph[1], sph[2], &Ynm);
  for (int n=0, k=0; n<=pcoeff; n++){
    for (int m=0; m<=n; m++, k++){
      int nm  = n * n + n - m;
      int nms = n * (n + 1) / 2 + m;
      coeff[nms] += mm * Ynm[nm];
    }
  }
}
/*------------------------------------------------------------------------*/
bool mpexp_theta(const struct mpexp *this1,const struct mpexp *that)
/* Theta criterion for accurate cluster to cluster interaction between
   two boxes this and that.

   This routine is a workhorse and needs to be fast.

   The maximum relative error in using a multipole expansion from a
   box 'this' to evaluate the field in a box 'that' is error <=
   constant * (R/|R+minimum distance between the boxes|)^(p+1) for
   points |z| > R. Using that the minimum distance is >= d-R-r (where
   d is the center-to-center distance and where r is the radius of the
   smaller box), we are led to the requirement that R/(d-r) <= theta
   and this is what is implemented below (see [4] for the full
   analysis).

   This also means that theta = 1/2 is equivalent to the boxes being
   classically well-separated as in [1,2]. With a general theta, any
   of the two sets may be expanded by a factor of 1/theta and
   arbitrarily rotated about its center point without touching the
   other set.

   However, in that case one should note that the theta used in the
   classical uniform multipole method [2] is actually 1/(2*sqrt(2)-1)
   = 0.5469181 and not 1/2 -- perhaps for convenience, given the
   quadratic boxes.

   Also, in [3] a different and less stringent criterion is used; R <=
   c*d. For uniform boxes, their c = 1/2 would mean theta = sqrt(2)/3
   = 0.4714045 here. Note that trivially, R <= theta/(1+theta)*d
   implies R+theta*r <= theta*d, but the former criterion is less
   sharp when the two boxes have different sizes.

   A slightly sharper result stems from squaring and using Young's
   weighted inequality; then R^2+theta*r^2 <= theta^2/(1+theta)*d^2
   implies R+theta*r <= theta*d. This is used below as a first check
   in order to avoid computing the square-root.

   Also, note that the GKZ-criterion in [3] yields an error
   proportional to (c/(1-2*c))^(p+1) and not to c^(p+1) as they
   claim. Hence c = theta/(1+2*theta) would be appropriate -- this is
   even less sharp compared to the trivial interpretation mentioned
   above. Unfortunately, following their method of proof, the
   currently implemented criterion yields an error proportional to
   (theta/(1-theta))^(p+1), which would suggest that the choice theta
   = 1/2 should be avoided.

   Notes: the GKZ-criterion is found in [3], Eq. (8.31) at p. 379 and
   the error estimate is derived at p. 386. The first estimate [3] in
   (8.40) is incorrect (change (1+2*theta) to (1-2*theta) and the
   second is unsharp (replace (1-2*theta) with 1/(1+2*theta)). In the
   proof that follows, the problematic factor (1-2*theta) is
   inadvertently omitted.

   References:
     [1] J. Carrier, L. Greengard, and V. Rokhlin: "A fast adaptive
         multipole algorithm for particle simulations". SIAM
         J. Sci. Stat. Comput., 9(4):669--686, 1988.
     [2] L. Greengard and V. Rokhlin: "A fast algorithm for particle
         simulations". J. Comput. Phys., 73:325--348, 1987.
     [3] M. Griebel, S. Knapek, and G. Zumbusch: "Numerical Simulation
         in Molecular Dynamics", volume 5 of Texts in Computational
         Science and Engineering. Springer Verlag, Berlin, 2007.
     [4] S. Engblom: "On well-separated sets and fast mutlipole
         methods". Appl. Numer. Math. 61(10):1096--1102, 2011.
*/
{
  const double theta2_ = theta*theta;
  const double thetac_ = theta2_/(1.0+theta);

  const double thisr = MPexp_distance(this1->dx0,this1->dy0,this1->dz0);
  const double thatr = MPexp_distance(that->dx0,that->dy0,that->dz0);
  const double r     = MPexp_distance(this1->x0-that->x0, this1->y0-that->y0, this1->z0-that->z0);

  const double thisrad = thisr*thisr;
  const double thatrad = thatr*thatr;
  const double d2 = r*r;
  
  /* note: criterion uses strict inequality to get rid of the singular
     case when both boxes are the same points (this can happen!) */
  //printf("theta crit: %f,%f\n",thisrad+theta*thatrad,thetac_*d2);
  if (thisrad >= thatrad) {
    // sufficient condition (return probability is about 80%)
    if (thisrad+theta*thatrad < thetac_*d2)
      return true;
    else
      return thisrad+theta2_*thatrad+
        2.0*theta*sqrt(thisrad*thatrad) < theta2_*d2;
  }
  else {
    if (thatrad+theta*thisrad < thetac_*d2)
      return true;
    else
      return thatrad+theta2_*thisrad+
        2.0*theta*sqrt(thatrad*thisrad) < theta2_*d2;
  }
}
/*------------------------------------------------------------------------*/
void mpexp_l2p(const struct mpexp *this1,dcmplx *Ynm,int pcoeff,
                double *vr,
                const double *zx,const double *zy,const double *zz,
                const int *ix,int begin,int end)
/* Evaluation of far field (local expansion) at coordinates (zx,zy,zz). */
{
  const dcmplx *coeff = this1->coeff2;
  
  for (int i = begin; i < end; i++) {
    dcmplx val = 0.0;
    dcmplx tmp = 0.0;
    double sph[3];
    cart2spher(zx[ix[i]]-this1->x0, zy[ix[i]]-this1->y0, zz[ix[i]]-this1->z0,sph);
    evalMultipole(sph[0], sph[1], sph[2], &Ynm);

    for (int n=0, k=0; n<=pcoeff; n++){
      int nm  = n * n + n;
      int nms = n * (n + 1) / 2;
      
      tmp = coeff[nms] * Ynm[nm];
      val += creal(tmp);
//      val += creal(coeff[nms] * Ynm[nm]);

      for (int m=1; m<=n; m++, k++) {
        int nm  = n * n + n + m;
        int nms = n * (n + 1) / 2 + m;
        tmp = coeff[nms] * Ynm[nm];
        val += 2.0*creal(tmp);
//        val += 2.0 * creal(coeff[nms] * Ynm[nm]);
      }
    }

    // add the result
    vr[ix[i]] += creal(val);
  }
}
/*------------------------------------------------------------------------*/
