/* fmm.h -- Fast multipole method for 2-D potentials in free space */

/* S. Engblom and J. Bull 2015-09-02 (Major port to 3D) */
/* S. Engblom 2009-10-08 (Revision) */
/* S. Engblom 2009-04-15 (Major revision, adaptivity) */
/* S. Engblom 2007-07-08 (Major revision, performance) */
/* S. Engblom 2007-06-19 */

#ifndef __fmm_h
#define __fmm_h

#include <math.h>
#include <float.h>

#if !defined FMM_MPI || !defined __cplusplus
#include <stdbool.h>
#endif

#ifndef M_PI
#define M_PI 3.1415926535897932384626434
#endif

// MPI bindings
#ifdef FMM_MPI
  #include <mpi.h>
#endif

#ifdef OMP
  #include <omp.h>
#endif

// C++ bindings
#if defined FMM_MPI || defined __cplusplus
  // Matlab external libraries
  #ifndef NOMEX
    #include "mat.h"
    #include "mex.h"
  #endif
  #include <complex>
  #include <vector>
  #include <assert.h>
  #include <cstdlib>
  #include <fstream>
  #include <iostream>
  #include <cstring>
  //#define creal(z) ((z).real())
  //#define cimag(z) ((z).imag())
  #define creal(z) (((double*)&(z))[0]) //code runs about 3% faster with these crude definitions,
  #define cimag(z) (((double*)&(z))[1])
  #define clog(x) (log(x))
  #define cabs(x) (abs(x))
  #define cexp(x) (exp(x))
  #define restrict __restrict
  
  inline double log2(double x)
  {
    return log(x)/0.69314718055994530942;
  }

  using namespace std;

  typedef complex<double> dcmplx;
  const dcmplx I = dcmplx(0.0,1.0);

// C bindings
#else
  #include "mat.h"
  #include "mex.h"
  #include <complex.h>
  #include <assert.h>
  #include <stdio.h>
  #include <string.h>

  typedef double _Complex dcmplx;

#endif

// type of smoother
typedef enum {DIRAC = 1,ERR_SMOOTHER = -1} SMOOTHER;

// for performance experiments:
#ifdef PRINTTIME
  #include <time.h>
  #ifdef FMM_MPI
    static double TIME_before,TIME_after;
    #define StartTime TIME_before = MPI_Wtime()
    #define StopTime TIME_after = MPI_Wtime()
    #define PrintTime(s) printf(s, \
              rank,TIME_after-TIME_before)
  #else
    static clock_t TIME_before,TIME_after;
    #define StartTime TIME_before = clock()
    #define StopTime TIME_after = clock()
    #define PrintTime(s) printf(s, \
              rank,(double)(TIME_after-TIME_before)/CLOCKS_PER_SEC)
  #endif /* FMM_MPI */
#else
  #define StartTime
  #define StopTime
  #define PrintTime(s)
#endif /* PRINTTIME */

/* Criterion for dividing a box. If one side is more than meshratio
   times the other, then the longer side is split into two. Otherwise
   it is split according to an ordering which yields good cache
   effects. */
#ifndef FMM_MESHRATIO
#define FMM_MESHRATIO 2.0
#endif
const static double meshratio = FMM_MESHRATIO; // 1.0 <= meshratio

/* Multipole expansion structure. The coefficients are to be
   understood as
      mpexp(z) = *** TODO
   or, depending on the stage of the algorithm,
      mpexp(z) = *** TODO
*/
struct mpexp {
  // midpoint coords and distance to upper right back corner
  double x0,y0,z0,dx0,dy0,dz0;
  // outgoing/incoming multipole coefficients
  dcmplx *coeff1,*coeff2;
};

/* Sparse connectivity matrix (special CCS format integer matrix): the
   row indices in ir of column j are found in [jcptr[j],jcptr[j+1]),
   the near connections (think of these as the value 1) in
   [jcptr[j],kcptr[j]) and the far connections (i.e. value -1) in
   [kcptr[j],jcptr[j+1]). The row indices of the near part is sorted
   in ascending- and the far part in descending order,
   respectively. Only the strictly lower triangular part is stored.

   For far connections between two levels (value -2), a negative index
   will occur in ir, pointing to a box at the previous levels. Then
   seven positive indices follow, pointing to children of the previous
   negative index. Those indices are ignored in the FMM assembly and are
   a by-product of the algorithm.

   At the finest level a special syntax is used in which the near part
   is grouped into a 'truly' near part (value 1) in [jptr[j],kcptr[j+N])
   and a 'less strongly' connected part (value 2) in
   [kcptr[j+N],kcptr[j]). In this case the near part is not sorted and
   also, the row indices are signed (positive for cluster to particle,
   negative for particle to cluster connection). */
struct mpSparse {
  int *jcptr,*kcptr; /* column pointer and pointer to the beginning of
                        the 'far'-part */
  int *ir;           // row indices
};

#ifdef FMM_MPI
/* Very sparse matrix ("index sparse") to store halo */
struct mpHalo {
  int halo_nnz; // length of list of nonzero entries
  int *ibox; // 'from' box pointer on this proc
  int *jbox; // 'to' box pointer to boxes on other procs
  int *sptr; // pointer to end of the 'near' part
  int *wptr; // pointer to end of the 'far' part
};
#endif

/* Multipole expansion tree in vectorized form. First comes the root,
   then the 8 children of the root and so on */
// this type is on the tree level
struct MPexp {
  double xmin,xmax,
    ymin,ymax,
    zmin,zmax;       // bounding box
  int pcoeff;        /* number of coefficients offset by one; coeff[0]
                        through coeff[pcoeff] */
  int nlevel;        /* number of levels offset by one; level =
                        0..nlevel */
  int *lptr;         /* pointer to levels in tree; level l can be
                        found in root[i] where lptr[l] <= i <
                        lptr[l+1] and lptr[nlevel+1] is the total
                        number of boxes. */
  struct mpexp *root;       // box 0, the root of the tree
  struct mpSparse *connect; // connectivity information for each level
  int *ix,*ixptr;    // permutation of potentials
  int *jx,*jxptr;    // permutation of evaluation points
    int *iOrder, *jOrder;

#ifdef FMM_MPI
  int nprocs;
  int rank;
  struct mpHalo *halo; // connectivity information between subdomains (procs)
#endif
};

/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (mm) at the positions (zx,zy,zz). */
void fmm3dInteract(int rank,int nprocs,int N,
                   const double *zx,const double *zy,const double *zz,
                   const double *mm,
                   int NE,
                   const double *ex,const double *ey,const double *ez,
                   double *pr,double *qr,
                   int pot,SMOOTHER smooth,
                   double tol,int pcoeff,double theta,
                   int *Ndirect,int *Nlevels,double eta);

/* Returns the multipole mesh and sparse connection matrices. For
   debugging and pedagogical purposes. */
#ifdef FMM_MPI
#ifndef NOMEX
void fmm3dGetMesh(mxArray *m[],mxArray *c[],mxArray *h[],mxArray *s[],mxArray *t[],
      int rank,int nprocs,int N,
                  const double *zx,const double *zy,const double *zz,
                  const double *mrm,
                  int NE,
                  const double *ex,const double *ey,const double *ez,
                  int pot,SMOOTHER smooth,
                  double tol,int pcoeff,double theta,
                  int *Ndirect,int *Nlevels,double eta);
#endif
#else
void fmm3dGetMesh(int nlhs,mxArray *plhs[],
                  int N,
                  const double *zx,const double *zy,const double *zz,
                  const double *mm,
                  int NE,
                  const double *ex,const double *ey,const double *ez,
                  int pot,SMOOTHER smooth,
                  double tol,int pcoeff,double theta,
                  int *Ndirect,int *Nlevels,double eta);
#endif

#endif /* __fmm_h */
 
