/* fmm3d.c -- Fast multipole method for 3-D potentials in free space */

/* J. Bull 2015-09-02 (Major port to 3D) */
/* S. Engblom 2009-05-08 (Revision) */
/* S. Engblom 2007-07-08 (Major revision) */
/* S. Engblom 2006-10-11 (Port to Mex and a major revision) */
/* S. Engblom 2005-01-05 */

#include <string.h>

#include "mex.h"
#include "matrix.h"
#include "fmm.h"
#include "direct.h"

// forward declarations
typedef enum {ETA = 1,PCOEFF = 2,NLEVELS = 3,SMOOTH = 4,
	      POT = 6,OUT = 7,NDIRECT = 9,TOL = 10,THETA = 12,
	      ERR_PROPERTY = -1} PROPERTY;
typedef void (*getvalFun)(void *,const mxArray *);

void getETA(void *,const mxArray *);
void getPCOEFF(void *,const mxArray *);
void getNLEVELS(void *,const mxArray *);
void getSMOOTH(void *,const mxArray *);
void getPOT(void *,const mxArray *);
void getOUT(void *,const mxArray *);
void getNDIRECT(void *,const mxArray *);
void getTOL(void *,const mxArray *);
void getTHETA(void *,const mxArray *);

PROPERTY getprop(const char *str);
typedef enum {MESH = 0,SOL = 1,TIME = 2,ERR_OUT = -1} OUT_TYPE;
OUT_TYPE getout(const char *str);
SMOOTHER getsmooth(const char *str);

#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
			   mxGetNumberOfDimensions(A) == 2)
#define ISREALSCALAR(A) (!mxIsCell(A) && !mxIsStruct(A) && \
			 !mxIsComplex(A) && mxGetNumberOfElements(A) == 1)

/*------------------------------------------------------------------------*/
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  // input
  const double *zx,*zy,*zz,*ex,*ey,*ez,*mm;
  int N,NE = 0;
  const int rank = 0,nprocs = 1;

  // values of properties with defaults (-1 for empty)
  double tol = -1.0,eta = 1.0,theta = 0.5;
  int pot = 1,Ndirect = -1,pcoeff = -1,Nlevels = -1;
  OUT_TYPE out = SOL;
  SMOOTHER smooth = DIRAC;
  // (note: defaults for tol and Ndirect handled later)

  // output
  double *pr = NULL,*qr = NULL,*timing = NULL;
  mxArray *Timing = NULL;
  int eval_syntax = 2 < nrhs && ISDOUBLEMATRIX(prhs[2]);

  // parsing functions
  static const getvalFun getval[] = {NULL,&getETA,&getPCOEFF,&getNLEVELS,
				     &getSMOOTH,NULL,&getPOT,&getOUT,NULL,
				     &getNDIRECT,&getTOL,NULL,&getTHETA};
  void *val[] = {NULL,&eta,&pcoeff,&Nlevels,&smooth,NULL,&pot,&out,NULL,
		 &Ndirect,&tol,NULL,&theta};

  // syntax
  if (nrhs < 2 || (nrhs-eval_syntax)&1)
    mexErrMsgTxt("Expecting two or three inputs "
		 "followed by property/value-pairs.");

  // inputs Z and M
  if (!ISDOUBLEMATRIX(prhs[0]) || !ISDOUBLEMATRIX(prhs[1]) ||
      mxIsComplex(prhs[0]) || mxIsComplex(prhs[1]) ||
      mxGetM(prhs[0]) != mxGetM(prhs[1]) ||
      mxGetN(prhs[0]) != 3 || mxGetN(prhs[1]) != 1)
    mexErrMsgTxt("First two arguments must be double real matrices of "
		 "dimensions N-by-3 and N, respectively.");
  N = mxGetM(prhs[0]);
  // replace mxGetPr (soon to be deprecated) with mxGetDoubles (introduced in R2018a)
#if MX_HAS_INTERLEAVED_COMPLEX
  zx = mxGetDoubles(prhs[0]);
  zy = mxGetDoubles(prhs[0])+N;
  zz = mxGetDoubles(prhs[0])+2*N;
  mm = mxGetDoubles(prhs[1]);
#else
  zx = mxGetPr(prhs[0]);
  zy = mxGetPr(prhs[0])+N;
  zz = mxGetPr(prhs[0])+2*N;
  mm = mxGetPr(prhs[1]);
#endif

  // input ZE -- or just (ex,ey,ez) for short
  if (eval_syntax) {
    if (!ISDOUBLEMATRIX(prhs[2]) || mxIsComplex(prhs[2]) ||
	mxGetN(prhs[2]) != 3)
      mexErrMsgTxt("Third argument must be double real matrix of "
		   "dimensions NE-by-3.");
    NE = mxGetM(prhs[2]);
#if MX_HAS_INTERLEAVED_COMPLEX
    ex = mxGetDoubles(prhs[2]);
    ey = mxGetDoubles(prhs[2])+NE;
    ez = mxGetDoubles(prhs[2])+2*NE;
#else
    ex = mxGetPr(prhs[2]);
    ey = mxGetPr(prhs[2])+NE;
    ez = mxGetPr(prhs[2])+2*NE;
#endif
  }

  // parse property/value-pairs in an automatic fashion
  for (int i = 2+eval_syntax; i < nrhs-1; i += 2) {
    PROPERTY prop;
    char str[10];

    // determine property
    if (!mxIsChar(prhs[i]))
      mexErrMsgTxt("Expecting a property name.");
    if (mxGetString(prhs[i],str,10) != 0 ||
	(prop = getprop(str)) == ERR_PROPERTY)
      mexErrMsgTxt("Unknown property.");

    // get its value
    getval[prop](val[prop],prhs[i+1]);
  }
  
  // defaults for tol and Ndirect
  if (tol < 0.0 && pcoeff < 0)
    tol = 1e-6;
  else if (tol > 0 && pcoeff > 0)
    mexErrMsgTxt("Specifying both options 'tol' and 'pcoeff' not allowed.");
  if (Ndirect < 0 && Nlevels < 0)
    Ndirect = 100;
  else if (Ndirect > 0 && Nlevels > 0)
    mexErrMsgTxt("Specifying both options 'ndirect' and 'nlevels' not allowed.");

  // remaining checks
  if (out != MESH) {
    if (nlhs > 1 && !eval_syntax)
      mexErrMsgTxt("Three input matrices required for two outputs.");
  }
  else {
    if (nlhs > 3+eval_syntax)
      mexErrMsgTxt("Too many outputs for this syntax.");
  }

  // allocate and produce output
  if (out == MESH) {
    // special syntax: return FMM-mesh only
    fmm3dGetMesh(nlhs,plhs,N,zx,zy,zz,mm,NE,ex,ey,ez,
		 pot,smooth,
		 tol,pcoeff,theta,
		 &Ndirect,&Nlevels,eta
		 );
    return;
  }
  else {
    if (out == TIME) {
      Timing = mxCreateDoubleMatrix(32,1,mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
      timing = mxGetDoubles(Timing);
#else
      timing = mxGetPr(Timing);
#endif
    }
  
    if (nlhs > 1) {
      // evaluate both at potentials Z and at coordinates ZE
      plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),1,mxREAL);
      plhs[1] = mxCreateDoubleMatrix(mxGetM(prhs[2]),1,mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
      pr = mxGetDoubles(plhs[0]);
      qr = mxGetDoubles(plhs[1]);
#else
      pr = mxGetPr(plhs[0]);
      qr = mxGetPr(plhs[1]);
#endif
    }
    else if (!eval_syntax || (zx == ex && zy == ey && zz == ez && N == NE)) {
      // only at potentials Z
      plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[0]),1,mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
      pr = mxGetDoubles(plhs[0]);
#else
      pr = mxGetPr(plhs[0]);
#endif
    }
    else {
      // at coordinates ZE
      plhs[0] = mxCreateDoubleMatrix(mxGetM(prhs[2]),1,mxREAL);
#if MX_HAS_INTERLEAVED_COMPLEX
      qr = mxGetDoubles(plhs[0]);
#else
      qr = mxGetPr(plhs[0]);
#endif
    }

    // evaluate
    if (tol == 0.0)
      directInteract(N,zx,zy,zz,mm,NE,ex,ey,ez,pr,qr,
		     pot,smooth);
    else
      fmm3dInteract(rank,nprocs, // MPI
		    N,zx,zy,zz,mm,NE,ex,ey,ez,pr,qr, // input
		    pot,smooth, // potential
		    tol,pcoeff,theta, // FMM-expansion
		    &Ndirect,&Nlevels,eta // FMM-mesh
		    );
  }

  // possibly return only the time
  if (out == TIME) {
    mxDestroyArray(plhs[0]);
    plhs[0] = Timing;
  }
}
/*------------------------------------------------------------------------*/
PROPERTY getprop(const char *str)
/* Small perfect hash for the supported properties. Note that str[0]
   must be readable for this to work. Returns ERR_PROPERTY on
   failure. */
{
  static const char *tab[] = {"","eta","pcoeff","nlevels","smooth","",
			      "pot","out","","ndirect","tol","","theta"};
  const unsigned hash = (((unsigned)str[0] * 1)+((unsigned)str[1] * 9))%13;

  return str[0] != '\0' && strcmp(str,tab[hash]) == 0 ? 
    (PROPERTY)hash : ERR_PROPERTY;
}
/*------------------------------------------------------------------------*/
void getETA(void *eta_,const mxArray *rhs)
{
  double *eta = (double *)eta_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'eta'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *eta = *mxGetDoubles(rhs);
#else
  *eta = *mxGetPr(rhs);
#endif
  if (*eta < 0 || 1 < *eta)
    mexErrMsgTxt("Property 'eta' must be in [0,1].");
}
/*------------------------------------------------------------------------*/
void getPCOEFF(void *pcoeff_,const mxArray *rhs)
{
  int *pcoeff = (int *)pcoeff_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'pcoeff'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *pcoeff = *mxGetInt8s(rhs);
#else
  *pcoeff = *mxGetPr(rhs);
#endif
  if (*pcoeff <= 0)
    mexErrMsgTxt("Property 'pcoeff' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getNLEVELS(void *Nlevels_,const mxArray *rhs)
{
  int *Nlevels = (int *)Nlevels_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'nlevels'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *Nlevels = *mxGetInt8s(rhs);
#else
  *Nlevels = *mxGetPr(rhs);
#endif
  if (*Nlevels < 0)
    mexErrMsgTxt("Property 'nlevels' must be non-negative.");
}
/*------------------------------------------------------------------------*/
void getSMOOTH(void *smooth_,const mxArray *rhs)
{
  SMOOTHER *smooth = (SMOOTHER *)smooth_;
  char str[10];

  if (!mxIsChar(rhs))
    mexErrMsgTxt("Expecting a character array as option 'smooth'.");
  if (mxGetString(rhs,str,10) != 0 || 
      (*smooth = getsmooth(str)) == ERR_SMOOTHER)
    mexErrMsgTxt("Unknown smoother.");
}
/*------------------------------------------------------------------------*/
void getPOT(void *pot_,const mxArray *rhs)
{
  int *pot = (int *)pot_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'pot'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *pot = *mxGetInt8s(rhs);
#else
  *pot = *mxGetPr(rhs);
#endif
  if (*pot != 1)
    mexErrMsgTxt("Property 'pot' must be 1.");
}
/*------------------------------------------------------------------------*/
void getOUT(void *out_,const mxArray *rhs)
{
  OUT_TYPE *out = (OUT_TYPE *)out_;
  char str[10];

  if (!mxIsChar(rhs))
    mexErrMsgTxt("Expecting a character array as option 'out'.");
  if (mxGetString(rhs,str,10) != 0 || 
      (*out = getout(str)) == ERR_OUT)
    mexErrMsgTxt("Unknown case of option 'out'.");
}
/*------------------------------------------------------------------------*/
void getNDIRECT(void *Ndirect_,const mxArray *rhs)
{
  int *Ndirect = (int *)Ndirect_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'ndirect'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *Ndirect = *mxGetInt8s(rhs);
#else
  *Ndirect = *mxGetPr(rhs);
#endif
  if (*Ndirect <= 0)
    mexErrMsgTxt("Property 'ndirect' must be strictly positive.");
}
/*------------------------------------------------------------------------*/
void getTOL(void *tol_,const mxArray *rhs)
{
  double *tol = (double *)tol_;

  if (!ISREALSCALAR(rhs))
      mexErrMsgTxt("Expecting a real and scalar option 'tol'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *tol = *mxGetDoubles(rhs);
#else
  *tol = *mxGetPr(rhs);
#endif
  if (*tol < 0 || 1 < *tol)
    mexErrMsgTxt("Property 'tol' must be in (0,1).");
}
/*------------------------------------------------------------------------*/
void getTHETA(void *theta_,const mxArray *rhs)
{
  double *theta = (double *)theta_;

  if (!ISREALSCALAR(rhs))
    mexErrMsgTxt("Expecting a real and scalar option 'theta'.");
#if MX_HAS_INTERLEAVED_COMPLEX
  *theta = *mxGetDoubles(rhs);
#else
  *theta = *mxGetPr(rhs);
#endif
  if (*theta <= 0 || 1 <= *theta)
    mexErrMsgTxt("Property 'theta' must be in (0,1).");
}
/*------------------------------------------------------------------------*/
OUT_TYPE getout(const char *str)
/* Small perfect hash for the supported out-options. Note that str[0]
   must be readable for this to work. Returns ERR_OUT on failure. */
{
  static const char *tab[] = {"mesh","sol","time"};
  const unsigned hash = ((unsigned)str[0]+(unsigned)str[1])%3;

  return strcmp(str,tab[hash]) == 0 ? (OUT_TYPE)hash : ERR_OUT;
}
/*------------------------------------------------------------------------*/
SMOOTHER getsmooth(const char *str)
/* Small perfect hash for the supported smoothers. Note that str[0]
   must be readable for this to work. Returns ERR_SMOOTHER on
   failure. */
{
  static const char *tab[] = {"dirac"};
  const unsigned hash = ((unsigned)str[0])%1;

  return strcmp(str,tab[hash]) == 0 ? (SMOOTHER)hash : ERR_SMOOTHER;
}
/*------------------------------------------------------------------------*/
