/* fmm3d.c -- Fast multipole method for 3-D potentials in free space */

/* J. Bull 2015-11-16 (Port main function to C and parallelise) */
/* J. Bull 2015-09-02 (3D) */
/* S. Engblom 2009-05-08 (Revision) */
/* S. Engblom 2007-07-08 (Major revision) */
/* S. Engblom 2006-10-11 (Port to Mex and a major revision) */
/* S. Engblom 2005-01-05 */

#include <string.h>

#include "fmm.h"
#include "direct.h"

#ifdef FMM_MPI
  #include <fstream>
  #include "io.h"
#endif

/* Arguments are:
   argv(0) input file (points z and masses m),
   argv(1) output file (potentials or mesh),
   argv(2) optional: additional input file (sampling locations ze),
   the rest can be in any order and are all optional (use defaults):
   tolerance, potential type, flag to output potentials or mesh, Ndirect,
   smooth, output file
*/
/*------------------------------------------------------------------------*/
#ifdef FMM_MPI
int main(int argc, char *argv[])
{
  // for MPI
  int rank = 0;
  int nprocs = 1;

  // Initialize MPI
  int already_initialised,already_finalised;
  MPI_Initialized(&already_initialised);
  
  if (!already_initialised) {
    MPI_Init(NULL, NULL);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  }
  
  StartTime;

  if(rank == 0) {
    cout << "Running ";
    if(nprocs > 1)
      cout << "mpirun -n " << nprocs << " ";
    for(int i=0; i<argc; i++) {
      cout << argv[i] << " ";
    }
    cout << endl;
  }
  
  // main input
  double *zx,*zy,*zz,*ex,*ey,*ez,*mm;
  int N,NE = 0;

  // output
  double *pr = NULL,*qr = NULL;
#ifndef NOMEX
  mxArray *m = NULL,*c = NULL,*h = NULL,*s = NULL,*t = NULL;
#endif
  char *output_file = NULL;
  
  // values of properties with defaults (-1 for empty)
  double tol = -1.0,eta = 1.0,theta = 0.5;
  int pot = 1,Ndirect = -1,pcoeff = -1,Nlevels = -1;
  int out = 0;
  SMOOTHER smooth = DIRAC;

  double starttime = MPI_Wtime();

  // not all of the options are currently supported
  ReadInput(nprocs,rank,argc,argv,&output_file,
	    &N,&NE,&zx,&zy,&zz,&mm,&ex,&ey,&ez,
            &tol,&pot,&Ndirect,&pcoeff,
            &Nlevels,&out,&smooth,&eta,&theta);

  ofstream fp;

  // allocate output arrays
  if (out == -1) {
#ifdef NOMEX
    cout << "Warning: if compiled with NOMEX, you cannot output mesh at present." << endl;
#else
    // special syntax: return FMM-mesh only
    fmm3dGetMesh(&m,&c,&h,&s,&t,rank,nprocs,
    N,zx,zy,zz,mm,NE,ex,ey,ez,
    pot,smooth,
    tol,pcoeff,theta,
    &Ndirect,&Nlevels,eta);
#endif
  }
  else if (out == 0) {
    // only at potentials Z
    pr = (double*) malloc(N*sizeof(double));
    for (int i=0; i<N; i++) {
      pr[i] = 0.0;
    }
  }
  else if (out == 1) {
    // at coordinates E
    qr = (double*) malloc(NE*sizeof(double));
    for (int i=0; i<NE; i++) {
      qr[i] = 0.0;
    }
  }
  else if (out == 2) {
    // evaluate both at potentials Z and at coordinates E
    pr = (double*) malloc(N*sizeof(double));
    qr = (double*) malloc(NE*sizeof(double));
    for (int i=0; i<N; i++) {
      pr[i] = 0.0;
    }
    for (int i=0; i<NE; i++) {
      qr[i] = 0.0;
    }
  }

  // timing for read and allocate output arrays
  if (!rank)
  {
    fp.open("perf.log", ofstream::app);
  
    if(!fp.is_open()) {
      cout << "perf.log is unavailable" << endl;
    }
    fp << nprocs << " 1 " << N << " " << tol << " " << Nlevels << " " << MPI_Wtime() - starttime << endl;
    fp.close();
  }

  // evaluate
  if(out != -1) {
    if (tol == 0.0)
      directInteract(N,zx,zy,zz,mm,NE,ex,ey,ez,pr,qr,
                     pot,smooth);
    else
      fmm3dInteract(rank,nprocs, // MPI
                    N,zx,zy,zz,mm,NE,ex,ey,ez,pr,qr, // input
                    pot,smooth, // potential
                    tol,pcoeff,theta, // FMM-expansion
                    &Ndirect,&Nlevels,eta // FMM-mesh
                    );
  }
  
  starttime = MPI_Wtime();

  // write potentials or mesh to file
  if (out == -1) {
#ifndef NOMEX
    WriteMesh(rank,out,output_file,N,NE,Nlevels,nprocs,m,c,h,s,t);
#endif
  }
  else {
    WritePotential(rank,out,output_file,N,NE,Nlevels,nprocs,pr,qr);
  }
  delete [] output_file;

  if (out == 0) {
    free(pr);
  }
  else if (out == 1) {
    free(qr);
  }
  else if (out == 2) {
    free(pr);
    free(qr);
  }
  
  // timing for write and free arrays
  if (!rank)
  {
    fp.open("perf.log", ofstream::app);
  
    if(!fp.is_open()) {
      cout << "perf.log is unavailable" << endl;
    }
    fp << nprocs << " 4 " << N << " " << tol << " " << Nlevels << " " << MPI_Wtime() - starttime << endl;
    fp.close();
  }

  StopTime;
  PrintTime("rank %d, Total: %f\n");
  MPI_Finalized(&already_finalised);
  if (!already_finalised) {
    MPI_Finalize();
  }
  
  return 0;
}
#endif
/*------------------------------------------------------------------------*/
