/* fmmshift.c */

/* S. Engblom 2010-02-18
   J. Bull 2019-05-02 - added 3D shifts */

#include "fmmshift.h"

// deallocate persistent data
void shift_atExit(void);

// radial distance between 2 points (duplicate of fmm.c function)
double sdistance(const double dx,const double dy,const double dz);

// workspace arrays
static dcmplx *Yi, *Yj;

static double *Yd;
static double *Wigner_d;
static dcmplx *coeff_beta;
static dcmplx *rotcoeff_shift;
static dcmplx *rotcoeff;
static dcmplx *rotcoeff_shift_j;
static dcmplx *rotcoeff_j;


// scalars defining shift-type and sizes
static int pshift,potshift,maxm2lshift;
static double rminshift;

static inline double oddoreven(int n) {
  return (((n) & 1) == 1) ? -1 : 1;
}

static inline double ipow2n(int n) {
  return (n >= 0) ? 1 : oddoreven(n);
}

/*------------------------------------------------------------------------*/
// Cartesian to spherical coordinate transformation given coordinates of a point and a reference point (origin)
void cart2spher(const double dx, const double dy, const double dz,double *sph)
{
  sph[0] = sdistance(dx,dy,dz);
  sph[1] = sph[0] == 0 ? 0 : acos((dz)/sph[0]);
  sph[2] = atan2(dy,dx);
}
/*------------------------------------------------------------------------*/
// distance between 2 points (duplicate of fmm.c function)
double sdistance(const double dx,const double dy,const double dz)
{
  return sqrt(dx*dx + dy*dy + dz*dz);
}
/*------------------------------------------------------------------------*/
//! Evaluate harmonic function \f$ r^n Y_{n}^{m} \f$
void evalMultipole(double rho, double alpha, double beta, dcmplx **Y) {

  const double x = cos(alpha);
  const double y = sin(alpha);
  const dcmplx ei = cexp(I * beta);
  double pn = 1;                                      // Initialize Legendre polynomial Pn
  double rhom = 1;                                    // Initialize rho^m
  double fact = 1;                                    // Initialize 2 * m + 1
  dcmplx expim = 1.0;                                 // Initialize exp(i * m * beta)

  // Loop over m in Y
  for (int m=0; m<pshift; m++) {
    double p = pn;                                    // Associated Legendre polynomial Pnm
    double p1 = p;                                    // Pnm-1
    int npn = m * m + 2 * m;                          // Index of Ynm for m > 0
    int nmn = m * m;                                  // Index of Ynm for m < 0
    (*Y)[npn] = rhom * p * expim;                     // rho^m * Ynm for m > 0
    (*Y)[nmn] = conj((*Y)[npn]);                      // Use conjugate relation for m < 0
    p = x * (2 * m + 1) * p1;                         // Update Pnm using recurrence relation
    rhom *= rho;                                      // Update rho^m
    double rhon = rhom;                               // Initialize rho^n
    
    // Loop over n in Ynm
    for (int n=m+1; n<pshift; n++) {
      int npm = n * n + n + m;                        // Index of Ynm for m > 0
      int nmm = n * n + n - m;                        // Index of Ynm for m < 0
      double p2 = p1;                                 // Pnm-2
      p1 = p;                                         // Pnm-1
      rhon /= -(n + m);                               // Update rho^n
      (*Y)[npm] = rhon * p * expim;                   // rho^n * Ynm
      (*Y)[nmm] = conj((*Y)[npm]);                    // Use conjugate relation for m < 0
      p = (x * (2*n+1) * p1 - (n+m) * p2) / (n-m+1);  // Update Pnm using recurrence relation
      rhon *= rho;                                    // Update rho^n
    }
    
    rhom /= -(2 * m + 2) * (2 * m + 1);               // Update rho^m
    pn = -pn * fact * y;                              // Update Legendre polynomial Pn
    fact += 2;                                        // Update 2 * m + 1
    expim *= ei;                                      // Update exp(i * m * beta)
  }
}
/*------------------------------------------------------------------------*/
//! Evaluate harmonic function \f$ r^n Y_{n}^{m} \f$
void evalMultipoleZ(double rho, double *Y) {
  //evalMultipole assuming that alpha = 0 and beta = 0, used for rotated shifts
  
  Y[0] = 1;
  double rhon = 1;
  for (int n=1; n<pshift; n++) {
    rhon *= -rho/n;
    Y[n] = rhon;
  }
}
/*------------------------------------------------------------------------*/
//! Evaluate singular harmonics \f$ r^{-n-1} Y_n^m \f$
void evalLocal(double rho, double alpha, double beta, dcmplx **Y) {
  const double x = cos(alpha);                                        // x = cos(alpha)
  const double y = sin(alpha);                                        // y = sin(alpha)
  double fact = 1;                                              // Initialize 2 * m + 1
  double pn = 1;                                                // Initialize Legendre polynomial Pn
  const double invR = -1.0 / rho;                                     // - 1 / rho
  double rhom = -invR;                                          // Initialize rho^(-m-1)
  const dcmplx ei = cexp(I * beta);                                   // exp(i * beta)
  dcmplx eim = 1.0;                                             // Initialize exp(i * m * beta)
  for (int m=0; m<pshift; m++) {                                // Loop over m in Y
    double p = pn;                                              //  Associated Legendre polynomial Pnm
    int npn = m * m + 2 * m;                                    //  Index of Y for m > 0
    int nmn = m * m;                                            //  Index of Y for m < 0
    (*Y)[npn] = rhom * p * eim;                               //  rho^(-m-1) * Y for m > 0
    (*Y)[nmn] = conj((*Y)[npn]);                            //  Use conjugate relation for m < 0
    double p1 = p;                                              //  Pnm-1
    p = x * (2 * m + 1) * p1;                                   //  Pnm using recurrence relation
    rhom *= invR;                                               //  rho^(-m-1)
    double rhon = rhom;                                         //  rho^(-n-1)
    for (int n=m+1; n<pshift; n++) {                            //  Loop over n in Y
      int npm = n * n + n + m;                                  //   Index of Y for m > 0
      int nmm = n * n + n - m;                                  //   Index of Y for m < 0
      (*Y)[npm] = rhon * p * eim;                             //   rho^n * Y for m > 0
      (*Y)[nmm] = conj((*Y)[npm]);                          //   Use conjugate relation for m < 0
      double p2 = p1;                                           //   Pnm-2
      p1 = p;                                                   //   Pnm-1
      p = (x * (2 * n + 1) * p1 - (n + m) * p2) / (n - m + 1);  //   Pnm using recurrence relation
      rhon *= invR * (n - m + 1);                               //   rho^(-n-1)
    }                                                           //  End loop over n in Y
    pn = -pn * fact * y;                                        //  Pn
    fact += 2;                                                  //  2 * m + 1
    eim *= ei;                                                  //  Update exp(i * m * beta)
  }                                                             // End loop over m in Y
}
/*------------------------------------------------------------------------*/
//! Evaluate singular harmonics \f$ r^{-n-1} Y_n^m \f$
void evalLocalZ(double rho, double *Y)
//evalLocal assuming that alpha = 0 and beta = 0, used for rotated shifts
{
  const double invR = -1.0 / rho;
  double rhon = -invR;
  Y[0] = rhon;

  for (int n=1; n<pshift; n++) {
    rhon *= invR * n;
    Y[n] = rhon;
  }
}
/*------------------------------------------------------------------------*/
void shift_alloc(int p,int pot,int maxm2l)
/* Allocates and initializes static data given a polynomial order p,
   type of potential pot and maximum number of cluster to cluster
   interactions maxm2l. */
{
  if (pshift != p) {
    free(Yi);
    free(Yj);
    Yi = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    Yj = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    Yd = (double *) malloc(p*p*sizeof(double));
    Wigner_d = (double *) malloc((p+1)*(p+2)*(4*p+3)/6*sizeof(double)); //size is sum[i = 1-->p](i*(2i-1)
    coeff_beta = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    rotcoeff_shift = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    rotcoeff = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    rotcoeff_shift_j = (dcmplx *) malloc(p*p*sizeof(dcmplx));
    rotcoeff_j = (dcmplx *) malloc(p*p*sizeof(dcmplx));

    pshift = p;
    potshift = pot;
  }
}
/*------------------------------------------------------------------------*/
void shift_atExit(void)
/* Deallocates persistent data. */
{
  free(Yi);
  free(Yj);

  free(Yd);
  free(Wigner_d);
  free(coeff_beta);
  free(rotcoeff_shift);
  free(rotcoeff);
  free(rotcoeff_shift_j);
  free(rotcoeff_j);

  pshift = 0;
}
/*------------------------------------------------------------------------*/
void shift_m2m(struct mpexp *restrict this1, const struct mpexp *restrict that1)
/* Multipole to multipole shift from 8 children (that) to 1 parent (this); this1 +=
 shift(that1[0...p*p]) {child 0} + shift(that1[p*p+1...2*p*p]) {child 1} +
 ...  and so on. */
{
  // coords and coeffs of parent
  const double x0 = this1->x0;
  const double y0 = this1->y0;
  const double z0 = this1->z0;
  dcmplx *thism = this1->coeff1;
  
  // loop over 8 children
  for (int i = 0; i < 8; i++) {

    // ptr to child
    const struct mpexp *that1i = &that1[i];

    // coords and coeffs of child
    const double xj = that1i->x0;
    const double yj = that1i->y0;
    const double zj = that1i->z0;
    const dcmplx *thatm = that1i->coeff1;
    
    double sph[3];
    // spherical coords of child w.r.t. parent
    cart2spher(x0-xj,y0-yj,z0-zj,sph);
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];

    // compute spherical harmonic coeffs
    evalMultipole(rho, alpha, beta, &Yi);
    
    for (int j=0; j<pshift; j++) {
      
      for (int k=0; k<=j; k++) {
        int jks = j * (j + 1) / 2 + k;
        dcmplx M = 0;
        
        for (int n=0; n<=j; n++) {
          int mmax = ( (-n < -j+k+n) ? -j+k+n : -n );
          int mmin = ( (k-1 < n) ? k-1 : n );
          
          int jnkms = (j - n) * (j - n + 1) / 2 + k - mmax;
          int nm    = n * n + n - mmax;
          dcmplx Mt = 0;
          for (int m=mmax; m<=mmin; m++,jnkms--,nm--) {
            //int jnkms = (j - n) * (j - n + 1) / 2 + k - m;
            //int nm    = n * n + n - m;
            Mt += thatm[jnkms] * Yi[nm] * ipow2n(m);
          }
          M += Mt*oddoreven(n);
          int nmin = ( (n < j+k-n) ? n : j+k-n );
          
          jnkms = (j - n) * (j - n + 1) / 2;
          nm    = n * n + n - k;
          for (int m=k; m<=nmin; m++,jnkms++,nm--) {
            //int jnkms2 = (j - n) * (j - n + 1) / 2 - k + m;
            //int nm2    = n * n + n - m;
            M += conj(thatm[jnkms]) * Yi[nm] * oddoreven(k+n+m);
          }
        }
        thism[jks] += M;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_m2mZ(struct mpexp *restrict this1, const struct mpexp *restrict that1)
/* p3 version of multipole to multipole shift, which is based on rotating the shift to perform it along the z-axis
 Multipole to multipole shift from 8 children (that) to 1 parent (this); this1 +=
 shift(that1[0...p*p]) {child 0} + shift(that1[p*p+1...2*p*p]) {child 1} +
 ...  and so on. */
{
  // coords and coeffs of parent
  const double x0 = this1->x0;
  const double y0 = this1->y0;
  const double z0 = this1->z0;
  dcmplx *thism = this1->coeff1;
  
  // loop over 8 children
  for (int i = 0; i < 8; i++) {

    // ptr to child
    const struct mpexp *that1i = &that1[i];

    // coords and coeffs of child
    const double xj = that1i->x0;
    const double yj = that1i->y0;
    const double zj = that1i->z0;
    const dcmplx *thatm = that1i->coeff1;
    
    double sph[3];
    // spherical coords of child w.r.t. parent
    cart2spher(x0 - xj, y0 - yj, z0 - zj, sph);
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];

    // generate the scaled version of the Wigner d matrix, which is used to rotate the coefficients
    buildWigner_d_scaled(fabs(alpha), Wigner_d);

    //for rotation around the z-axis
    dcmplx beta_rot = cexp(I * beta);                                   // exp(i * beta)

    //rotate the coefficients
    rotate_Multipole_scaled(rotcoeff, thatm, Wigner_d, beta_rot);

    // compute spherical harmonic coeffs
    evalMultipoleZ(rho, Yd);

    //perform the p3 shift
    for (int j = 0; j < pshift; j++) {
      int jks = j * (j + 1) / 2;
      double Mkre = 0;
      double Mkim = 0;

      //the special case with k = 0, where the complex conjugate should be used
      for (int n = 0; n <= j; n++) {
        const int jnkms = (j - n) * (j - n + 1) / 2;
        Mkre += creal(rotcoeff[jnkms]) * Yd[n] * oddoreven(n);
        Mkim -= cimag(rotcoeff[jnkms]) * Yd[n] * oddoreven(n);
      }
      rotcoeff_shift[jks] = Mkre + I*Mkim;//Mk;
      jks++;
      //for the rest of the coefficients, no complex conjugates are required
      for (int k = 1; k <= j; k++, jks++) {
        double Mre = 0;
        double Mim = 0;

        for (int n = 0; n <= j - k; n++) {
          const int jnkms = (j - n) * (j - n + 1) / 2 + k;
          Mre += creal(rotcoeff[jnkms]) * Yd[n] * oddoreven(n);
          Mim += cimag(rotcoeff[jnkms]) * Yd[n] * oddoreven(n);
        }
        rotcoeff_shift[jks] = Mre + I*Mim;
      }
    }
    //inverse rotate and store results
    inv_rotate_Multipole_scaled(thism, rotcoeff_shift, Wigner_d, beta_rot);
  }
}
/*------------------------------------------------------------------------*/
void shift_m2msimple(double *restrict sph,dcmplx *restrict thism,const dcmplx *restrict thatm)
/* Multipole to multipole shift from 1 child (that) to 1 parent (this);
this1 += shift(that1). */
{
  // spherical coords of this w.r.t. that
  const double rho = sph[0];
  const double alpha = sph[1];
  const double beta = sph[2];
  
  // compute spherical harmonic coeffs
  evalMultipole(rho, alpha, beta, &Yi);
  
  for (int j=0; j<pshift; j++) {

    for (int k=0; k<=j; k++) {
      int jks = j * (j + 1) / 2 + k;
      dcmplx M = 0;

      for (int n=0,s=0; n<=j; n++) {
        int mmax = ( (-n < -j+k+n) ? -j+k+n : -n );
        int mmin = ( (k-1 < n) ? k-1 : n );

        for (int m=mmax; m<=mmin; m++,s++) {
          int jnkms = (j - n) * (j - n + 1) / 2 + k - m;
          int nm    = n * n + n - m;
          M += thatm[jnkms] * Yi[nm] * ipow2n(m) * oddoreven(n);
        }
        int nmin = ( (n < j+k-n) ? n : j+k-n );
        
        for (int m=k; m<=nmin; m++,s++) {
          int jnkms = (j - n) * (j - n + 1) / 2 - k + m;
          int nm    = n * n + n - m;
          M += conj(thatm[jnkms]) * Yi[nm] * oddoreven(k+n+m);
        }
      }
      thism[jks] += M;
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_m2l(struct mpexp *restrict this1, struct mpexp *restrict that1,
               const int *restrict ix,int begin,int end,int symm)
/* Symmetric multipole to local shift; thisl +=
   shift(thatm[ix[i]]) followed by thatl[ix[i]] += shift(thism) for
   index i in [begin,end). Center coordinates of the boxes are given
   in x0,y0,z0 and xj,yj,zj (in which each entry is separated by siz bytes).

   A somewhat singular feature is that a negative index in ix,
   pointing to a box at a previous level, is detected and implies that
   the 7 following indices are ignored. This ensures that connections
   between levels are properly handled and that the connectivity
   indices are convenient to assemble. */
{

  // coords and coeffs of ibox
  const double x0 = this1->x0;
  const double y0 = this1->y0;
  const double z0 = this1->z0;
  const dcmplx *thism = this1->coeff1;
  dcmplx *thisl = this1->coeff2;
  
  // loop over connection list for ibox
  for (int i = begin; i < end; i++) {
    int offset = ix[i];

    // ptr to jbox
    struct mpexp *that1i = &that1[offset];

    // coords and coeffs of jbox
    const double xj = that1i->x0;
    const double yj = that1i->y0;
    const double zj = that1i->z0;
    const dcmplx *thatmi = that1i->coeff1;
    dcmplx *thatli;
    if(symm) {
      thatli = that1i->coeff2;
    }
    
    double sph[3];
    cart2spher(x0-xj,y0-yj,z0-zj,sph);
//    printf("x0 %f, y0 %f, z0 %f, j %d, xj %f, yj %f, zj %f\n",x0,y0,z0,i,xj,yj,zj);
    // spherical coords of this w.r.t. that
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];
    
    evalLocal(rho, alpha, beta, &Yi);
    if(symm) {
      evalLocal(rho, alpha+M_PI, beta, &Yj);
    }
    
    for (int j=0; j<pshift; j++) {
      double Cnm = oddoreven(j);
      
      for (int k=0; k<=j; k++) {
        int jks = j * (j + 1) / 2 + k;
        dcmplx Li = 0, Lj = 0;
#ifdef FMM_INLINECOMPLEX
        double Lir = 0, Lim = 0, Ljr = 0, Ljm = 0;
#endif
        
        if(symm) {
          for (int n=0; n<pshift-j; n++) {
          
            int nms  = n * (n + 1) / 2 + n;
            int jnkm = (j + n) * (j + n) + j - k;
#ifdef FMM_INLINECOMPLEX
            double Litr = 0, Litm = 0, Ljtr = 0, Ljtm = 0;
            for (int m=-n; m<0; m++, nms--, jnkm++) {
              Litr += creal(thatmi[nms])*creal(Yi[jnkm]) + cimag(thatmi[nms])*cimag(Yi[jnkm]);
              Litm += creal(thatmi[nms])*cimag(Yi[jnkm]) - cimag(thatmi[nms])*creal(Yi[jnkm]);
              Ljtr += creal(thism[nms])*creal(Yj[jnkm]) + cimag(thism[nms])*cimag(Yj[jnkm]);
              Ljtm += creal(thism[nms])*cimag(Yj[jnkm]) - cimag(thism[nms])*creal(Yj[jnkm]);
              //Lit += conj(thatmi[nms]) * Yi[jnkm];
              //Ljt += conj(thism[nms]) * Yj[jnkm];
            }

            Lir += Litr*Cnm;
            Lim += Litm*Cnm;
            Ljr += Ljtr*Cnm;
            Ljm += Ljtm*Cnm;
            //Li += Lit*Cnm;
            //Lj += Ljt*Cnm;
          
            for (int m=0; m<=n; m++, nms++, jnkm++) {
              
              if((j&1)^(((k-m)*(k<m)+m)&1)) {
                Lir -= (creal(thatmi[nms])*creal(Yi[jnkm]) - cimag(thatmi[nms])*cimag(Yi[jnkm]));
                Lim -= (creal(thatmi[nms])*cimag(Yi[jnkm]) + cimag(thatmi[nms])*creal(Yi[jnkm]));
                Ljr -= (creal(thism[nms])*creal(Yj[jnkm]) - cimag(thism[nms])*cimag(Yj[jnkm]));
                Ljm -= (creal(thism[nms])*cimag(Yj[jnkm]) + cimag(thism[nms])*creal(Yj[jnkm]));
              }
              else {
                Lir += (creal(thatmi[nms])*creal(Yi[jnkm]) - cimag(thatmi[nms])*cimag(Yi[jnkm]));
                Lim += (creal(thatmi[nms])*cimag(Yi[jnkm]) + cimag(thatmi[nms])*creal(Yi[jnkm]));
                Ljr += (creal(thism[nms])*creal(Yj[jnkm]) - cimag(thism[nms])*cimag(Yj[jnkm]));
                Ljm += (creal(thism[nms])*cimag(Yj[jnkm]) + cimag(thism[nms])*creal(Yj[jnkm]));
              }
            
              //double Cnm2 = Cnm * oddoreven((k-m)*(k<m)+m);
              //Lir += (creal(thatmi[nms])*creal(Yi[jnkm]) - cimag(thatmi[nms])*cimag(Yi[jnkm])) * Cnm2;
              //Lim += (creal(thatmi[nms])*cimag(Yi[jnkm]) + cimag(thatmi[nms])*creal(Yi[jnkm])) * Cnm2;
              //Ljr += (creal(thism[nms])*creal(Yj[jnkm]) - cimag(thism[nms])*cimag(Yj[jnkm])) * Cnm2;
              //Ljm += (creal(thism[nms])*cimag(Yj[jnkm]) + cimag(thism[nms])*creal(Yj[jnkm])) * Cnm2;
              //Li += thatmi[nms] * Yi[jnkm] * Cnm2;
              //Lj += thism[nms] * Yj[jnkm] * Cnm2;
            }
          }
          thisl[jks] += Lir+I*Lim;
          thatli[jks] += Ljr+I*Ljm;

#else
            dcmplx Lit = 0, Ljt = 0;
            for (int m=-n; m<0; m++, nms--, jnkm++) {
              //int nms  = n * (n + 1) / 2 - m;
              //int jnkm = (j + n) * (j + n) + j + n + m - k;
            
              Lit += conj(thatmi[nms]) * Yi[jnkm];
              Ljt += conj(thism[nms]) * Yj[jnkm];
            }

            Li += Lit*Cnm;
            Lj += Ljt*Cnm;
          
            for (int m=0; m<=n; m++, nms++, jnkm++) {
              //int nms  = n * (n + 1) / 2 + m;
              //int jnkm = (j + n) * (j + n) + j + n + m - k;
              double Cnm2 = Cnm * oddoreven((k-m)*(k<m)+m);
            
              Li += thatmi[nms] * Yi[jnkm] * Cnm2;
              Lj += thism[nms] * Yj[jnkm] * Cnm2;
            }
          }
          thisl[jks] += Li;
          thatli[jks] += Lj;
#endif
        }
        else {
          for (int n=0; n<pshift-j; n++) {
        
            int nms  = n * (n + 1) / 2 + n;
            int jnkm = (j + n) * (j + n) + j - k;
            dcmplx Lit = 0;
            for (int m=-n; m<0; m++, nms--, jnkm++) {
              Lit += conj(thatmi[nms]) * Yi[jnkm];
            }
            Li += Lit*Cnm;
          
            for (int m=0; m<=n; m++, nms++, jnkm++) {
              double Cnm2 = Cnm * oddoreven((k-m)*(k<m)+m);
          
              Li += thatmi[nms] * Cnm2 * Yi[jnkm];
            }
          }
          thisl[jks] += Li;
        }
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_m2lZ(struct mpexp *restrict this1, struct mpexp *restrict that1,
               const int *restrict ix,int begin,int end,int symm)
/* Symmetric multipole to local shift; thisl +=
   shift(thatm[ix[i]]) followed by thatl[ix[i]] += shift(thism) for
   index i in [begin,end). Center coordinates of the boxes are given
   in x0,y0,z0 and xj,yj,zj (in which each entry is separated by siz bytes).
   
   The rotated form is used. */
{

  // coords and coeffs of ibox
  const double x0 = this1->x0;
  const double y0 = this1->y0;
  const double z0 = this1->z0;
  const dcmplx *restrict thism = this1->coeff1;
  dcmplx *thisl = this1->coeff2;
  
  // loop over connection list for ibox
  for (int i = begin; i < end; i++) {

    // ptr to jbox
    struct mpexp *that1i = &that1[ix[i]];

    // coords and coeffs of jbox
    const double xj = that1i->x0;
    const double yj = that1i->y0;
    const double zj = that1i->z0;
    const dcmplx *restrict thatmi = that1i->coeff1;
    dcmplx *thatli;
    if(symm) {
      thatli = that1i->coeff2;
    }
    
    double sph[3];
    cart2spher(x0-xj,y0-yj,z0-zj,sph);
//    printf("x0 %f, y0 %f, z0 %f, j %d, xj %f, yj %f, zj %f\n",x0,y0,z0,i,xj,yj,zj);
    // spherical coords of this w.r.t. that
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];

    // generate the scaled version of the Wigner d matrix, which is used to rotate the coefficients
    buildWigner_d_scaled(fabs(alpha), Wigner_d);

    //for rotation around the z-axis
    const dcmplx beta_rot = cexp(I * beta);                                   // exp(i * beta)

    //rotate the coefficients
    rotate_Multipole_scaled(rotcoeff, thatmi, Wigner_d, beta_rot); //some speed improvements may be obtained if a special version of rotate_Multipole_scaled is created that takes two arguments
    if (symm) {
      rotate_Multipole_scaled(rotcoeff_j, thism, Wigner_d, beta_rot);
      memset(rotcoeff_shift_j, 0, pshift*pshift*sizeof(dcmplx));
    }
    memset(rotcoeff_shift, 0, pshift*pshift*sizeof(dcmplx)); //code is not populating all elements of the matrix, keep the rest to zero for safety

    // compute spherical harmonic coeffs
    evalLocalZ(rho, Yd);

    //perform the p3 shift
    for (int j=0; j<pshift; j++) {


      for (int k=0; k<=j; k++) {
        const double Cnm = oddoreven(j + k);
        const int jks = j * (j + 1) / 2 + k;

        double Lir = 0, Lim = 0, Ljr = 0, Ljm = 0;

        if(symm) {
          for (int n = k; n < pshift - j; n++) {

            const int nms  = n * (n + 1) / 2 + k;
            //int jnkm = (j + n) * (j + n) + j - k;

            const double tmp = Cnm * Yd[j + n];
            const double tmpj = oddoreven(j + n)*tmp; //I have not checked if this is faster than generating Ydj to remove oddoreven
            Lir += creal(rotcoeff[nms])*tmp;
            Lim += cimag(rotcoeff[nms])*tmp;

            Ljr += creal(rotcoeff_j[nms])*tmpj;
            Ljm += cimag(rotcoeff_j[nms])*tmpj;

          }
          rotcoeff_shift[jks] += Lir + I*Lim;
          rotcoeff_shift_j[jks] += Ljr + I*Ljm;

        }
        else {
          for (int n = k; n < pshift - j; n++) {
            const int nms  = n * (n + 1) / 2 + k;

            const double tmp = Cnm * Yd[j + n];
            Lir += creal(rotcoeff[nms])*tmp;
            Lim += cimag(rotcoeff[nms])*tmp;
          }
          rotcoeff_shift[jks] += Lir + I*Lim;
        }
      }
    }

    //inverse rotate and store results
    inv_rotate_Local_scaled_permuted(thisl, rotcoeff_shift, Wigner_d, beta_rot);
    if (symm)
      inv_rotate_Local_scaled_permuted(thatli, rotcoeff_shift_j, Wigner_d, beta_rot);
  }
}
/*------------------------------------------------------------------------*/
void shift_m2lZsimple(int rank,double *restrict sph,
               const dcmplx *restrict thism,dcmplx *restrict thisl,
               const dcmplx *restrict thatm,dcmplx *restrict thatl)
/* Symmetric multipole to local shift; thisl +=
   shift(thatm[ix[i]]) followed by thatl[ix[i]] += shift(thism) for
   index i in [begin,end). Center coordinates of the boxes are given
   in x0,y0,z0 and xj,yj,zj (in which each entry is separated by siz bytes).

   A somewhat singular feature is that a negative index in ix,
   pointing to a box at a previous level, is detected and implies that
   the 7 following indices are ignored. This ensures that connections
   between levels are properly handled and that the connectivity
   indices are convenient to assemble. */
{

  // spherical coords of this w.r.t. that
  const double rho = sph[0];
  const double alpha = sph[1];
  const double beta = sph[2];

  // if shift is one-way or two-way
  int symm=0;
  if(thism != NULL && thatl != NULL) symm = 1;
  
  // generate the scaled version of the Wigner d matrix, which is used to rotate the coefficients
  buildWigner_d_scaled(fabs(alpha), Wigner_d);
  
  //for rotation around the z-axis
  const dcmplx beta_rot = cexp(I * beta);                                   // exp(i * beta)
  
  //rotate the coefficients
  rotate_Multipole_scaled(rotcoeff, thatm, Wigner_d, beta_rot); //some speed improvements may be obtained if a special version of rotate_Multipole_scaled is created that takes two arguments
  if (symm) {
    rotate_Multipole_scaled(rotcoeff_j, thism, Wigner_d, beta_rot);
    memset(rotcoeff_shift_j, 0, pshift*pshift*sizeof(dcmplx));
  }
  memset(rotcoeff_shift, 0, pshift*pshift*sizeof(dcmplx)); //code is not populating all elements of the matrix, keep the rest to zero for safety
  
  // compute spherical harmonic coeffs
  evalLocalZ(rho, Yd);

  //perform the p3 shift
  for (int j=0; j<pshift; j++) {
    
    
    for (int k=0; k<=j; k++) {
      const double Cnm = oddoreven(j + k);
      const int jks = j * (j + 1) / 2 + k;
      
      double Lir = 0, Lim = 0, Ljr = 0, Ljm = 0;
      
      if(symm) {
        for (int n = k; n < pshift - j; n++) {
          
          const int nms  = n * (n + 1) / 2 + k;
          
          const double tmp = Cnm * Yd[j + n];
          const double tmpj = oddoreven(j + n)*tmp;
          Lir += creal(rotcoeff[nms])*tmp;
          Lim += cimag(rotcoeff[nms])*tmp;
          
          Ljr += creal(rotcoeff_j[nms])*tmpj;
          Ljm += cimag(rotcoeff_j[nms])*tmpj;
          
        }
        rotcoeff_shift[jks] += Lir + I*Lim;
        rotcoeff_shift_j[jks] += Ljr + I*Ljm;
        
      }
      else {
        for (int n = k; n < pshift - j; n++) {
          const int nms  = n * (n + 1) / 2 + k;
          
          const double tmp = Cnm * Yd[j + n];
          Lir += creal(rotcoeff[nms])*tmp;
          Lim += cimag(rotcoeff[nms])*tmp;
        }
        rotcoeff_shift[jks] += Lir + I*Lim;
      }
    }
  }
  
  //inverse rotate and store results
  inv_rotate_Local_scaled_permuted(thisl, rotcoeff_shift, Wigner_d, beta_rot);
  if (symm)
    inv_rotate_Local_scaled_permuted(thatl, rotcoeff_shift_j, Wigner_d, beta_rot);
}
/*------------------------------------------------------------------------*/
void shift_m2lsimple(int rank,double *restrict sph,
               const dcmplx *restrict thism,dcmplx *restrict thisl,
               const dcmplx *restrict thatm,dcmplx *restrict thatl)
{
  // spherical coords of this w.r.t. that
  const double rho = sph[0];
  const double alpha = sph[1];
  const double beta = sph[2];

  // if shift is one-way or two-way
  int symm=0;
  if(thism != NULL && thatl != NULL) symm = 1;
  
  evalLocal(rho, alpha, beta, &Yi);
  if(symm) {
    evalLocal(rho, alpha+M_PI, beta, &Yj);
  }
  
  for (int j=0; j<pshift; j++) {
    const double Cnm = oddoreven(j);
    
    for (int k=0; k<=j; k++) {
      const int jks = j * (j + 1) / 2 + k;
      dcmplx Li = 0, Lj = 0;
      
      for (int n=0; n<pshift-j; n++) {

        for (int m=-n; m<0; m++) {
          const int nms  = n * (n + 1) / 2 - m;
          const int jnkm = (j + n) * (j + n) + j + n + m - k;

          Li += conj(thatm[nms]) * Cnm * Yi[jnkm];
          if(symm) {
            Lj += conj(thism[nms]) * Cnm * Yj[jnkm];
          }
        }

        for (int m=0; m<=n; m++) {
          const int nms  = n * (n + 1) / 2 + m;
          const int jnkm = (j + n) * (j + n) + j + n + m - k;
          double Cnm2 = Cnm * oddoreven((k-m)*(k<m)+m);

          Li += thatm[nms] * Cnm2 * Yi[jnkm];
          if(symm) {
            Lj += thism[nms] * Cnm2 * Yj[jnkm];
          }
        }
      }
      thisl[jks] += Li;
      if(symm) {
        thatl[jks] += Lj;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_l2l(struct mpexp *restrict this1, const struct mpexp *restrict that1)
/* Local to local shift from 1 parent to 8 children;
   this1[0] += shift(that1) {child 0},
   this1[1] += shift(that1) {child 1},
   and so on. */
{
  // coords and coeffs of parent
  const double xj = that1->x0;
  const double yj = that1->y0;
  const double zj = that1->z0;
  const dcmplx *thatl = that1->coeff2;
  
  // loop over 8 children
  for (int i = 0; i < 8; i++) {

    // ptr to child
    struct mpexp *this1i = &this1[i];

    // coords and coeffs of jbox
    const double x0 = this1i->x0;
    const double y0 = this1i->y0;
    const double z0 = this1i->z0;
    dcmplx *thisl = this1i->coeff2;
    
    double sph[3];
    // spherical coords of child w.r.t. parent
    cart2spher(x0-xj,y0-yj,z0-zj,sph);
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];

    evalMultipole(rho, alpha, beta, &Yi);
    
    for (int j=0; j<pshift; j++) {
      
      for (int k=0; k<=j; k++) {
        int jks = j * (j + 1) / 2 + k;
        dcmplx L = 0;
        
        for (int n=j; n<pshift; n++) {
          
          for (int m=j+k-n; m<0; m++) {
            int jnkm = (n - j) * (n - j) + n - j + m - k;
            int nms  = n * (n + 1) / 2 - m;
            
            L += conj(thatl[nms]) * Yi[jnkm] * oddoreven(k);
          }
          
          for (int m=0; m<=n; m++) {
            if( n-j >= abs(m-k) ) {
              int jnkm = (n - j) * (n - j) + n - j + m - k;
              int nms  = n * (n + 1) / 2 + m;
              
              L += thatl[nms] * Yi[jnkm] * oddoreven((m-k)*(m<k));
            }
          }
        }
        thisl[jks] += L;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void shift_l2lZ(struct mpexp *this1, const struct mpexp *that1)
/* Local to local shift from 1 parent to 8 children;
   this1[0] += shift(that1) {child 0},
   this1[1] += shift(that1) {child 1},
   and so on. */
{
  // coords and coeffs of parent
  const double xj = that1->x0;
  const double yj = that1->y0;
  const double zj = that1->z0;
  const dcmplx *thatl = that1->coeff2;
  
  // loop over 8 children
  for (int i = 0; i < 8; i++) {

    // ptr to child
    struct mpexp *this1i = &this1[i];

    // coords and coeffs of jbox
    const double x0 = this1i->x0;
    const double y0 = this1i->y0;
    const double z0 = this1i->z0;
    dcmplx *thisl = this1i->coeff2;
    
    double sph[3];
    // spherical coords of child w.r.t. parent
    cart2spher(x0-xj,y0-yj,z0-zj,sph);
    const double rho = sph[0];
    const double alpha = sph[1];
    const double beta = sph[2];

    // generate the scaled version of the Wigner d matrix, which is used to rotate the coefficients
    // note that these coefficients are scaled for the multipole shift. It would be possible to create a version of
    // the Wigner matrix code that scales them directly for local shifts, which would improve memory access patterns.
    // It is questionable how much difference it would make though, as L2L and M2M runs basically at the same speed at the moment
    buildWigner_d_scaled(fabs(alpha), Wigner_d);

    //for rotation around the z-axis
    dcmplx beta_rot = cexp(I * beta);                                   // exp(i * beta)

    // compute spherical harmonic coeffs
    evalMultipoleZ(rho, Yd);

    //rotate the coefficients
    rotate_Local_scaled_permuted(rotcoeff, thatl, Wigner_d, beta_rot);

    //perform the p3 shift
    for (int j=0; j<pshift; j++) {

      for (int k=0; k<=j; k++) {
        int jks = j * (j + 1) / 2 + k;
        double Lre = 0;
        double Lim = 0;

        for (int n=j; n<pshift; n++) {
          int nms  = n * (n + 1) / 2 + k;
          Lre += creal(rotcoeff[nms]) * Yd[n-j];
          Lim += cimag(rotcoeff[nms]) * Yd[n-j];
        }
        rotcoeff_shift[jks] = Lre + I*Lim;
      }
    }

    //inverse rotate and store results
    inv_rotate_Local_scaled_permuted(thisl, rotcoeff_shift, Wigner_d, beta_rot);
  }
}
/*------------------------------------------------------------------------*/
void shift_l2lsimple(double *restrict sph,dcmplx *restrict thisl,const dcmplx *restrict thatl)
/* Local to local shift from 1 parent (that) to 1 child (this);
   this1 += shift(that1) */
{
  // spherical coords of this w.r.t. that
  const double rho = sph[0];
  const double alpha = sph[1];
  const double beta = sph[2];
  
  evalMultipole(rho, alpha, beta, &Yi);
  
  for (int j=0; j<pshift; j++) {

    for (int k=0; k<=j; k++) {
      int jks = j * (j + 1) / 2 + k;
      dcmplx L = 0;

      for (int n=j; n<pshift; n++) {

        for (int m=j+k-n; m<0; m++) {
          int jnkm = (n - j) * (n - j) + n - j + m - k;
          int nms  = n * (n + 1) / 2 - m;

          L += conj(thatl[nms]) * Yi[jnkm] * oddoreven(k);
        }

        for (int m=0; m<=n; m++) {
          if( n-j >= abs(m-k) ) {
            int jnkm = (n - j) * (n - j) + n - j + m - k;
            int nms  = n * (n + 1) / 2 + m;

            L += thatl[nms] * Yi[jnkm] * oddoreven((m-k)*(m<k));
          }
        }
      }
      thisl[jks] += L;
    }
  }
}
/*------------------------------------------------------------------------*/
void rotate_Multipole_scaled(dcmplx *thism, const dcmplx *restrict thatm, const double *restrict Wigner, const dcmplx beta_rot)
  /*Rotates multipole coefficients using a scaled version of the Wigner d matrix to avoid having to 
  calculate any extra factors (especially involving sqrt) */
{
  //perform the rotation around the z - axis first to get beta = 0
  dcmplx rot = beta_rot;
  for (int n = 0; n < pshift; n++) {
    int nms = n * (n + 1) / 2;
    coeff_beta[nms] = thatm[nms];
  }
  for (int m = 1; m < pshift; m++) {
    for (int n = m; n < pshift; n++) {
      int nms = n * (n + 1) / 2 + m;
      coeff_beta[nms] = rot*thatm[nms];
    }
    rot *= beta_rot;
  }

  //start the rotation witht the Wigner matrices
  int baseind = 0;
  for (int n = 0; n < pshift; n++) {
    baseind += n*(2*n - 1);
    int baseind2 = baseind + n;
    for (int m = 0; m <= n; m++) {
      int nmps = n * (n + 1) / 2;
      int nms = n * (n + 1) / 2 + m;

      double Mre = Wigner[baseind2] * creal(coeff_beta[nmps]);
      double Mim = Wigner[baseind2] * cimag(coeff_beta[nmps]);

      //note that only mp > 0 are stored, while the shift also requires the negative values, hence the two terms in the shift, 
      //where the conjugate has to be used for mp < 0
      for (int  mp = 1; mp <= n; mp++) {
        Mre += (Wigner[baseind2 - mp] + Wigner[baseind2 + mp])*creal(coeff_beta[nmps + mp]);
        Mim += (-Wigner[baseind2 - mp] + Wigner[baseind2 + mp])*cimag(coeff_beta[nmps + mp]);
      }
      thism[nms] = Mre + I*Mim;
      baseind2 += 2*n + 1;
    }
  }
}
/*------------------------------------------------------------------------*/
void inv_rotate_Multipole_scaled(dcmplx *thism, const dcmplx *restrict thatm, const double *restrict Wigner, const dcmplx beta_rot)
  /*Reverses the rotation of the multipole coefficients that was made by rotate_Multipole_scaled. 
  This function uses scaled Wigner d matrix
  The function is based on the symmetry that Wigner_d(-alpha) = oddoreven(m + mp)*Wigner_d(alpha)*/
{
  
  int baseind = 0;
  for (int n = 0; n < pshift; n++) {
    baseind += n*(2*n - 1);
    int baseind2 = baseind + n;

    for (int m = 0; m <= n; m++) {
      int nmps = n * (n + 1) / 2;
      int nms = n * (n + 1) / 2 + m;
      double tmp = oddoreven(m)*Wigner[baseind2];

      double Mre = tmp * creal(thatm[nmps]);
      double Mim = tmp * cimag(thatm[nmps]);

      //handle both mp > 0 and mp < 0 at the same time, as only half the elements are stored
      for (int  mp = 1; mp <= n; mp++) {
        tmp = oddoreven(m + mp); //due to the negative angle
        Mre += tmp*(Wigner[baseind2 - mp] + Wigner[baseind2 + mp])*creal(thatm[nmps + mp]);
        Mim += tmp*(-Wigner[baseind2 - mp] + Wigner[baseind2 + mp])*cimag(thatm[nmps + mp]);
      }
      coeff_beta[nms] = Mre + I*Mim;
      baseind2 += 2*n + 1;

    }
  }

  //perform the rotation around the z-axis last for the inverse
  dcmplx cbeta_rot = conj(beta_rot);
  dcmplx rot = cbeta_rot;
  for (int n = 0; n < pshift; n++) {
    int nms = n * (n + 1) / 2;
    thism[nms] += coeff_beta[nms];
  }
  for (int m = 1; m < pshift; m++) {
    for (int n = m; n < pshift; n++) {
      int nms = n * (n + 1) / 2 + m;
      thism[nms] += rot*coeff_beta[nms];
    }
    rot *= cbeta_rot;
  }
}
/*------------------------------------------------------------------------*/
void rotate_Local_scaled_permuted(dcmplx *thism, const dcmplx *restrict thatm, const double *restrict Wigner, const dcmplx beta_rot)
  /* rotates local coefficients using the scaled wigner_d matrix, where the matrix is scaled for the multipole rotation
  This version uses that the coefficients scaled for the local coefficients can be obtained by switching the indices
  m <--> mp and multiplying by oddoreven(m + mp)*/
{
  //perform the rotation around the z - axis first to get beta = 0
  dcmplx rot = beta_rot;
  for (int n = 0; n < pshift; n++) {
    int nms = n * (n + 1) / 2;
    coeff_beta[nms] = thatm[nms];
  }
  for (int m = 1; m < pshift; m++) {
    for (int n = m; n < pshift; n++) {
      int nms = n * (n + 1) / 2 + m;
      coeff_beta[nms] = rot*thatm[nms];
    }
    rot *= beta_rot;
  }

  //start the rotation witht the Wigner matrices
  int baseind = 0;
  for (int n = 0; n < pshift; n++) {
    baseind += n*(2*n - 1);
    int baseind2 = baseind + n;
    int basefact = 2*n + 1;

    for (int m = 0; m <= n; m++) {
      int nmps = n * (n + 1) / 2;
      int nms = n * (n + 1) / 2 + m;

      double tmp = Wigner[baseind2 + m]*oddoreven(m);

      //mp = 0
      double Mre = tmp * creal(coeff_beta[nmps]);
      double Mim = tmp * cimag(coeff_beta[nmps]);

      //somewhat more complicated indexing at the gain of not having to calculate the scaling coefficient
      for (int  mp = 1; mp <= n; mp++) {
        double tmp = oddoreven(m + mp);
        Mre += tmp*(Wigner[baseind2 + basefact*mp - m] + Wigner[baseind2 + basefact*mp + m])*creal(coeff_beta[nmps + mp]);
        Mim += tmp*(-Wigner[baseind2 + basefact*mp - m] + Wigner[baseind2 + basefact*mp + m])*cimag(coeff_beta[nmps + mp]);
      }
      thism[nms] = Mre + I*Mim;
    }
  }
}
/*------------------------------------------------------------------------*/
void inv_rotate_Local_scaled_permuted(dcmplx *thism, const dcmplx *restrict thatm, const double *restrict Wigner, const dcmplx beta_rot)
  /*Reverses the rotation of the local coefficients that was made by rotate_Local_scaled. 
  This function uses the scaled wigner d matrix, where the scaling is used for the multipole coefficients
  The function is based on the symmetry that Wigner_d(-alpha) = oddoreven(m + mp)*Wigner_d(alpha) 
  and that the coefficients scaled for the local coefficients can be obtained by switching the indices
  m <--> mp and multiplying by oddoreven(m + mp)*/
{
  int baseind = 0;
  for (int n = 0; n < pshift; n++) {
    baseind += n*(2*n - 1);
    int baseind2 = baseind + n;
    int basefact = 2*n + 1;
    for (int m = 0; m <= n; m++) {
      int nmps = n * (n + 1) / 2;
      int nms = n * (n + 1) / 2 + m;

      double tmp = Wigner[baseind2 + m];

      double Mre = tmp * creal(thatm[nmps]);
      double Mim = tmp * cimag(thatm[nmps]);

      //note that the factor oddoreven(m + mp) appears twice and hence cancels
      for (int  mp = 1; mp <= n; mp++) {
        Mre += (Wigner[baseind2 + basefact*mp - m] + Wigner[baseind2 + basefact*mp + m])*creal(thatm[nmps + mp]);
        Mim += (-Wigner[baseind2 + basefact*mp - m] + Wigner[baseind2 + basefact*mp + m])*cimag(thatm[nmps + mp]);
      }
      coeff_beta[nms] = Mre + I*Mim;

    }
  }

  //perform the rotation around the z-axis last for the inverse
  dcmplx cbeta_rot = conj(beta_rot);
  dcmplx rot = cbeta_rot;
  for (int n = 0; n < pshift; n++) {
    int nms = n * (n + 1) / 2;
    thism[nms] += coeff_beta[nms];
  }
  for (int m = 1; m < pshift; m++) {
    for (int n = m; n < pshift; n++) {
      int nms = n * (n + 1) / 2 + m;
      thism[nms] += rot*coeff_beta[nms];
    }
    rot *= cbeta_rot;
  }
}
/*------------------------------------------------------------------------*/
void buildWigner_d_scaled(double alpha, double *Wigner_d)
/* Generates the Wigner H matrix used to rotate spherical harmonics
   results are in Wigner_d(n,m,mp).
   the indices are ind = n*(n + 1)*(4*n - 1)/6 + (2*n + 1)*m + n + mp
   where n=[0,p], m=[0,n], mp=[-n,n]
   Comments are given as Wigner_d_{n}^{m',m}
   
   The algorithm follows 
   Nail A. Gumerov, Ramani Duraiswami,
   "Recursive computation of spherical harmonic rotation coefficients of large degree"
   http://arxiv.org/abs/1403.7698
   See  page 24 for the step by step description
   Note that the implemenation below calculates the coefficients for all levels,
   while the description in the paper is for one individual level.

   In the scaled version, the recursion scheme has been modified to include the scaling factor
   which gives a version that does not require any sqrt calculations during the recursions.
   Note that the stability properties have changed with the modified recursion, and hence, for large
   number of coefficients, additional testing is required.
   */
{
  //Step 1 - omitted as it is automatically obtained from step 2
  double x = cos(alpha);
  double y = sin(alpha);
  
  //step 2 - Calculate H_{n}^{0,m}
  double fact = 1;                                              // Initialize 2 * m + 1
  double pn = 1;                                                // Initialize Legendre polynomial Pn
  double rhom = 1;                                              // Initialize rho, which here only contains the factor
  //the mirror value is added to the algorithm due to the scaling
  //as the constant in front of the symmetry points in the matrix
  //have changed
  double rhom_mirror = 1;

  for (int m=0; m<=pshift; m++) {
    double p = pn;
    int mmn = m*(m + 1)*(4*m + 11)/6;                           //  Wigner_d_{n}^{0, n}
    
    Wigner_d[mmn] = rhom * p;
    if (m > 0) {
      int mpmn = m*(m + 1)*(4*m - 1)/6 + 2*m;                   //  Wigner_d_{n}^{n, 0}
      int nmpmn = m*(m + 1)*(4*m - 1)/6;                        //  Wigner_d_{n}^{-n, 0}
      Wigner_d[mpmn] = oddoreven(m) * rhom_mirror * p;
      Wigner_d[nmpmn] = Wigner_d[mpmn];
    }
    
    double p1 = p;                                              //  Pnm-1
    p = x * (2 * m + 1) * p1;                                   //  Pnm using recurrence relation
    double rhon = rhom;
    double rhon_mirror = rhom_mirror;
    for (int n = m + 1; n<=pshift; n++) {
      int npm = n*(n + 1)*(4*n - 1)/6 + (2*n + 1)*m + n;        //   Wigner_d_[n}^{0, m}
      rhon *= (double)n/(n + m);                                //   Update factorial
      rhon_mirror *= (double)(n - m)/n;
      Wigner_d[npm] = rhon * p;
      if (m > 0) {                                              //   symmetry points
        int npmp = n*(n + 1)*(4*n - 1)/6 + n + m;               //   Wigner_d_[n}^{m, 0}
        int npnmp = n*(n + 1)*(4*n - 1)/6 + n - m;              //   Wigner_d_[n}^{-m, 0}
        Wigner_d[npmp] = oddoreven(m) * rhon_mirror * p;
        Wigner_d[npnmp] = Wigner_d[npmp];
      }
      
      double p2 = p1;                                           //   Pnm-2
      p1 = p;                                                   //   Pnm-1
      p = (x * (2 * n + 1) * p1 - (n + m) * p2) / (n - m + 1);  //   Pnm using recurrence relation
    }                                                           //  End loop over n in Y
    rhom *= (double)(m + 1)/((2 * m + 2) * (2 * m + 1));        //  Update factorial
    rhom_mirror /= m + 1;
    pn = pn * fact * y;                                         //  Pn
    fact += 2;                                                  //  2 * m + 1
  }                                                             // End loop over m in Y

  //the remaining steps can be handled one level at a time, level 0 is already completed
  int baseind = 1;                                              // store n*(n + 1)*(4*n - 1)/6
  for (int n = 1; n < pshift; n++ )  {
    //step 3 - Calculate H_{n}^{1,m}
    int nextind = baseind + (n + 1)*(2*n + 1);                  // the start index of the next level


    //step 3
    double invn = 1.0/n;
    double factorm = 1;                                         // scaling factor due to the scaled Wigner matrix calculation
    for (int m = 1; m <= n; m++) {
      //use the mirror symmetry to simplify the expression

      //note that this first recursion uses the symmetry points as input
      //as the recursion equation is simpler with these
      int H1_m_n_mirror = baseind + 3*n + 1 + m;                // Wigner_d_{n}^{m, 1}
      int H1_m_n = baseind + (2*n + 1)*m + n + 1;               // Wigner_d_{n}^{1, m}
      
      int H0_m_np1 = nextind + n + 1 + m;                       // Wigner_d_{n + 1}^{m, 0}
      int H0_mp1_np1 = nextind + n + 2 + m;                     // Wigner_d_{n + 1}^{m + 1, 0}
      int H0_mm1_np1 = nextind + n + m;                         // Wigner_d_{n + 1}^{m - 1, 0}

      //the recursion
      Wigner_d[H1_m_n_mirror] = (0.5*(1 - x)*Wigner_d[H0_mp1_np1] + 0.5*(1 + x)*Wigner_d[H0_mm1_np1] - y*Wigner_d[H0_m_np1]);
      if (m > 1) {                                              // symmetry points
        Wigner_d[H1_m_n] = -oddoreven(m)*factorm*Wigner_d[H1_m_n_mirror];
      }

      //first iteration of step 5 with m'=0
      int H_mpm1_m = baseind + (2*n + 1)*m + n - 1;             // Wigner_d_{n}^{-1, m}
      int H_mp_mm1 = baseind + (2*n + 1)*(m - 1) + n;           // Wigner_d_{n}^{0, m - 1}
      int H_mp_mp1 = baseind + (2*n + 1)*(m + 1) + n;           // Wigner_d_{n}^{0, m + 1}
      
      //the recursion
      if (n == m) // necessary as last term should be zero, which is not fulfilled in the current case due to division by zero in the derivation
        Wigner_d[H_mpm1_m] = -Wigner_d[H1_m_n] + invn*Wigner_d[H_mp_mm1];
      else
        Wigner_d[H_mpm1_m] = invn*(-n*Wigner_d[H1_m_n] + (n - m + 1)*Wigner_d[H_mp_mm1] - (n + m + 1)*Wigner_d[H_mp_mp1]);
      if (m > 1) {                                              // symmetry points
        int H_mpm1_m_mirror = baseind + 3*n + 1 - m;            // Wigner_d_{n}^{-m, 1}
        Wigner_d[H_mpm1_m_mirror] = -oddoreven(m)/factorm*Wigner_d[H_mpm1_m];
      }
      factorm *= (double)(n - m)/(n + m + 1);                   //update scaling factor
    }

    //step 4 - calculate Wigner_d_{n}^{m'+1,m}
    for (int mp = 1; mp < n; mp++) {
      double inv_d_mp_n = 1.0/(n - mp);
      double factorm = 1;
      for (int m = mp + 1; m < n; m ++) {
        
        int H_mpp1_m = baseind + (2*n + 1)*m + n + mp + 1;      // Wigner_d_{n}^{mp + 1, m}
        
        int H_mpm1_m = baseind + (2*n + 1)*m + n + mp - 1;      // Wigner_d_{n}^{mp - 1, m}
        int H_mp_mm1 = baseind + (2*n + 1)*(m - 1) + n + mp;    // Wigner_d_{n}^{mp, m - 1}
        int H_mp_mp1 = baseind + (2*n + 1)*(m + 1) + n + mp;    // Wigner_d_{n}^{mp, m + 1}
        //the recursion
        Wigner_d[H_mpp1_m] = inv_d_mp_n*((n + mp)*Wigner_d[H_mpm1_m] + (n - m + 1)*Wigner_d[H_mp_mm1] - (n + m + 1)*Wigner_d[H_mp_mp1]);

        //step 5
        int H_nmpp1_m = baseind + (2*n + 1)*m + n - mp + 1;     // Wigner_d_{n}^{-mp + 1, m}
        
        int H_nmpm1_m = baseind + (2*n + 1)*m + n - mp - 1;     // Wigner_d_{n}^{-mp - 1, m}
        int H_nmp_mm1 = baseind + (2*n + 1)*(m - 1) + n - mp;   // Wigner_d_{n}^{-mp, m - 1}
        int H_nmp_mp1 = baseind + (2*n + 1)*(m + 1) + n - mp;   // Wigner_d_{n}^{-mp, m + 1}

        //the recursion
        Wigner_d[H_nmpm1_m] = inv_d_mp_n*((n + mp)*Wigner_d[H_nmpp1_m] + (n - m + 1)*Wigner_d[H_nmp_mm1] - (n + m + 1)*Wigner_d[H_nmp_mp1]);

        if (m > mp + 1) {                                       // symmetry points
          int H_mpp1_m_mirror = baseind + (2*n + 1)*(mp + 1) + n + m;  // Wigner_d_{n}^{m, mp + 1}
          int H_nmpp1_m_mirror = baseind + (2*n + 1)*(mp + 1) + n - m; // Wigner_d_{n}^{-m, mp + 1}
          Wigner_d[H_mpp1_m_mirror] = -oddoreven(m + mp)*factorm*Wigner_d[H_mpp1_m];
          Wigner_d[H_nmpp1_m_mirror] = -oddoreven(m + mp)*factorm*Wigner_d[H_nmpm1_m];
        }
        factorm *= (double)(n + m + 1)/(n - m);
        
      }
      //step 4
      //the final step where n == m, give a recursion with one term less
      int H_mpp1_n = baseind + (2*n + 2)*n + mp + 1;            // Wigner_d_{n}^{mp + 1, n}
      int H_mpm1_n = baseind + (2*n + 2)*n + mp - 1;            // Wigner_d_{n}^{mp - 1, n}
      int H_mp_nm1 = baseind + 2*n*n - 1 + mp;                  // Wigner_d_{n}^{mp, n - 1}
      
      //the recursion
      Wigner_d[H_mpp1_n] = inv_d_mp_n*((n + mp)*Wigner_d[H_mpm1_n] + Wigner_d[H_mp_nm1]);
      
      //step 5
      int H_nmpp1_n = baseind + (2*n + 2)*n - mp + 1;           // Wigner_d_{n}^{-mp + 1, n}
      
      int H_nmpm1_n = baseind + (2*n + 2)*n - mp - 1;           // Wigner_d_{n}^{-mp - 1, n}
      int H_nmp_nm1 = baseind + 2*n*n - 1 - mp;                 // Wigner_d_{n}^{-mp, n - 1}

      //the recursion
      Wigner_d[H_nmpm1_n] = inv_d_mp_n*((n + mp)*Wigner_d[H_nmpp1_n] + Wigner_d[H_nmp_nm1]);
      if (n > mp + 1) {                                         // symmetry points
        int H_mpp1_n_mirror = baseind + (2*n + 1)*(mp + 1) + 2*n; // Wigner_d_{n}^{n, mp + 1}
        int H_nmpp1_n_mirror = baseind + (2*n + 1)*(mp + 1);      // Wigner_d_{n}^{-n, mp + 1}
        Wigner_d[H_mpp1_n_mirror] = -oddoreven(n + mp)*factorm*Wigner_d[H_mpp1_n];
        Wigner_d[H_nmpp1_n_mirror] = -oddoreven(n + mp)*factorm*Wigner_d[H_nmpm1_n];
      }
    }
    baseind = nextind;
  }
}
