/* io.c */

/* J. Bull 2015-11-18 */

#include "io.h"

#ifndef NOMEX
#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
			   mxGetNumberOfDimensions(A) == 2)
#endif

using namespace std;

// local functions
void ReadCoords(char *file, int& dim1, int& dim2, double **data);
void ReadMasses(char *file, int& dim1, double **data);
void WriteData(const char *out_fname, int N, const double *data, int N2=0, const double *data2=NULL);
void ReadVTK(const char *file, const double *array, int& dim1, int& dim2);
void WriteVTKData(const char *out_fname,
                  int N,const double *data);
#ifndef NOMEX
void ReadMatlab(char *file, int& dim1, int& dim2,double **data);
void WriteMatlabMesh(const char *out_fname, int N,int Nlevels,int nprocs,int rank,
                     mxArray *m, mxArray *c, mxArray *h, mxArray *s, mxArray *t);
void WriteMatlabData(const char *out_fname,
                     int N, const double *data,
                     int N2=0, const double *data2=NULL);
#endif
// Check and parse the command line input, plus allocate output arrays
void ReadInput(int nprocs,int rank, int argc, char *const *argv, char **output_file, int *N, int *NE,
               double **zx,double **zy,double **zz,double **mm,
               double **ex,double **ey,double **ez,
               double *tol,int *pot,int *Ndirect,int *pcoeff,
               int *Nlevels,int *out,
               SMOOTHER *smooth,double *eta,double *theta)
{
  // used to read and write data
  char *in_zname;
  char *in_mname;
  char *in_ename;
  char *in_fname;
  // local char arrays
  char *tempz;
  char *tempm;
  char *tempo;
  char *tempe;
  char *pend;
  char *in_fext;
  string in_fext_s;
  double *array = NULL;
  int dim1,dim2;
  int zflag=0,mflag=0,eflag=0,fflag=0;
  int c;
  int opt_index = 0;
  int sizem;
  int tempi;
  double tempf;

  size_t pos;
  string root;
  string rank_s = to_string(static_cast<long long>(rank));
  //char *tempr; 
  //sprintf(tempr,"%d",rank);
  //string rank_s = string(tempr);
  stringstream tmp;

  // reset global variable optind
  optind = 1;

  if (argc < 3) {
    cout << "Basic usage: " << argv[0] <<
    " -z <point-coords-file> -m <point-masses-file> -f <output-file>" <<
    " (optional: -e <eval-coords-file>). Exiting." << endl;
    return;
  }

  struct option longopts[] = {
    { "coords",      required_argument, &zflag, 'z' },
    { "mass",        required_argument, &mflag, 'm' },
    { "eval",        required_argument, &eflag, 'e' },
    { "output-file", required_argument, &fflag, 'f' },
    { "out",         required_argument, NULL, 'o' },
    { "tol",         required_argument, NULL, 't' },
    { "pot",         required_argument, NULL, 'p' },
    { "Ndirect",     required_argument, NULL, 'N' },
    { "Nlevels",     required_argument, NULL, 'l' },
    //{ "smooth",      required_argument, NULL, 's' },
    { "theta",       required_argument, NULL, 'T' },
    { "eta",         required_argument, NULL, 'n' },
    { "pcoeff",      required_argument, NULL, 'c' },
    { 0, 0, 0, 0 }
  };
  // TODO: find out why catch is not working
  while ((c = getopt_long(argc, argv, "z:m:e:f:o:t:p:N:l:T:n:c:", longopts, &opt_index)) != -1) {

    switch (c) {

      // Points
      case 'z':
        zflag = 1;
        try {
          tempz = optarg;
        }
        catch (exception& e) {
          cout << "Error: must provide valid name of point coordinates file to --coords (-z). " <<
          "Exiting with exception: " << e.what() << endl;
          return;
        }
        break;
      // Masses
      case 'm':
        mflag = 1;
        try {
          tempm = optarg;
        }
        catch (exception& e) {
          cout << "Error: must provide valid name of point masses file to --mass (-m). " <<
          "Exiting with exception: " << e.what() << endl;
          return;
        }
        break;
      // (optional) sampling points
      case 'e':
        eflag = 1;
        try {
          tempe = optarg;
        }
        catch (exception& e) {
          cout << "Error: must provide valid name of evaluation point coords file to --eval (-e). " <<
          "Exiting with exception: " << e.what() << endl;
          return;
        }
        break;
      // file to write out potentials
      case 'f':
        fflag = 1;
        try {
          tempo = optarg;
        }
        catch (exception& e) {
          cout << "Error: must provide valid name of output file to --output-file (-f). " <<
          "Exiting with exception: " << e.what() << endl;
          return;
        }
        break;
      // flag for what variables to write out
      case 'o':
        tempi = strtol(optarg,&pend,10);
        if(!pend[0] && (tempi == -1 || tempi == 0 || tempi == 1 || tempi == 2)) {
          *out = tempi;
        }
        else {
          cout << "Argument to --out (-o) is invalid." << endl;
          cout << "Choose 0 (default: write potential at point coords), " <<
          "1 (write potential at eval pt coords), " <<
          "2 (write potential at points and eval pts) or -1 (write mesh only). " <<
          "Exiting." << endl;
          return;
        }
        break;
      // error tolerance
      case 't':
        tempf = strtof(optarg,&pend);
        if(!pend[0] && tempf >= 0) {
          *tol = tempf;
        }
        else {
          cout << "Argument to --tol (-t) is invalid." << endl;
          cout << "Specify non-negative tolerance. Exiting." << endl;
          return;
        }
        break;
      // type of potential (only inverse supported right now)
      case 'p':
        tempi = strtol(optarg,&pend,10);
        if(!pend[0] && tempi == 1) {
          *pot = tempi;
        }
        else {
          cout << "Argument to --pot (-p) is invalid." << endl;
          cout << "Only a value of 1 (inverse kernel) is accepted at present. Exiting." << endl;
          return;
        }
        break;
      // degree of spherical harmonic expansion
      case 'c':
        tempi = strtol(optarg,&pend,10);
        if(!pend[0] && tempi > 0) {
          *pcoeff = tempi;
        }
        else {
          cout << "Argument to --pcoeff (-c) is invalid. Exiting." << endl;
          return;
        }
        break;
      // max. no. of points in leaf nodes
      case 'N':
        tempi = strtol(optarg,&pend,10);
        if(!pend[0] && tempi > 0) {
          *Ndirect = tempi;
        }
        else {
          cout << "Argument to --Ndirect (-N) is invalid. Exiting." << endl;
          return;
        }
        break;
      // max. no. of levels in tree
      case 'l':
        tempi = strtol(optarg,&pend,10);
        if(!pend[0] && tempi > 0) {
          *Nlevels = tempi;
        }
        else {
          cout << "Argument to --Nlevels (-l) is invalid. Exiting." << endl;
          return;
        }
        
        break;
      // another smoothing parameter
      case 'n':
        tempf = strtof(optarg,&pend);
        if(!pend[0] && tempf >= 0 && tempf <= 1) {
          *eta = tempf;
        }
        else {
          cout << "Argument to --eta (-n) is invalid." << endl;
          cout << "Specify a value in the range [0:1]. Exiting." << endl;
          return;
        }
        break;
      // theta criterion (well-separatedness parameter)
      case 'T':
        tempf = strtof(optarg,&pend);
        if(!pend[0] && tempf >= 0 && tempf <= 1) {
          *theta = tempf;
        }
        else {
          cout << "Argument to --theta (-T) is invalid." << endl;
          cout << "theta must be a value in the range [0:1]. Exiting." << endl;
          return;
        }
        break;
      case '?':
        cout << "?" << endl;
        if (optopt == 'z' || optopt == 'm' || optopt == 'e' ||
            optopt == 'o' || optopt == 't' || optopt == 'p' ||
            optopt == 'N' || optopt == 'n' || optopt == 'f' ||
            optopt == 'l' || optopt == 'T' || optopt == 'c')
          cout << "Error: option -" << (char)optopt << " requires an argument.\n" << endl;
        else if (isprint (optopt))
          cout << "Error: unknown option -" << (char)optopt << "\n" << endl;
        else
          cout << "Error: unknown option character -" << (char)optopt << "\n" << endl;
      default:
        abort ();
    }
  }
  
  // defaults for tol and Ndirect
  if (*tol < 0.0 && *pcoeff < 0)
    *tol = 1e-6;
  else if (*tol > 0 && *pcoeff > 0) {
    cout << "Error: specifying both options 'tol' and 'pcoeff' not allowed. Exiting." << endl;
    return;
  }
  if (*Ndirect < 0 && *Nlevels < 0)
    *Ndirect = 100;
  else if (*Ndirect > 0 && *Nlevels > 0) {
    cout << "Error: specifying both options 'Ndirect' and 'Nlevels' not allowed. Exiting." << endl;
    return;
  }

  // more error checking
  if(zflag==0) {
    cout << "Error: point coords have not been specified. Exiting." << endl;
    return;
  }
  if(mflag==0) {
    cout << "Error: point masses not been specified. Exiting." << endl;
    return;
  }
  if((*out == 1 || *out == 2) && !eflag) {
    cout << "Error: if providing evaluation points via --eval (-e)," <<
    " then output argument '--out (-o)' must be equal to 1 or 2. Exiting." << endl;
    return;
  }

  // 1. points file
  // TODO: find a way to do this safely (no segfault)
  try {
    in_fext = strrchr(tempz,'.');
    in_fext_s = in_fext;
  }
  catch (exception& e) {
    cout << "1st argument must be a valid file name including extension. " <<
    "Exiting with exception: " << e.what() << endl;
  }

  if(nprocs>1) {
    pos = string(tempz).find_last_of('.');
    root = string(tempz).substr(0,pos);
    tmp << root << "_" << rank_s << in_fext_s;
  }
  else
    tmp << tempz;

  in_zname = new char [tmp.str().size()+1];
  strcpy(in_zname,tmp.str().c_str());
  tmp.str("");

  // call relevant reader
  if(in_fext_s == ".mat") {
#ifndef NOMEX
    ReadMatlab(in_zname,dim1,dim2,&array);
#endif
  }
  else if(in_fext_s == ".dat") {
    ReadCoords(in_zname,dim1,dim2,&array);
  }
  else if(in_fext_s == ".vtk") {
  }
  else if(in_fext_s == ".npy") {
  }
  else {
    cout << "Error: invalid input option value=" << in_ename << ". Exiting." << endl;
    return;
  }
  
  assert(array != NULL);
  assert(dim2 == 3);
  
  // assumes that binary file format is {x_0,...,x_N,y_0,...,y_N,z_0,...,z_N}
  *N = dim1;
  *zx = &array[0];
  *zy = &array[(*N)];
  *zz = &array[2*(*N)];
  
  // 2. masses file
  try {
    in_fext = strrchr(tempm,'.');
    in_fext_s = in_fext;
  }
  catch (exception& e) {
    cout << "Error: 2nd argument must be a valid file name including extension. " <<
    "Exiting with exception: " << e.what() << endl;
  }

  if(nprocs>1) {
    pos = string(tempm).find_last_of('.');
    root = string(tempm).substr(0,pos);
    tmp << root << "_" << rank_s << in_fext_s;
  }
  else
    tmp << tempm;

  in_mname = new char [tmp.str().size()+1];
  strcpy(in_mname,tmp.str().c_str());
  tmp.str("");

  // call relevant reader
  if(in_fext_s == ".mat") {
#ifndef NOMEX
    ReadMatlab(in_mname,dim1,dim2,&array);
    // for error checking
    sizem = dim1;
#endif
  }
  else if(in_fext_s == ".dat") {
    ReadMasses(in_mname,dim1,&array);
    sizem = dim1;
  }

  *mm = array;
  
  // 3. eval points file
  if(eflag) {
    try {
      in_fext = strrchr(tempe,'.');
      in_fext_s = in_fext;
    }
    catch (exception& e) {
      cout << "Error: 3rd argument must be a valid file name including extension. " <<
      "Exiting with exception: " << e.what() << endl;
    }
    
    if(nprocs>1) {
      pos = string(tempe).find_last_of('.');
      root = string(tempe).substr(0,pos);
      tmp << root << "_" << rank_s << in_fext_s;
    }
    else
      tmp << tempe;

    in_ename = new char [tmp.str().size()+1];
    strcpy(in_ename,tmp.str().c_str());
    tmp.str("");

    // call relevant reader
    if(in_fext_s == ".mat") {
#ifndef NOMEX
      // read file
      ReadMatlab(in_ename,dim1,dim2,&array);
#endif
    }
    else if(in_fext_s == ".dat") {
      ReadCoords(in_ename,dim1,dim2,&array);
      sizem = dim1;
    }
    else if(in_fext_s == ".vtk") {
    }
    else if(in_fext_s == ".npy") {
    }
    else {
      cout << "Error: invalid input option value=" << in_ename << ". Exiting." << endl;
    }

    if(dim2 != 3) {
      cout << "Error: array must be double matrix of size NE*3. Exiting." << endl;
      return;
    }
    
    *NE = dim1;
    *ex = array;
    *ey = array+(*NE);
    *ez = array+2*(*NE);
  }

  // catch invalid array sizes
  if(*N!=sizem) {
    cout << "Error: sizes of z and m must match. Exiting." << endl;
    return;
  }

  // 4. output file
  // default output file name, type
  // default directory is where fmm3d is executed from ($PWD)
  if(fflag==0) {
  
    // write out potentials to file p.mat
    if(*out != -1)
      tmp << "pot";
    // write out mesh to file mesh.mat
    else
      tmp << "mesh";
    
    if(nprocs>1)
      tmp << "_" << rank_s.c_str();

    // get file extension from coords file
    in_fext = strrchr(tempz,'.');
    tmp << in_fext;
  }
  // if an output file name has been provided
  else {
    // get file extension from output file
    try {
      in_fext = strrchr(tempo,'.');
      in_fext_s = in_fext;
    }
    catch (exception& e) {
      cout << "Error: output file must be a valid file name including directory and file extension. " <<
      "Exiting with exception: " << e.what() << endl;
    }

    if(nprocs>1) {
      pos = string(tempo).find_last_of('.');
      root = string(tempo).substr(0,pos);
      tmp << root << "_" << rank_s << in_fext_s;
    }
    else {
      tmp << tempo;
    }
  }
  // put string into char* output_file
  *output_file = new char[tmp.str().size()+1];
  strcpy(*output_file,tmp.str().c_str());
  tmp.str("");
}

#ifndef NOMEX
void WriteMesh(int rank, int out, char *output_file, int N, int NE,
                 int Nlevels,int nprocs,
                 mxArray *m, mxArray *c, mxArray *h, mxArray *s, mxArray *t)
{
  char *out_fext = strrchr(output_file,'.');
  string out_fext_s = out_fext;
  
  // call relevant reader
  if(out_fext_s == ".mat") {
    WriteMatlabMesh(output_file,N,Nlevels,nprocs,rank,m,c,h,s,t);
  }
  else {
    cout << "Error: invalid output file format specified. Exiting." << endl;
    return;
  }
}
#endif

void WritePotential(int rank, int out, char *output_file, int N, int NE,
                 int Nlevels,int nprocs,
                 const double *pr,const double *qr)
{
  char *out_fext = strrchr(output_file,'.');
  string out_fext_s = out_fext;
  
  // call relevant reader
  if(out_fext_s == ".mat") {
#ifndef NOMEX
    if(out == 0) {
      WriteMatlabData(output_file,N,pr);
    }
    else if(out == 1) {
      WriteMatlabData(output_file,NE,qr);
    }
    else if(out == 2) {
      WriteMatlabData(output_file,N,pr,NE,qr);
    }
#endif
  }
  else if(out_fext_s == ".dat") {
    if(out == 0) {
      WriteData(output_file,N,pr);
    }
    else if(out == 1) {
      WriteData(output_file,NE,qr);
    }
    else if(out == 2) {
      WriteData(output_file,N,pr,NE,qr);
    }
  }
  // Write out potentials to VTK files.
  else if(out_fext_s == ".vtk") {
    if(out == 0) {
      //WriteVTKData(output_file,N,pr);
    }
    else if(out == 1) {
      //WriteVTKData(NE,qr,qi);
    }
    else if(out == 2) {
      //WriteVTKData(N,pr,pi,NE,qr,qi);
    }
  }
  // NumPy format? Points and masses in same file.
  else if(out_fext_s == ".npy") {
    //WriteNumpy();
  }
  else {
    cout << "Error: invalid output file format specified. Exiting." << endl;
    return;
  }

}

#ifndef NOMEX
// read real-valued data from Matlab binary file
void ReadMatlab(char *in_fname, int& dim1, int& dim2, double **data)
{
  MATFile *zfile;
  const char **dir;
  const char *name;
  int ndir,i;
  mxArray *pa;
  
  zfile = matOpen(in_fname, "r");
  if(zfile == NULL) {
    cout << "Error opening Matlab MAT file " << in_fname << ". Exiting." << endl;
    return;
  }
  
  // get list of variables
  dir = (const char **)matGetDir(zfile,&ndir);
  if(dir == NULL) {
    cout << "Error reading Matlab directory of file " << in_fname << ". Exiting." << endl;
    return;
  }
  mxFree(dir);
  
  // In order to use matGetNextXXX correctly, reopen file
  if (matClose(zfile) != 0) {
    cout << "Error closing file " << in_fname << ". Exiting." << endl;
    return;
  }
  
  zfile = matOpen(in_fname, "r");
  if(zfile == NULL) {
    cout << "Error opening Matlab MAT file " << in_fname << ". Exiting." << endl;
    return;
  }
  
  //for (i=0; i<ndir; i++) {
    pa = matGetNextVariable(zfile, &name);
    if (pa == NULL) {
      cout << "Error reading data in file " << in_fname << ". Exiting." << endl;
      return;
    }
  
    // check that data is type real double
    if(!ISDOUBLEMATRIX(pa)) {
      cout << "ERROR: Array must be double (real or complex) matrix. Exiting." << endl;
      return;
    }
    // set output
#if MX_HAS_INTERLEAVED_COMPLEX
    *data = mxGetDoubles(pa);
#else
    *data = mxGetPr(pa);
#endif
    dim1 = mxGetM(pa);
    dim2 = mxGetN(pa);
  //}
  
  if (matClose(zfile) != 0) {
    cout << "Error closing file " << in_fname << ". Exiting." << endl;
    return;
  }
}
#endif

// read coordinate data from binary file
void ReadCoords(char *in_fname, int& dim1, int& dim2, double **data)
{
  ifstream file;
  int size;
  dim1=0,dim2=0;
  double *pa;
  
  file.open(in_fname, ios::in | ios::binary);
  
  if(file.is_open()) {
  
    // get size of data
    struct stat res;
    if(stat(in_fname,&res) == 0) {
      size = res.st_size;
      if(size % 24 != 0) {
        cout << "Error: input data dimensions are not consistent with expected dimensions." << endl;
        cout << "Check that data is type(double) and in the format(x_0,...,x_N,y_0,...,y_N,z_0,...,z_N)" << endl;
        return;
      }
      else {
        // this is N
        dim1 = size/24;
        // 3D data
        dim2 = 3;
      }
    }
    else {
      cout << "Error getting in_file size" << endl;
      return;
    }
    
    pa = new double [dim1*dim2];
    file.read((char *) pa, size);

    // copy to output pointer
    *data = pa;

    file.close();
  }
  else {
    cerr << "Error opening binary coords file " << in_fname << ". Exiting." << endl;
    return;
  }
}

// read mass data from binary file
void ReadMasses(char *in_fname, int& dim1, double **data)
{
  ifstream file;
  int size;
  dim1=0;
  double *pa;
  
  file.open(in_fname, ios::in | ios::binary);
  
  if(file.is_open()) {
  
    // get size of data
    struct stat res;
    if(stat(in_fname,&res) == 0) {
      size = res.st_size;
      if(size % 8 != 0) {
        cout << "Error: input data dimensions are not consistent with expected dimensions." << endl;
        cout << "Check that data is type(double) and in the format(m_0,...,m_N)" << endl;
        return;
      }
      else {
        dim1 = size/8;
      }
    }
    else {
      cout << "Error getting in_file size" << endl;
      return;
    }
    
    pa = new double [dim1];
    file.read((char *) pa, size);
    
    // copy to output pointer
    *data = pa;
    
    file.close();
  }
  else {
    cerr << "Error opening binary masses file " << in_fname << ". Exiting." << endl;
    return;
  }
}

#ifndef NOMEX
// write out an array data to file. If int N2 and array data2 are provided, write them too.
void WriteMatlabData(const char *out_fname, int N, const double *data, int N2, const double *data2)
{
  MATFile *pfile;
  int status;
  mxArray *pa;

  if(data == NULL) {
    printf("Error writing data to file %s on line %d\n", __FILE__, __LINE__);
    printf("Exiting.");
    return;
  }

//  cout << "writing Matlab file " << out_fname << endl;
  
  // "w" overwrites previous file and data, "u" doesn't.
  // TODO: if solving a time-dependent problem,
  // enable a sequence of potentials to be written to a sequence of files.
  pfile = matOpen(out_fname, "w");
  if(pfile == NULL) {
    printf("Error opening Matlab MAT file %s on line %d\n", __FILE__, __LINE__);
    printf("Exiting.");
    return;
  }
  
  pa = mxCreateDoubleMatrix(N,1,mxREAL);
  if (pa == NULL) {
    printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
    printf("Unable to create mxArray. Exiting.");
    return;
  }

#if MX_HAS_INTERLEAVED_COMPLEX
  memcpy(mxGetDoubles(pa), data, N*sizeof(double));
#else
  memcpy(mxGetPr(pa), data, N*sizeof(double));
#endif
  
  status = matPutVariable(pfile, "p", pa);
  if (status != 0) {
    printf("%s : Error using matPutVariable on line %d\n", __FILE__, __LINE__);
    printf("Exiting.");
    return;
  }

  // write data2, if given
  if(N2 != 0 && data2 == NULL) {
    printf("Error writing data2: NULL pointer. Exiting.");
    return;
  }
  else if(N2 == 0 && data2 != NULL) {
    printf("Error writing data2: N2 = 0. Exiting.");
    return;
  }
  else if(N2 !=0 && data2 != NULL) {
    pa = mxCreateDoubleMatrix(N2,1,mxREAL);
    if (pa == NULL) {
      printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
      printf("Unable to create mxArray. Exiting.");
      return;
    }

#if MX_HAS_INTERLEAVED_COMPLEX
    memcpy(mxGetDoubles(pa), data2, N2*sizeof(double));
#else
    memcpy(mxGetPr(pa), data2, N2*sizeof(double));
#endif
    
    status = matPutVariable(pfile, "q", pa);
    if (status != 0) {
      printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
      printf("Exiting.");
      return;
    }
  }

  if (matClose(pfile) != 0) {
    printf("Error closing file %s\n",out_fname);
    printf("Exiting.");
    return;
  }

  mxDestroyArray(pa);
}
#endif

// write out an array data to binary file. If int N2 and array data2 are provided, write them too.
void WriteData(const char *out_fname, int N, const double *data, int N2, const double *data2)
{
  ofstream file;

  if(data == NULL) {
    cout << "Error writing data to file. Exiting." << endl;
    return;
  }

  file.open(out_fname, ios::out | ios::binary);
  
  if(file.is_open()) {
  
    // write data
    file.write((char *) data, N*8);
    file.close();
  }
  else {
    cout << "Error opening binary file " << out_fname << ". Exiting." << endl;
    return;
  }

  // write data2, if given
  if(N2 != 0 && data2 == NULL) {
    cout << "Error writing data2: NULL pointer. Exiting." << endl;
    return;
  }
  else if(N2 == 0 && data2 != NULL) {
    cout << "Error writing data2: N2 = 0. Exiting." << endl;
    return;
  }
  else if(N2 !=0 && data2 != NULL) {
    file.open(out_fname, ios::app | ios::binary);

    if(file.is_open()) {
      file.write((char *) data2, N2*8);
      file.close();
    }
    else {
      cout << "Error re-opening binary file " << out_fname << ". Exiting." << endl;
      return;
    }
  }
}

void WriteVTKData(const char *out_fname, int N, const double *data)
{
  // rank 0 write headers, all ranks write to subfolders
  // use VTK particles file format for points,
  // hierarchical box data format for mesh?
}

#ifndef NOMEX
void WriteMatlabMesh(const char *out_fname, int N, int Nlevels, int nprocs, int rank,
                     mxArray *m, mxArray *c, mxArray *h, mxArray *s, mxArray *t)
{
  MATFile *pfile;
  int status;
  mxArray *pa;

  // "w7.3" enables storage of large objects (>2GB) by using HDF5 format.
  pfile = matOpen(out_fname, "w7.3");
  if(pfile == NULL) {
    printf("Error opening Matlab MAT file %s on line %d\n", __FILE__, __LINE__);
    return;
  }
  
  if(m != NULL) {
    //cout << "writing mesh to Matlab file " << out_fname << endl;
    for(int i=0; i<=Nlevels; i++) {
      const char *name = ("M"+to_string(i)).c_str();
      pa = mxGetCell(m,i);
      if (pa == NULL) {
        printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
        printf("Unable to create mxArray. Exiting.");
        return;
      }
    
      // write m level by level:
      status = matPutVariable(pfile, name, pa);
      if (status != 0) {
        printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
        printf("Exiting.");
        return;
      }
      pa = NULL;
    }
    // write m directly:
//    status = matPutVariable(pfile, "M", m);
  }
  
  // write connect, if given
  if(c != NULL) {
    for(int i=0; i<=Nlevels; i++) {
      const char *name = ("C"+to_string(i)).c_str();
      pa = mxGetCell(c,i);
      if (pa == NULL) {
        printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
        printf("Unable to create mxArray. Exiting.");
        return;
      }
    
      // write c level by level:
      status = matPutVariable(pfile, name, pa);
      if (status != 0) {
        printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
        printf("Exiting.");
        return;
      }
      pa = NULL;
    }
    // write c directly:
    //status = matPutVariable(pfile, "C", c);
  }

  // write halo, if given
  if(h != NULL) {
    for(int i=0; i<=Nlevels; i++) {
      mwIndex ii = i; // not strictly necesary
      const char *name = ("H"+to_string(i)).c_str();
      pa = mxGetCell(h,i);
      
      if (pa == NULL) {
        printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
        printf("Unable to create mxArray. Exiting.");
      }
      
      // write h level by level:
      status = matPutVariable(pfile, name, pa);
      if (status != 0) {
        printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
        printf("Exiting.");
        return;
      }
      
      pa = NULL;
    }
    // write h directly:
//    status = matPutVariable(pfile, "H", h);
  }

  // write sparse, if given
  if(s != NULL) {
    pa = mxCreateDoubleMatrix(sizeof(s),1,mxREAL);
    if (pa == NULL) {
      printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
      printf("Unable to create mxArray. Exiting.");
      return;
    }

#if MX_HAS_INTERLEAVED_COMPLEX
    memcpy((void *)(mxGetDoubles(pa)), (void *)s, sizeof(s));
#else
    memcpy((void *)(mxGetPr(pa)), (void *)s, sizeof(s));
#endif
    
    status = matPutVariable(pfile, "S", pa);
    if (status != 0) {
      printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
      printf("Exiting.");
      return;
    }
  }

  // write evaluation point matrix, if given
  if(t != NULL) {
    pa = mxCreateDoubleMatrix(sizeof(t),1,mxREAL);
    if (pa == NULL) {
      printf("%s : Out of memory on line %d\n", __FILE__, __LINE__);
      printf("Unable to create mxArray. Exiting.");
      return;
    }

#if MX_HAS_INTERLEAVED_COMPLEX
    memcpy((void *)(mxGetDoubles(pa)), (void *)t, sizeof(t));
#else
    memcpy((void *)(mxGetPr(pa)), (void *)t, sizeof(t));
#endif
    
    status = matPutVariable(pfile, "T", pa);
    if (status != 0) {
      printf("%s :  Error using matPutVariable on line %d\n", __FILE__, __LINE__);
      printf("Exiting.");
      return;
    }
  }

  // close output file
  if (matClose(pfile) != 0) {
    printf("Error closing file %s\n",out_fname);
    printf("Exiting.");
    return;
  }
  
  // clean up memory
  mxDestroyArray(pa);
}
#endif

