function make(varargin)
%MAKE Makefile for FMM3D.
%   MAKE(ARG1,...) compiles FMM3D with the compiler arguments ARG1,
%   ... using the MEX-command. With no arguments, default
%   platform-dependent define arguments are used (the -D switch).

%   TODO: update to latest options and defines.

% S. Engblom 2020-02-11 (mexa64, 9.6)
% J. Bull    2017-09-02 (mexmaci64, 8.5)
% S. Engblom 2011-04-17 (mexmaci64, 7.11)
% S. Engblom 2011-04-17 (mexmaci64, 7.10)
% S. Engblom 2011-03-07 (mexa64, 7.11)
% S. Engblom 2010-02-02 (mexs64, 7.7)
% S. Engblom 2010-02-02 (mexa64, 7.8)
% S. Engblom 2010-01-13 (mexmaci)
% S. Engblom 2007-09-01 (mexs64)
% S. Engblom 2006-07-10 (mexa64)
% S. Engblom 2005-03-22 (mexmac)

% Public defines:
%
% -use '-DmwIndex=int' for platforms where mex is not automatically
% linked against the library defining the mwIndex type,
%
% -use '-DPRINTTIME' to print precise timings of the FMM-algorithm,
%
% For experimental use:
%
% '-DFMM_INLINECOMPLEX' (recommended) avoids complex arithmetic in
% various places in the FMM-algorithm and uses inline versions
% instead.  TODO: update to work with new Matlab interleaved complex
% type,
%
% '-DFMM_INLINECOMPLEX_DIRECT' (Not implemented in 3D code) uses
% "Quick-and-dirty" versions of direct evaluations. On some platforms,
% it is faster by a factor of up to 8 not to rely on certain
% math-functions and/or complex arithmetics, but to write the complex
% arithmetic inline,
%
% '-DFMM_SHRINK' (recommended) sets Postsplit shrinking of boxes on/off,
%
% '-DFMM_THETA=<value>' (default 0.5) sets the constant of
% well-separatedness in the theta-criterion,
%
% '-DFMM_MESHRATIO=<value>' (Not implemented in 3D code. default 2.0)
% tunes the criterion for splitting long edges in the mesh.
%
% Make some effort in finding good optimization flags!

mx = mexext;
ver = version;

% possible defines
define = {'-DmwIndex=int' ...              % (1)
          '-DPRINTTIME' ...                % (2)
          '-DFMM_INLINECOMPLEX' ...        % (3)
          '-DFMM_INLINECOMPLEX_DIRECT' ... % (4) 
          '-DFMM_SHRINK'};                 % (5)
% (edit defaults below)

% path, include and link path, source
s = pwd;
include = ['-I' s '/source'];
link = ['-L' s '/source'];

source = {[s '/source/fmm3d.c'] ...
          [s '/source/fmm.c'] ...
          [s '/source/fmmshift.c'] ...
          [s '/source/direct.c']};

if strcmp(mx,'mexa64')
  if ~strncmp(version,'7.8',3) && ~strncmp(version,'7.11',4) ...
        && ~strncmp(version,'8.4',3) && ~strncmp(version,'9.6',3)
    warning(['Extension .' mexext ...
             ' tested with Matlab version(s) 7.8, 7.11, 8.4, and 9.6 only.']);
  end
  cflags = ['CFLAGS=-fPIC -fno-omit-frame-pointer -std=c99 -O3 ' ...
            '-D_GNU_SOURCE -pthread -fexceptions'];
  link = link;
  if strncmp(version,'7.11',4)
    link = [link ' -L' matlabroot '/sys/os/glnxa64'];
  end
  define = define(3:5);
  if nargin > 0, define = varargin; end
  mex(cflags,define{:},include,link, ...
      '-largeArrayDims','-outdir',s, ...
      source{:});
elseif strcmp(mx,'mexmaci64')
  if ~strncmp(version,'8.4',3)
    warning(['Extension .' mexext ...
             ' tested with Matlab version(s) 8.4.']);
  end
  cflags = 'CC=clang -std=c99 -fast -w';
  link = link;
  
  define = define(3:5);
  if nargin > 0, define = varargin; end
  mex(cflags,define{:},include,link, ...
      '-largeArrayDims','-outdir',s, ...
      source{:});
elseif strcmp(mx,'mexw32')
  warning('Platform never tried.');
  cflags = ['COMPFLAGS=-c -DMATLAB_MEX_FILE -fno-omit-frame-pointer -std=c99' ...
            '-D_GNU_SOURCE -fexceptions'];
  include(include == '/') = '\';
  link(link == '/') = '\';
  for i = 1:numel(source)
    source{i}(source{i} == '/') = '\';
  end
  define = define(3:5);
  if nargin > 0, define = varargin; end
  mex(cflags,define{:},include,link, ...
      '-outdir',s, ...
      source{:},[matlabroot '\extern\lib\win32\microsoft\libmwblas.lib']); 
elseif strcmp(mx,'mexw64')
  warning('Platform never tried.');
  cflags = ['COMPFLAGS=-c -DMATLAB_MEX_FILE ' ...
            '-D_GNU_SOURCE -TP -EHsc -FAsc'];
  include(include == '/') = '\';
  link(link == '/') = '\';
  for i = 1:numel(source)
    source{i}(source{i} == '/') = '\';
  end
  define = define(2:5);
  if nargin > 0, define = varargin; end
  mex(cflags,define{:},include,link, ...
      '-outdir',s, '-largeArrayDims', ...
      source{:}); 
else
  error('Unknown platform.');
end

