function meshstats(M,C)
%MESHSTATS Display mesh statistics.
%   MESHSTATS(M,C) displays lots of interesting information of the
%   FMM-mesh M with connectivity matrix C.
%
%   No error-checking is performed.
%
%   See also FMM3D, MESHPLOT, MESHQUAL.

% S. Engblom 2010-03-10

nlevels = size(C,2);
fprintf('\n');
fprintf('FMM mesh statistics\n');
fprintf('-------------------\n\n');
fprintf(['Mesh: %d levels, %d boxes at the finest level, %d boxes ' ...
         'in total.\n\n'],nlevels,size(M{end},2),sum(cellfun('prodofsize',M)/6));

% expand to full format
if nnz(diag(C{1})) == 0
  for l = 1:nlevels
    C{l} = speye(size(C{l}))+C{l}''+C{l}';
  end
end

percent = 1;
if(percent)
  fprintf('Connectivity per box (mean +- std (percent)):\n');
else
  fprintf('Connections per box (mean +- std):\n');
end
far_all=0;
near_all=0;
zero_all=0;
factor = 1;
for l = 1:nlevels
  factor = 100/8^(l-1);
  far = full(sum(C{l} == -1));
  near = full(sum(C{l} == 1));
  zero = full(sum(C{l} == 0));
  far_all=far_all+mean(far)*8^(l-1);
  near_all=near_all+mean(near)*8^(l-1);
  zero_all=zero_all+mean(zero)*8^(l-1);
  if(percent)
    fprintf(['  Level  %3d: %% far = (%5.2f +- %5.2f), ' ...
               '%% near = (%5.2f +- %4.2f), ' ...
               '%% unconnected = (%5.2f +- %4.2f),\n'], ...
              l-1,mean(far)*factor,std(far)*factor,mean(near)*factor, ...
              std(near)*factor,mean(zero)*factor,std(zero))*factor;
  else
    fprintf(['  Level  %3d: N far = (%5.2f +- %5.2f), ' ...
               'N near = (%5.2f +- %4.2f), ' ...
               'N unconnected = (%5.2f +- %4.2f),\n'], ...
              l-1,mean(far),std(far),mean(near), ...
              std(near),mean(zero),std(zero));
  end
end
total=far_all+near_all+zero_all;
if(percent)
  fprintf(['  All levels total: %% far = %5.1f, ' ...
               '%% near = %5.1f, ' ...
               '%% unconnected = %5.1f\n\n'], ...
              far_all/total*100.,near_all/total*100.,zero_all/total*100.);
else
  fprintf(['  All levels total: N far = %5.1f, ' ...
               'N near = %5.1f, ' ...
               'N unconnected = %5.1f\n\n'], ...
              far_all,near_all,zero_all);
end
fprintf('Miscellaneous information:\n');
for l = 1:nlevels
  r = sqrt(M{l}(4,:).^2+M{l}(5,:).^2+M{l}(6,:).^2);
  [ii,jj] = find(C{l});
  dsq = (M{l}(1,ii)-M{l}(1,jj)).^2 + (M{l}(2,ii)-M{l}(2,jj)).^2 + (M{l}(3,ii)-M{l}(3,jj)).^2;
  theta0 = max(r(ii),r(jj)).^2+0.5*min(r(ii),r(jj)).^2 < 0.5*0.5/(1+0.5)*dsq;
  theta1 = max(r(ii),r(jj))+0.5*min(r(ii),r(jj)) < 0.5*sqrt(dsq);
  fprintf(['  Level %3d: simplified theta-criterion positive %4.1f%%,\n' ...
           '            ditto false and should be false %4.1f%%,\n' ...
           '            ditto false and should be true %4.1f%%.\n'], ...
          l-1,100*mean(theta0), ...
          100*mean(~theta0 & ~theta1), ...
          100*mean(~theta0 & theta1));
end
fprintf('\n');
