function partition_points(xyzm,p,N,base)

% partition a set of points/masses geometrically into p bins.
% the bins are arranged into a compact grid.
% load balancing is achieved by splitting the number of points
% equally between the 2 children of a splitting operation.
% user has to supply the points in an N*4 array
% where each row contains x,y,z,m.

% simple example
%rand('state',131345);
%xx = rand(N,1);
%yy = rand(N,1);
%zz = rand(N,1);
%m = rand(N,1);
%xyzm = cat(2,xx,yy,zz,m);

dz=floor(N/p);

if(p==1),
  z = xyzm(:,1:3);
  m = xyzm(:,4);
  save(strcat(base,'z',int2str(N),'.mat'),'z');
  save(strcat(base,'m',int2str(N),'.mat'),'m');
  return;
end

%find 3 integer factors to split in (x,y,z) axes
[a,b,c] = findthreeintegerfactors(p);
split = sort([a,b,c],'descend');

% sort sequentially: x, y, z
zsort = sortcoords(xyzm,split,p,N);

% make sub-arrays for each process
for j=1:p,
  [j,min(zsort(dz*(j-1)+1:dz*j,1))/1.e20,max(zsort(dz*(j-1)+1:dz*j,1))/1.e20,min(zsort(dz*(j-1)+1:dz*j,2))/1.e20,max(zsort(dz*(j-1)+1:dz*j,2))/1.e20,min(zsort(dz*(j-1)+1:dz*j,3))/1.e20,max(zsort(dz*(j-1)+1:dz*j,3))/1.e20]
  % save arrays to file
  z = zsort(dz*(j-1)+1:dz*j,1:3);
  m = zsort(dz*(j-1)+1:dz*j,4);
  save(strcat(base,'z',int2str(N),'_',int2str(j-1),'.mat'),'z');
  save(strcat(base,'m',int2str(N),'_',int2str(j-1),'.mat'),'m');
end

% helper functions

% function to recursively sort point coords
% WARNING: may not work for large p and odd N
function xyzm = sortcoords(xyzm,split,p,N)
  [A,I] = sort(xyzm(:,1));
  xyzm = xyzm(I,:);
  tmp2=zeros(size(xyzm));
  for j = 1:split(1)
    div = floor(N/split(1));
    st = (j-1)*div+1;
    en = j*div;
    if(j==split(1))
      en = en + rem(N,split(1));
    end
    [A,I] = sort(xyzm(st:en,2));
    tmp = xyzm(st-1+I,:);

    for k = 1:split(2)
      div2 = floor(div/split(2));
      st2 = (k-1)*div2+1;
      en2 = k*div2;
      if(k==split(2))
        en2 = en2 + rem(div,split(2));
      end
      [A,I] = sort(tmp(st2:en2,3));
      tmp2(st-1+st2:st-1+en2,:) = tmp(st2-1+I,:);
    end
  end
  xyzm=tmp2;
  %return;

% function to factorise an integer into the 3 closest factors
function [a, b, c] =  findthreeintegerfactors(n)
  % a cannot be greater than the cube root of n
  % b cannot be smaller than the cube root of n
  % we get the maximum allowed value of a
  amax = floor(nthroot(n,3));
  if amax^3 == n
    % special case where n is a cube number
    a = amax;
    b = a;
    c = a;
    return;
  end
  % Get its prime factors of n
  primeFactors  = factor(n);
  % Start with a factor 1 in the list of candidates for a
  candidates = [1];
  for i=1:numel(primeFactors)
    % get the next prime factor
    f = primeFactors(i);
    % Add new candidates which are obtained by multiplying
    % existing candidates with the new prime factor f
    % Set union ensures that duplicate candidates are removed
    candidates  = union(candidates, f .* candidates);
    % throw out candidates which are larger than amax
    candidates(candidates > amax) = [];
  end
  % Take the largest factor in the list d
  a = candidates(end);
  % call similar function for remaining 2 factors
  [b, c] = findtwointegerfactors(n/a);


% function to factorise an integer into the 2 closest factors
function [a, b] =  findtwointegerfactors(n)
  % a cannot be greater than the square root of n
  % b cannot be smaller than the square root of n
  % we get the maximum allowed value of a
  amax = floor(sqrt(n));
  if 0 == rem(n, amax)
    % special case where n is a square number
    a = amax;
    b = n / a;
    return;
  end
  % Get its prime factors of n
  primeFactors  = factor(n);
  % Start with a factor 1 in the list of candidates for a
  candidates = [1];
  for i=1:numel(primeFactors)
    % get the next prime factr
    f = primeFactors(i);
    % Add new candidates which are obtained by multiplying
    % existing candidates with the new prime factor f
    % Set union ensures that duplicate candidates are removed
    candidates  = union(candidates, f .* candidates);
    % throw out candidates which are larger than amax
    candidates(candidates > amax) = [];
  end
  % Take the largest factor in the list d
  a = candidates(end);
  b = n / a;

